@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013 Red Hat.
  * Copyright (c) 1995,2003,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -108,7 +109,7 @@ agent_creds(__pmPDU *pb)
 	handshake[0].c_vala = PDU_VERSION;
 	handshake[0].c_valb = 0;
 	handshake[0].c_valc = 0;
-	if ((sts = __pmSendCreds(outfd, getpid(), 1, handshake)) < 0)
+	if ((sts = __pmSendCreds(outfd, (int)getpid(), 1, handshake)) < 0)
 	    return sts;
     }
 
@@ -120,23 +121,25 @@ pmdaversion(void)
 {
     int		sts;
     __pmPDU	*ack;
+    int		pinpdu;
 
-    sts = __pmGetPDU(infd, ANY_SIZE, _creds_timeout, &ack);
+    pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, _creds_timeout, &ack);
     if (sts == PDU_CREDS) {
 	if ((sts = agent_creds(ack)) < 0) {
 	    fprintf(stderr, "Warning: version exchange failed "
 		"for PMDA %s: %s\n", myPmdaName, pmErrStr(sts));
-	    return;
 	}
     }
     else {
 	if (sts < 0)
 	    fprintf(stderr, "__pmGetPDU(%d): %s\n", infd, pmErrStr(sts));
-	fprintf(stderr, "Warning: no version exchange with PMDA %s: "
-			"assuming PCP 1.x PMDA.\n", myPmdaName);
-	__pmSetVersionIPC(infd, PDU_VERSION1);
-	__pmSetVersionIPC(outfd, PDU_VERSION1);
+	else
+	    fprintf(stderr, "pmdaversion: expecting PDU_CREDS, got PDU type %d\n", sts);
+	fprintf(stderr, "Warning: no version exchange with PMDA %s\n",
+			myPmdaName);
     }
+    if (pinpdu > 0)
+	__pmUnpinPDUBuf(ack);
 }
 
 void
@@ -159,7 +162,7 @@ openpmda(char *fname)
 	printf(" %s", param.argv[i]);
     putchar('\n');
 
-    if (__pmProcessCreate(param.argv, &infd, &outfd) < 0) {
+    if (__pmProcessCreate(param.argv, &infd, &outfd) < (pid_t)0) {
 	fprintf(stderr, "openpmda: create process: %s\n", osstrerror());
     }
     else {
@@ -172,9 +175,9 @@ openpmda(char *fname)
     }
 }
 
-#ifdef HAVE_SYS_UN_H
+#ifdef HAVE_STRUCT_SOCKADDR_UN
 void
-opensocket(char *fname)
+open_unix_socket(char *fname)
 {
     int			fd;
     struct stat		buf;
@@ -219,12 +222,69 @@ opensocket(char *fname)
 }
 #else
 void
-opensocket(char *fname)
+open_unix_socket(char *fname)
 {
-	__pmNotifyErr(LOG_CRIT, "UNIX domain sockets unsupported\n");
+    __pmNotifyErr(LOG_CRIT, "UNIX domain sockets unsupported\n");
 }
 #endif
 
+static void
+open_socket(int port, int family, const char *protocol)
+{
+    __pmSockAddr	*addr;
+    int			fd, sts;
+    char		socket[64];
+
+    fd = (family == AF_INET) ? __pmCreateSocket() : __pmCreateIPv6Socket();
+    if (fd < 0) {
+	fprintf(stderr, "opensocket: socket: %s\n", netstrerror());
+	return;
+    }
+
+    addr = __pmLoopBackAddress(family);
+    if (addr == NULL) {
+	fprintf(stderr, "opensocket: loopback: %s\n", netstrerror());
+	__pmCloseSocket(fd);
+	return;
+    }
+
+    closepmda();
+
+    sts = __pmConnectTo(fd, addr, port);
+    __pmSockAddrFree(addr);
+
+    if (sts < 0) {
+	fprintf(stderr, "opensocket: connect: %s\n", netstrerror());
+	__pmCloseSocket(fd);
+	return;
+    }
+
+    infd = fd;
+    outfd = fd;
+
+    sprintf(socket, "%s port %d", protocol, port);
+    printf("Connect to PMDA on %s\n", socket);
+
+    connmode = CONN_DAEMON;
+    reset_profile();
+    if (myPmdaName != NULL)
+	free(myPmdaName);
+    myPmdaName = strdup(socket);
+    pmdaversion();
+}
+
+void
+open_inet_socket(int port)
+{
+    open_socket(port, AF_INET, "inet");
+}
+
+void
+open_ipv6_socket(int port)
+{
+    open_socket(port, AF_INET6, "ipv6");
+}
+
 void
 closepmda(void)
 {
@@ -246,12 +306,13 @@ closepmda(void)
 int
 dopmda_desc(pmID pmid, pmDesc *desc, int print)
 {
-    int sts;
-    __pmPDU *pb;
-    int i;
+    int		sts;
+    __pmPDU	*pb;
+    int		i;
+    int		pinpdu;
 
     if ((sts = __pmSendDescReq(outfd, FROM_ANON, pmid)) >= 0) {
-	if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_DESC) {
+	if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_DESC) {
 	    if ((sts = __pmDecodeDesc(pb, desc)) >= 0) {
 		if (print)
 		    __pmPrintDesc(stdout, desc);
@@ -273,6 +334,9 @@ dopmda_desc(pmID pmid, pmDesc *desc, int print)
 	    printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 	else
 	    printf("Error: __pmGetPDU() failed: wrong PDU (%x)\n", sts);
+
+	if (pinpdu > 0)
+	    __pmUnpinPDUBuf(pb);
     }
     else
 	printf("Error: __pmSendDescReq() failed: %s\n", pmErrStr(sts));
@@ -285,20 +349,23 @@ dopmda(int pdu)
 {
     int			sts;
     pmDesc		desc;
-    pmDesc		*desc_list = NULL;	/* initialize to pander to gcc */
-    pmResult		*result;
+    pmDesc		*desc_list = NULL;
+    pmResult		*result = NULL;
     __pmInResult	*inresult;
     __pmPDU		*pb;
     int			i;
     int			j;
     int			ident;
+    int			length;
     char		*buffer;
     struct timeval	start;
     struct timeval	end;
+    char		name[32];
     char		**namelist;
     int			*statuslist;
     int			numnames;
     pmID		pmid;
+    int			pinpdu;
 
     if (timer != 0)
 	__pmtimevalNow(&start);
@@ -328,7 +395,7 @@ dopmda(int pdu)
 			return;
                     }
 		} 
-            }/*get_desc*/
+            }
 
 	    sts = 0;
 	    if (profile_changed) {
@@ -339,7 +406,7 @@ dopmda(int pdu)
 	    }
 	    if (sts >= 0) {
 		if ((sts = __pmSendFetch(outfd, FROM_ANON, 0, NULL, param.numpmid, param.pmidlist)) >= 0) {
-		    if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_RESULT) {
+		    if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_RESULT) {
 			if ((sts = __pmDecodeResult(pb, &result)) >= 0) {
 			    if (get_desc) 
 				_dbDumpResult(stdout, result, desc_list);
@@ -360,18 +427,25 @@ dopmda(int pdu)
 			printf("Error: __pmGetPDU() failed: PDU empty, PMDA may have died\n");
 		    else
 			printf("Error: __pmGetPDU() failed: wrong PDU (%x)\n", sts);
+
+		    if (pinpdu > 0)
+			__pmUnpinPDUBuf(pb);
 		}
 		else
 		    printf("Error: __pmSendFetch() failed: %s\n", pmErrStr(sts));
 	    }
+	    if (desc_list)
+		free(desc_list);
 	    break;
 
 	case PDU_INSTANCE_REQ:
 	    printf("pmInDom: %s\n", pmInDomStr(param.indom));
 	    if ((sts = __pmSendInstanceReq(outfd, FROM_ANON, &now, param.indom, param.number, param.name)) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_INSTANCE) {
-		    if ((sts = __pmDecodeInstance(pb, &inresult)) >= 0)
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_INSTANCE) {
+		    if ((sts = __pmDecodeInstance(pb, &inresult)) >= 0) {
 			printindom(stdout, inresult);
+			__pmFreeInResult(inresult);
+		    }
 		    else
 			printf("Error: __pmDecodeInstance() failed: %s\n", pmErrStr(sts));
 		}
@@ -383,6 +457,9 @@ dopmda(int pdu)
 		}
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else
 		printf("Error: __pmSendInstanceReq() failed: %s\n", pmErrStr(sts));
@@ -407,9 +484,10 @@ dopmda(int pdu)
 	    }
 
 	    printf("Getting Result Structure...\n");
+	    pinpdu = 0;
 	    if ((sts = __pmSendFetch(outfd, FROM_ANON, 0, NULL, 
 				    1, &(desc.pmid))) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, 
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, 
 				     &pb)) == PDU_RESULT) {
 		    if ((sts = __pmDecodeResult(pb, &result)) < 0)
 			printf("Error: __pmDecodeResult() failed: %s\n", 
@@ -430,20 +508,31 @@ dopmda(int pdu)
 	    }
 	    else
 		printf("Error: __pmSendFetch() failed: %s\n", pmErrStr(sts));
-
-	    if (sts < 0)
-		return;
-
-	    sts = fillResult(result, desc.type);
+	    /*
+	     * pb is still pinned, and result may contain pointers into
+	     * a second PDU buffer from __pmDecodeResult() ... need to
+	     * ensure all PDU buffers are unpinned once we're done with
+	     * result or giving up
+	     */
 
 	    if (sts < 0) {
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
+		return;
+	    }
+
+	    if ((sts = fillResult(result, desc.type)) < 0) {
 		pmFreeResult(result);
+		__pmUnpinPDUBuf(pb);
 		return;
 	    }
 
 	    printf("Sending Result...\n");
-	    if ((sts = __pmSendResult(outfd, FROM_ANON, result)) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, 
+	    sts = __pmSendResult(outfd, FROM_ANON, result);
+	    pmFreeResult(result);	
+	    __pmUnpinPDUBuf(pb);
+	    if (sts >= 0) {
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, 
 				     &pb)) == PDU_ERROR) {
 		    if ((i = __pmDecodeError(pb, &sts)) >= 0) {
 			if (sts < 0)
@@ -454,11 +543,13 @@ dopmda(int pdu)
 		}
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else
 		printf("Error: __pmSendResult() failed: %s\n", pmErrStr(sts));
 
-	    pmFreeResult(result);	
 	    break;
 
 	case PDU_TEXT_REQ:
@@ -481,7 +572,7 @@ dopmda(int pdu)
 		}
 
 		if ((sts = __pmSendTextReq(outfd, FROM_ANON, ident, param.number)) >= 0) {
-		    if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_TEXT) {
+		    if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_TEXT) {
 			if ((sts = __pmDecodeText(pb, &i, &buffer)) >= 0) {
 			    if (j == 0) {
 				if (*buffer != '\0')
@@ -506,6 +597,9 @@ dopmda(int pdu)
 		    }
 		    else
 			printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+		    if (pinpdu > 0)
+			__pmUnpinPDUBuf(pb);
+
 		}
 		else
 		    printf("Error: __pmSendTextReq() failed: %s\n", pmErrStr(sts));
@@ -516,7 +610,7 @@ dopmda(int pdu)
 	case PDU_PMNS_IDS:
             printf("PMID: %s\n", pmIDStr(param.pmid));
 	    if ((sts = __pmSendIDList(outfd, FROM_ANON, 1, &param.pmid, 0)) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
 		    if ((sts = __pmDecodeNameList(pb, &numnames, &namelist, NULL)) >= 0) {
 			for (i = 0; i < sts; i++) {
 			    printf("   %s\n", namelist[i]);
@@ -534,6 +628,9 @@ dopmda(int pdu)
 		}
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else
 		printf("Error: __pmSendIDList() failed: %s\n", pmErrStr(sts));
@@ -542,11 +639,11 @@ dopmda(int pdu)
 	case PDU_PMNS_NAMES:
             printf("Metric: %s\n", param.name);
 	    if ((sts = __pmSendNameList(outfd, FROM_ANON, 1, &param.name, NULL)) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_IDS) {
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_IDS) {
 		    int		xsts;
-		    if ((sts = __pmDecodeIDList(pb, 1, &pmid, &xsts)) >= 0) {
+
+		    if ((sts = __pmDecodeIDList(pb, 1, &pmid, &xsts)) >= 0)
 			printf("   %s\n", pmIDStr(pmid));
-		    }
 		    else
 			printf("Error: __pmDecodeIDList() failed: %s\n", pmErrStr(sts));
 		}
@@ -556,8 +653,11 @@ dopmda(int pdu)
 		    else
 			printf("Error: __pmDecodeError() failed: %s\n", pmErrStr(i));
 		}
-		else
+		else 
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else
 		printf("Error: __pmSendIDList() failed: %s\n", pmErrStr(sts));
@@ -566,7 +666,7 @@ dopmda(int pdu)
 	case PDU_PMNS_CHILD:
             printf("Metric: %s\n", param.name);
 	    if ((sts = __pmSendChildReq(outfd, FROM_ANON, param.name, 1)) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
 		    if ((sts = __pmDecodeNameList(pb, &numnames, &namelist, &statuslist)) >= 0) {
 			for (i = 0; i < numnames; i++) {
 			    printf("   %8.8s %s\n", statuslist[i] == 1 ? "non-leaf" : "leaf", namelist[i]);
@@ -585,6 +685,9 @@ dopmda(int pdu)
 		}
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else
 		printf("Error: __pmSendChildReq() failed: %s\n", pmErrStr(sts));
@@ -593,8 +696,8 @@ dopmda(int pdu)
 	case PDU_PMNS_TRAVERSE:
             printf("Metric: %s\n", param.name);
 	    if ((sts = __pmSendTraversePMNSReq(outfd, FROM_ANON, param.name)) >= 0) {
-		if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
-		    if ((sts = __pmDecodeNameList(pb, &numnames, &namelist, &statuslist)) >= 0) {
+		if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_NEVER, &pb)) == PDU_PMNS_NAMES) {
+		    if ((sts = __pmDecodeNameList(pb, &numnames, &namelist, NULL)) >= 0) {
 			for (i = 0; i < numnames; i++) {
 			    printf("   %s\n", namelist[i]);
 			}
@@ -611,16 +714,35 @@ dopmda(int pdu)
 		}
 		else
 		    printf("Error: __pmGetPDU() failed: %s\n", pmErrStr(sts));
+
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else
 		printf("Error: __pmSendTraversePMNS() failed: %s\n", pmErrStr(sts));
 	    break;
 
-	default:
-	    printf("Error: Daemon PDU (%s) botch!\n", __pmPDUTypeStr(pdu));
+	case PDU_AUTH:
+	    j = param.number;			/* attribute key */
+	    buffer = param.name;		/* attribute value */
+	    length = !buffer ? 0 : strlen(buffer) + 1;	/* value length */
+	    i = 0;				/* client ID */
+
+	    __pmAttrKeyStr_r(j, name, sizeof(name)-1);
+	    name[sizeof(name)-1] = '\0';
+
+	    printf("Attribute: %s=%s\n", name, buffer ? buffer : "''");
+	    if ((sts = __pmSendAuth(outfd, (int)getpid(), j, buffer, length)) >= 0)
+		printf("Success\n");
+	    else
+		printf("Error: __pmSendAuth() failed: %s\n", pmErrStr(sts));
 	    break;
 
-	}
+	default:
+	    printf("Error: Daemon PDU (%s) botch!\n", __pmPDUTypeStr(pdu));
+	    sts = PDU_ERROR;
+	    break;
+    }
 
     if (sts >= 0 && timer != 0) {
 	__pmtimevalNow(&end);
@@ -653,7 +775,7 @@ fillResult(pmResult *result, int type)
     case PM_TYPE_FLOAT:
 	atom.d = strtod(param.name, &endbuf);
 	if (atom.d < FLT_MIN || atom.d > FLT_MAX)
-	    sts = ERANGE;
+	    sts = -ERANGE;
 	else {
 	    atom.f = atom.d;
 	}
@@ -662,9 +784,9 @@ fillResult(pmResult *result, int type)
 	atom.d = strtod(param.name, &endbuf);
 	break;
     case PM_TYPE_STRING:
-	atom.cp = (char *)malloc(strlen(param.name));
+	atom.cp = (char *)malloc(strlen(param.name) + 1);
 	if (atom.cp == NULL)
-	    sts = ENOMEM;
+	    sts = -ENOMEM;
 	else {
 	    strcpy(atom.cp, param.name);
 	    endbuf = "";
@@ -674,12 +796,12 @@ fillResult(pmResult *result, int type)
 	printf("Error: dbpmda does not support storing into %s metrics\n", pmTypeStr(type));
 	sts = PM_ERR_TYPE;
     }
-    
+
     if (sts < 0) {
 	if (sts != PM_ERR_TYPE)
 	    printf("Error: Decoding value: %s\n", pmErrStr(sts));
     }
-    else if (*endbuf != '\0') {
+    else if (endbuf != NULL && *endbuf != '\0') {
 	printf("Error: Value \"%s\" is incompatible with metric type (PM_TYPE_%s)\n",
 	       param.name, pmTypeStr(type));
 	sts = PM_ERR_VALUE;
