@@ -31,27 +31,83 @@
  * 	call pmNewContext().
  * 4.	There is no pmUnregisterDerived(), so once registered a derived
  * 	metric persists for the life of the application.
+ *
+ * Thread-safe notes
+ *
+ * Need to call PM_INIT_LOCKS() in pmRegisterDerived() because we may
+ * be called before a context has been created, and missed the
+ * lock initialization in pmNewContext().
+ *
+ * registered.mutex is held throughout pmRegisterDerived() and this
+ * protects all of the lexical scanner and parser state, i.e. tokbuf,
+ * tokbuflen, string, lexpeek and this.  Same applies to pmid within
+ * pmRegisterDerived().
+ *
+ * The return value from pmRegisterDerived is either a NULL or a pointer
+ * back into the expr argument, so use of "this" to carry the return
+ * value in the error case is OK.
+ *
+ * All access to registered is controlled by the registered.mutex.
+ *
+ * No locking needed in init() to protect need_init and the getenv()
+ * call, as we always lock the registered.mutex before calling init().
+ *
+ * The context locking protocol ensures that when any of the routines
+ * below are called with a __pmContext * argument, that argument is
+ * not NULL and is associated with a context that is ALREADY locked
+ * via ctxp->c_lock.  We should not unlock the context, that is the
+ * responsibility of our callers.
+ *
+ * derive_errmsg needs to be thread-private
  */
 
 #include <inttypes.h>
-#include "derive.h"
+#include <assert.h>
+#include <ctype.h>
+#include "pmapi.h"
+#include "impl.h"
+#include "internal.h"
+#include "fault.h"
+#ifdef IS_MINGW
+extern const char *strerror_r(int, char *, size_t);
+#endif
 
-static int	need_init = 1;
-static ctl_t	registered;
+static int		need_init = 1;
+static ctl_t		registered = {
+#ifdef PM_MULTI_THREAD
+#ifdef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP,
+#else
+    PTHREAD_MUTEX_INITIALIZER,
+#endif
+#endif
+	0, NULL, 0, 0 };
 
 /* parser and lexer variables */
-static char	*tokbuf = NULL;
-static int	tokbuflen;
-static char	*this;		/* start of current lexicon */
-static int	lexpeek = 0;
-static char	*string;
-static char	*errmsg;
+static char		*tokbuf = NULL;
+static int		tokbuflen;
+static const char	*this;		/* start of current lexicon */
+static int		lexpeek = 0;
+static const char	*string;
 
-static char *type_dbg[] = { "ERROR", "EOF", "UNDEF", "NUMBER", "NAME", "PLUS", "MINUS", "STAR", "SLASH", "LPAREN", "RPAREN", "AVG", "COUNT", "DELTA", "MAX", "MIN", "SUM", "ANON" };
-static char type_c[] = { '\0', '\0', '\0', '\0', '\0', '+', '-', '*', '/', '(', ')', '\0' };
+#ifdef PM_MULTI_THREAD
+#ifdef HAVE___THREAD
+/* using a gcc construct here to make derive_errmsg thread-private */
+static __thread char	*derive_errmsg;
+#endif
+#else
+static char		*derive_errmsg;
+#endif
+
+static const char	*type_dbg[] = {
+	"ERROR", "EOF", "UNDEF", "NUMBER", "NAME", "PLUS", "MINUS",
+	"STAR", "SLASH", "LPAREN", "RPAREN", "AVG", "COUNT", "DELTA",
+	"MAX", "MIN", "SUM", "ANON" };
+static const char	type_c[] = {
+	'\0', '\0', '\0', '\0', '\0', '+', '-', '*', '/', '(', ')', '\0' };
 
 /* function table for lexer */
-static struct {
+static const struct {
     int		f_type;
     char	*f_name;
 } func[] = {
@@ -74,15 +130,64 @@ static struct {
 #define P_FUNC_END	5
 #define P_END		99
 
-static char *state_dbg[] = { "INIT", "LEAF", "LEAF_PAREN", "BINOP", "FUNC_OP", "FUNC_END" };
+static const char	*state_dbg[] = {
+	"INIT", "LEAF", "LEAF_PAREN", "BINOP", "FUNC_OP", "FUNC_END" };
+
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+static void
+initialize_mutex(void)
+{
+    static pthread_mutex_t	init = PTHREAD_MUTEX_INITIALIZER;
+    static int			done = 0;
+    int				psts;
+    char			errmsg[PM_MAXERRMSGLEN];
+    if ((psts = pthread_mutex_lock(&init)) != 0) {
+	strerror_r(psts, errmsg, sizeof(errmsg));
+	fprintf(stderr, "initializ_mutex: pthread_mutex_lock failed: %s", errmsg);
+	exit(4);
+    }
+    if (!done) {
+	/*
+	 * Unable to initialize at compile time, need to do it here in
+	 * a one trip for all threads run-time initialization.
+	 */
+	pthread_mutexattr_t    attr;
+
+	if ((psts = pthread_mutexattr_init(&attr)) != 0) {
+	    strerror_r(psts, errmsg, sizeof(errmsg));
+	    fprintf(stderr, "initialize_mutex: pthread_mutexattr_init failed: %s", errmsg);
+	    exit(4);
+	}
+	if ((psts = pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE)) != 0) {
+	    strerror_r(psts, errmsg, sizeof(errmsg));
+	    fprintf(stderr, "initialize_mutex: pthread_mutexattr_settype failed: %s", errmsg);
+	    exit(4);
+	}
+	if ((psts = pthread_mutex_init(&registered.mutex, &attr)) != 0) {
+	    strerror_r(psts, errmsg, sizeof(errmsg));
+	    fprintf(stderr, "initialize_mutex: pthread_mutex_init failed: %s", errmsg);
+	    exit(4);
+	}
+	done = 1;
+    }
+    if ((psts = pthread_mutex_unlock(&init)) != 0) {
+	strerror_r(psts, errmsg, sizeof(errmsg));
+	fprintf(stderr, "initialize_mutex: pthread_mutex_unlock failed: %s", errmsg);
+	exit(4);
+    }
+}
+#endif
+#endif
 
 /* Register an anonymous metric */
 int
-__pmRegisterAnon(char *name, int type)
+__pmRegisterAnon(const char *name, int type)
 {
     char	*msg;
     char	buf[21];	/* anon(PM_TYPE_XXXXXX) */
 
+PM_FAULT_CHECK(PM_FAULT_PMAPI);
     switch (type) {
 	case PM_TYPE_32:
 	    snprintf(buf, sizeof(buf), "anon(PM_TYPE_32)");
@@ -106,7 +211,7 @@ __pmRegisterAnon(char *name, int type)
 	    return PM_ERR_TYPE;
     }
     if ((msg = pmRegisterDerived(name, buf)) != NULL) {
-	pmprintf("__pmRegisterAnon(%s, %d): Error: %s\n", name, type, pmDerivedErrStr());
+	pmprintf("__pmRegisterAnon(%s, %d): @ \"%s\" Error: %s\n", name, type, msg, pmDerivedErrStr());
 	pmflush();
 	return PM_ERR_GENERIC;
     }
@@ -116,24 +221,26 @@ __pmRegisterAnon(char *name, int type)
 static void
 init(void)
 {
-    char	*configpath;
+    if (need_init) {
+	char	*configpath;
 
-    if ((configpath = getenv("PCP_DERIVED_CONFIG")) != NULL) {
-	int	sts;
+	if ((configpath = getenv("PCP_DERIVED_CONFIG")) != NULL) {
+	    int	sts;
 #ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_DERIVE) {
-	    fprintf(stderr, "Derived metric initialization from $PCP_DERIVED_CONFIG\n");
-	}
+	    if (pmDebug & DBG_TRACE_DERIVE) {
+		fprintf(stderr, "Derived metric initialization from $PCP_DERIVED_CONFIG\n");
+	    }
 #endif
-	sts = pmLoadDerivedConfig(configpath);
+	    sts = pmLoadDerivedConfig(configpath);
 #ifdef PCP_DEBUG
-	if (sts < 0 && (pmDebug & DBG_TRACE_DERIVE)) {
-	    fprintf(stderr, "pmLoadDerivedConfig -> %s\n", pmErrStr(sts));
-	}
+	    if (sts < 0 && (pmDebug & DBG_TRACE_DERIVE)) {
+		char	errmsg[PM_MAXERRMSGLEN];
+		fprintf(stderr, "pmLoadDerivedConfig -> %s\n", pmErrStr_r(sts, errmsg, sizeof(errmsg)));
+	    }
 #endif
+	}
+	need_init = 0;
     }
-
-    need_init = 0;
 }
 
 static void
@@ -145,18 +252,15 @@ unget(int c)
 static int
 get()
 {
-    static int	eof = 0;
     int		c;
     if (lexpeek != 0) {
 	c = lexpeek;
 	lexpeek = 0;
 	return c;
     }
-    if (eof) return L_EOF;
     c = *string;
     if (c == '\0') {
 	return L_EOF;
-	eof = 1;
     }
     string++;
     return c;
@@ -174,7 +278,7 @@ lex(void)
     for ( ; ; ) {
 	c = get();
 	if (firstch) {
-	    if (isspace(c)) continue;
+	    if (isspace((int)c)) continue;
 	    this = &string[-1];
 	    firstch = 0;
 	}
@@ -191,6 +295,7 @@ lex(void)
 	if (p == NULL) {
 	    tokbuflen = 128;
 	    if ((p = tokbuf = (char *)malloc(tokbuflen)) == NULL) {
+		PM_UNLOCK(registered.mutex);
 		__pmNoMem("pmRegisterDerived: alloc tokbuf", tokbuflen, PM_FATAL_ERR);
 		/*NOTREACHED*/
 	    }
@@ -199,6 +304,7 @@ lex(void)
 	    int		x = p - tokbuf;
 	    tokbuflen *= 2;
 	    if ((tokbuf = (char *)realloc(tokbuf, tokbuflen)) == NULL) {
+		PM_UNLOCK(registered.mutex);
 		__pmNoMem("pmRegisterDerived: realloc tokbuf", tokbuflen, PM_FATAL_ERR);
 		/*NOTREACHED*/
 	    }
@@ -208,9 +314,9 @@ lex(void)
 	*p++ = (char)c;
 
 	if (ltype == L_UNDEF) {
-	    if (isdigit(c))
+	    if (isdigit((int)c))
 		ltype = L_NUMBER;
-	    else if (isalpha(c))
+	    else if (isalpha((int)c))
 		ltype = L_NAME;
 	    else {
 		switch (c) {
@@ -252,14 +358,14 @@ lex(void)
 	}
 	else {
 	    if (ltype == L_NUMBER) {
-		if (!isdigit(c)) {
+		if (!isdigit((int)c)) {
 		    unget(c);
 		    p[-1] = '\0';
 		    return L_NUMBER;
 		}
 	    }
 	    else if (ltype == L_NAME) {
-		if (isalpha(c) || isdigit(c) || c == '_' || c == '.')
+		if (isalpha((int)c) || isdigit((int)c) || c == '_' || c == '.')
 		    continue;
 		if (c == '(') {
 		    /* check for functions ... */
@@ -288,6 +394,7 @@ newnode(int type)
     node_t	*np;
     np = (node_t *)malloc(sizeof(node_t));
     if (np == NULL) {
+	PM_UNLOCK(registered.mutex);
 	__pmNoMem("pmRegisterDerived: newnode", sizeof(node_t), PM_FATAL_ERR);
 	/*NOTREACHED*/
     }
@@ -338,6 +445,7 @@ bind_expr(int n, node_t *np)
 	}
     }
     if ((new->info = (info_t *)malloc(sizeof(info_t))) == NULL) {
+	PM_UNLOCK(registered.mutex);
 	__pmNoMem("bind_expr: info block", sizeof(info_t), PM_FATAL_ERR);
 	/*NOTREACHED*/
     }
@@ -360,7 +468,8 @@ bind_expr(int n, node_t *np)
 	if (sts < 0) {
 #ifdef PCP_DEBUG
 	    if (pmDebug & DBG_TRACE_DERIVE) {
-		fprintf(stderr, "bind_expr: error: derived metric %s: operand: %s: %s\n", registered.mlist[n].name, new->value, pmErrStr(sts));
+		char	errmsg[PM_MAXERRMSGLEN];
+		fprintf(stderr, "bind_expr: error: derived metric %s: operand: %s: %s\n", registered.mlist[n].name, new->value, pmErrStr_r(sts, errmsg, sizeof(errmsg)));
 	    }
 #endif
 	    free(new->info);
@@ -371,7 +480,9 @@ bind_expr(int n, node_t *np)
 	if (sts < 0) {
 #ifdef PCP_DEBUG
 	    if (pmDebug & DBG_TRACE_DERIVE) {
-		fprintf(stderr, "bind_expr: error: derived metric %s: operand (%s [%s]): %s\n", registered.mlist[n].name, new->value, pmIDStr(new->info->pmid), pmErrStr(sts));
+		char	strbuf[20];
+		char	errmsg[PM_MAXERRMSGLEN];
+		fprintf(stderr, "bind_expr: error: derived metric %s: operand (%s [%s]): %s\n", registered.mlist[n].name, new->value, pmIDStr_r(new->info->pmid, strbuf, sizeof(strbuf)), pmErrStr_r(sts, errmsg, sizeof(errmsg)));
 	    }
 #endif
 	    free(new->info);
@@ -422,13 +533,13 @@ void report_sem_error(char *name, node_t *np)
 		pmprintf("botch @ node type #%d?", np->type);
 	    break;
     }
-    pmprintf(": %s\n", errmsg);
+    pmprintf(": %s\n", PM_TPD(derive_errmsg));
     pmflush();
-    errmsg = NULL;
+    PM_TPD(derive_errmsg) = NULL;
 }
 
 /* type promotion */
-static int promote[6][6] = {
+static const int promote[6][6] = {
     { PM_TYPE_32, PM_TYPE_U32, PM_TYPE_64, PM_TYPE_U64, PM_TYPE_FLOAT, PM_TYPE_DOUBLE },
     { PM_TYPE_U32, PM_TYPE_U32, PM_TYPE_64, PM_TYPE_U64, PM_TYPE_FLOAT, PM_TYPE_DOUBLE },
     { PM_TYPE_64, PM_TYPE_64, PM_TYPE_64, PM_TYPE_U64, PM_TYPE_FLOAT, PM_TYPE_DOUBLE },
@@ -438,7 +549,7 @@ static int promote[6][6] = {
 };
 
 /* time scale conversion factors */
-static int timefactor[] = {
+static const int timefactor[] = {
     1000,		/* NSEC -> USEC */
     1000,		/* USEC -> MSEC */
     1000,		/* MSEC -> SEC */
@@ -605,13 +716,13 @@ map_desc(int n, node_t *np)
     if (left->sem == PM_SEM_COUNTER) {
 	if (right->sem == PM_SEM_COUNTER) {
 	    if (np->type != L_PLUS && np->type != L_MINUS) {
-		errmsg = "Illegal operator for counters";
+		PM_TPD(derive_errmsg) = "Illegal operator for counters";
 		goto bad;
 	    }
 	}
 	else {
 	    if (np->type != L_STAR && np->type != L_SLASH) {
-		errmsg = "Illegal operator for counter and non-counter";
+		PM_TPD(derive_errmsg) = "Illegal operator for counter and non-counter";
 		goto bad;
 	    }
 	}
@@ -619,7 +730,7 @@ map_desc(int n, node_t *np)
     else {
 	if (right->sem == PM_SEM_COUNTER) {
 	    if (np->type != L_STAR) {
-		errmsg = "Illegal operator for non-counter and counter";
+		PM_TPD(derive_errmsg) = "Illegal operator for non-counter and counter";
 		goto bad;
 	    }
 	}
@@ -631,7 +742,7 @@ map_desc(int n, node_t *np)
 		 * arithmetic operators are supported and all are
 		 * acceptable here ... check added for completeness
 		 */
-		errmsg = "Illegal operator for non-counters";
+		PM_TPD(derive_errmsg) = "Illegal operator for non-counters";
 		goto bad;
 	    }
 	}
@@ -668,7 +779,7 @@ map_desc(int n, node_t *np)
 	case PM_TYPE_DOUBLE:
 	    break;
 	default:
-	    errmsg = "Non-arithmetic type for left operand";
+	    PM_TPD(derive_errmsg) = "Non-arithmetic type for left operand";
 	    goto bad;
     }
     switch (right->type) {
@@ -680,7 +791,7 @@ map_desc(int n, node_t *np)
 	case PM_TYPE_DOUBLE:
 	    break;
 	default:
-	    errmsg = "Non-arithmetic type for right operand";
+	    PM_TPD(derive_errmsg) = "Non-arithmetic type for right operand";
 	    goto bad;
     }
     np->desc.type = promote[left->type][right->type];
@@ -696,7 +807,7 @@ map_desc(int n, node_t *np)
 	if (left->units.dimCount != right->units.dimCount ||
 	    left->units.dimTime != right->units.dimTime ||
 	    left->units.dimSpace != right->units.dimSpace) {
-	    errmsg = "Dimensions are not the same";
+	    PM_TPD(derive_errmsg) = "Dimensions are not the same";
 	    goto bad;
 	}
 	map_units(np);
@@ -711,7 +822,7 @@ map_desc(int n, node_t *np)
 	    if (right->units.dimCount != 0 ||
 	        right->units.dimTime != 0 ||
 	        right->units.dimSpace != 0) {
-		errmsg = "Non-counter and not dimensionless for right operand";
+		PM_TPD(derive_errmsg) = "Non-counter and not dimensionless for right operand";
 		goto bad;
 	    }
 	}
@@ -719,7 +830,7 @@ map_desc(int n, node_t *np)
 	    if (left->units.dimCount != 0 ||
 	        left->units.dimTime != 0 ||
 	        left->units.dimSpace != 0) {
-		errmsg = "Non-counter and not dimensionless for left operand";
+		PM_TPD(derive_errmsg) = "Non-counter and not dimensionless for left operand";
 		goto bad;
 	    }
 	}
@@ -731,7 +842,7 @@ map_desc(int n, node_t *np)
      * instance domain
      */
     if (left->indom != PM_INDOM_NULL && right->indom != PM_INDOM_NULL && left->indom != right->indom) {
-	errmsg = "Operands should have the same instance domain";
+	PM_TPD(derive_errmsg) = "Operands should have the same instance domain";
 	goto bad;
     }
 
@@ -748,18 +859,23 @@ check_expr(int n, node_t *np)
     int		sts;
 
     assert(np != NULL);
+
     if (np->type == L_NUMBER || np->type == L_NAME)
 	return 0;
-    if (np->left != NULL)
-	if ((sts = check_expr(n, np->left)) < 0)
-	    return sts;
-    if (np->right != NULL)
+
+    /* otherwise, np->left is never NULL ... */
+    assert(np->left != NULL);
+
+    if ((sts = check_expr(n, np->left)) < 0)
+	return sts;
+    if (np->right != NULL) {
 	if ((sts = check_expr(n, np->right)) < 0)
 	    return sts;
-    /*
-     * np->left is never NULL ...
-     */
-    if (np->right == NULL) {
+	/* build pmDesc from pmDesc of both operands */
+	if ((sts = map_desc(n, np)) < 0)
+	    return sts;
+    }
+    else {
 	np->desc = np->left->desc;	/* struct copy */
 	/*
 	 * special cases for functions ...
@@ -787,7 +903,7 @@ check_expr(int n, node_t *np)
 		    case PM_TYPE_DOUBLE:
 			break;
 		    default:
-			errmsg = "Non-arithmetic operand for function";
+			PM_TPD(derive_errmsg) = "Non-arithmetic operand for function";
 			report_sem_error(registered.mlist[n].name, np);
 			return -1;
 		}
@@ -806,12 +922,6 @@ check_expr(int n, node_t *np)
 	    ;
 	}
     }
-    else {
-	/* build pmDesc from pmDesc of both operands */
-	if ((sts = map_desc(n, np)) < 0) {
-	    return sts;
-	}
-    }
     return 0;
 }
 
@@ -866,14 +976,17 @@ dump_value(int type, pmAtomValue *avp)
 void
 __dmdumpexpr(node_t *np, int level)
 {
-    if (level == 0) fprintf(stderr, "Derived metric expr dump from %p...\n", np);
+    char	strbuf[20];
+
+    if (level == 0) fprintf(stderr, "Derived metric expr dump from " PRINTF_P_PFX "%p...\n", np);
     if (np == NULL) return;
-    fprintf(stderr, "expr node %p type=%s left=%p right=%p save_last=%d", np, type_dbg[np->type+2], np->left, np->right, np->save_last);
+    fprintf(stderr, "expr node " PRINTF_P_PFX "%p type=%s left=" PRINTF_P_PFX "%p right=" PRINTF_P_PFX "%p save_last=%d", np, type_dbg[np->type+2], np->left, np->right, np->save_last);
     if (np->type == L_NAME || np->type == L_NUMBER)
 	fprintf(stderr, " [%s] master=%d", np->value, np->info == NULL ? 1 : 0);
     fputc('\n', stderr);
     if (np->info) {
-	fprintf(stderr, "    PMID: %s (%s from pmDesc) numval: %d", pmIDStr(np->info->pmid), pmIDStr(np->desc.pmid), np->info->numval);
+	fprintf(stderr, "    PMID: %s ", pmIDStr_r(np->info->pmid, strbuf, sizeof(strbuf)));
+	fprintf(stderr, "(%s from pmDesc) numval: %d", pmIDStr_r(np->desc.pmid, strbuf, sizeof(strbuf)), np->info->numval);
 	if (np->info->div_scale != 1)
 	    fprintf(stderr, " div_scale: %d", np->info->div_scale);
 	if (np->info->mul_scale != 1)
@@ -943,14 +1056,14 @@ parse(int level)
 	/* handle lexicons that terminate the parsing */
 	switch (type) {
 	    case L_ERROR:
-		errmsg = "Illegal character";
+		PM_TPD(derive_errmsg) = "Illegal character";
 		free_expr(expr);
 		return NULL;
 		break;
 	    case L_EOF:
 		if (level == 1 && (state == P_LEAF || state == P_LEAF_PAREN))
 		    return expr;
-		errmsg = "End of input";
+		PM_TPD(derive_errmsg) = "End of input";
 		free_expr(expr);
 		return NULL;
 		break;
@@ -961,7 +1074,7 @@ parse(int level)
 		}
 		if ((level > 1 && state == P_LEAF_PAREN) || state == P_LEAF)
 		    return expr;
-		errmsg = "Unexpected ')'";
+		PM_TPD(derive_errmsg) = "Unexpected ')'";
 		free_expr(expr);
 		return NULL;
 		break;
@@ -969,9 +1082,20 @@ parse(int level)
 
 	switch (state) {
 	    case P_INIT:
+		/*
+		 * Only come here at the start of parsing an expression.
+		 * The assert() is designed to stop Coverity flagging a
+		 * memory leak if we should come here after expr and/or
+		 * curr have already been assigned values either directly
+		 * from calling newnode() or via an assignment to np that
+		 * was previously assigned a value from newnode()
+		 */
+		assert(expr == NULL && curr == NULL);
+
 		if (type == L_NAME || type == L_NUMBER) {
 		    expr = curr = newnode(type);
 		    if ((curr->value = strdup(tokbuf)) == NULL) {
+			PM_UNLOCK(registered.mutex);
 			__pmNoMem("pmRegisterDerived: leaf node", strlen(tokbuf)+1, PM_FATAL_ERR);
 			/*NOTREACHED*/
 		    }
@@ -980,7 +1104,7 @@ parse(int level)
 			__uint64_t	check;
 			check = strtoull(tokbuf, &endptr, 10);
 			if (*endptr != '\0' || check > 0xffffffffUL) {
-			    errmsg = "Constant value too large";
+			    PM_TPD(derive_errmsg) = "Constant value too large";
 			    free_expr(expr);
 			    return NULL;
 			}
@@ -1004,8 +1128,10 @@ parse(int level)
 		    expr = curr = newnode(type);
 		    state = P_FUNC_OP;
 		}
-		else
+		else {
+		    free_expr(expr);
 		    return NULL;
+		}
 		break;
 
 	    case P_LEAF_PAREN:	/* fall through */
@@ -1047,6 +1173,7 @@ parse(int level)
 		if (type == L_NAME || type == L_NUMBER) {
 		    np = newnode(type);
 		    if ((np->value = strdup(tokbuf)) == NULL) {
+			PM_UNLOCK(registered.mutex);
 			__pmNoMem("pmRegisterDerived: leaf node", strlen(tokbuf)+1, PM_FATAL_ERR);
 			/*NOTREACHED*/
 		    }
@@ -1086,6 +1213,7 @@ parse(int level)
 		if (type == L_NAME) {
 		    np = newnode(type);
 		    if ((np->value = strdup(tokbuf)) == NULL) {
+			PM_UNLOCK(registered.mutex);
 			__pmNoMem("pmRegisterDerived: func op node", strlen(tokbuf)+1, PM_FATAL_ERR);
 			/*NOTREACHED*/
 		    }
@@ -1147,11 +1275,11 @@ checkname(char *p)
     for ( ; *p; p++) {
 	if (firstch) {
 	    firstch = 0;
-	    if (isalpha(*p)) continue;
+	    if (isalpha((int)*p)) continue;
 	    return -1;
 	}
 	else {
-	    if (isalpha(*p) || isdigit(*p) || *p == '_') continue;
+	    if (isalpha((int)*p) || isdigit((int)*p) || *p == '_') continue;
 	    if (*p == '.') {
 		firstch = 1;
 		continue;
@@ -1163,12 +1291,20 @@ checkname(char *p)
 }
 
 char *
-pmRegisterDerived(char *name, char *expr)
+pmRegisterDerived(const char *name, const char *expr)
 {
     node_t		*np;
     static __pmID_int	pmid;
     int			i;
 
+    PM_INIT_LOCKS();
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    initialize_mutex();
+#endif
+#endif
+    PM_LOCK(registered.mutex);
+
 #ifdef PCP_DEBUG
     if ((pmDebug & DBG_TRACE_DERIVE) && (pmDebug & DBG_TRACE_APPL0)) {
 	fprintf(stderr, "pmRegisterDerived: name=\"%s\" expr=\"%s\"\n", name, expr);
@@ -1178,22 +1314,26 @@ pmRegisterDerived(char *name, char *expr)
     for (i = 0; i < registered.nmetric; i++) {
 	if (strcmp(name, registered.mlist[i].name) == 0) {
 	    /* oops, duplicate name ... */
-	    errmsg = "Duplicate derived metric name";
-	    return expr;
+	    PM_TPD(derive_errmsg) = "Duplicate derived metric name";
+	    PM_UNLOCK(registered.mutex);
+	    return (char *)expr;
 	}
     }
 
-    errmsg = NULL;
+    PM_TPD(derive_errmsg) = NULL;
     string = expr;
     np = parse(1);
     if (np == NULL) {
 	/* parser error */
-	return this;
+	char	*sts = (char *)this;
+	PM_UNLOCK(registered.mutex);
+	return sts;
     }
 
     registered.nmetric++;
     registered.mlist = (dm_t *)realloc(registered.mlist, registered.nmetric*sizeof(dm_t));
     if (registered.mlist == NULL) {
+	PM_UNLOCK(registered.mutex);
 	__pmNoMem("pmRegisterDerived: registered mlist", registered.nmetric*sizeof(dm_t), PM_FATAL_ERR);
 	/*NOTREACHED*/
     }
@@ -1215,11 +1355,12 @@ pmRegisterDerived(char *name, char *expr)
     }
 #endif
 
+    PM_UNLOCK(registered.mutex);
     return NULL;
 }
 
 int
-pmLoadDerivedConfig(char *fname)
+pmLoadDerivedConfig(const char *fname)
 {
     FILE	*fp;
     int		buflen;
@@ -1241,6 +1382,7 @@ pmLoadDerivedConfig(char *fname)
     }
     buflen = 128;
     if ((buf = (char *)malloc(buflen)) == NULL) {
+	/* registered.mutex not locked in this case */
 	__pmNoMem("pmLoadDerivedConfig: alloc buf", buflen, PM_FATAL_ERR);
 	/*NOTREACHED*/
     }
@@ -1248,6 +1390,7 @@ pmLoadDerivedConfig(char *fname)
     while ((c = fgetc(fp)) != EOF) {
 	if (p == &buf[buflen]) {
 	    if ((buf = (char *)realloc(buf, 2*buflen)) == NULL) {
+		/* registered.mutex not locked in this case */
 		__pmNoMem("pmLoadDerivedConfig: expand buf", 2*buflen, PM_FATAL_ERR);
 		/*NOTREACHED*/
 	    }
@@ -1274,16 +1417,17 @@ pmLoadDerivedConfig(char *fname)
 		char	*errp;
 		buf[eq] = '\0';
 		if ((np = strdup(buf)) == NULL) {
+		    /* registered.mutex not locked in this case */
 		    __pmNoMem("pmLoadDerivedConfig: dupname", strlen(buf), PM_FATAL_ERR);
 		    /*NOTREACHED*/
 		}
 		/* trim white space from tail of metric name */
 		q = &np[eq-1];
-		while (q >= np && isspace(*q))
+		while (q >= np && isspace((int)*q))
 		    *q-- = '\0';
 		/* trim white space from head of metric name */
 		q = np;
-		while (*q && isspace(*q))
+		while (*q && isspace((int)*q))
 		    q++;
 		if (*q == '\0') {
 		    buf[eq] = '=';
@@ -1299,7 +1443,7 @@ pmLoadDerivedConfig(char *fname)
 		    goto next_line;
 		}
 		ep = &buf[eq+1];
-		while (*ep != '\0' && isspace(*ep))
+		while (*ep != '\0' && isspace((int)*ep))
 		    ep++;
 		if (*ep == '\0') {
 		    buf[eq] = '=';
@@ -1314,7 +1458,7 @@ pmLoadDerivedConfig(char *fname)
 		    pmprintf("%s\n", &buf[eq+1]);
 		    for (q = &buf[eq+1]; *q; q++) {
 			if (q == errp) *q = '^';
-			else if (!isspace(*q)) *q = ' ';
+			else if (!isspace((int)*q)) *q = ' ';
 		    }
 		    pmprintf("%s\n", &buf[eq+1]);
 		    q = pmDerivedErrStr();
@@ -1340,6 +1484,7 @@ next_line:
 	else
 	    *p++ = c;
     }
+    fclose(fp);
     free(buf);
     return sts;
 }
@@ -1347,7 +1492,8 @@ next_line:
 char *
 pmDerivedErrStr(void)
 {
-    return errmsg;
+    PM_INIT_LOCKS();
+    return PM_TPD(derive_errmsg);
 }
 
 /*
@@ -1362,7 +1508,13 @@ __dmtraverse(const char *name, char ***namelist)
     char	**list = NULL;
     int		matchlen = strlen(name);
 
-    if (need_init) init();
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    initialize_mutex();
+#endif
+#endif
+    PM_LOCK(registered.mutex);
+    init();
 
     for (i = 0; i < registered.nmetric; i++) {
 	/*
@@ -1374,6 +1526,7 @@ __dmtraverse(const char *name, char ***namelist)
 	      registered.mlist[i].name[matchlen] == '\0'))) {
 	    sts++;
 	    if ((list = (char **)realloc(list, sts*sizeof(list[0]))) == NULL) {
+		PM_UNLOCK(registered.mutex);
 		__pmNoMem("__dmtraverse: list", sts*sizeof(list[0]), PM_FATAL_ERR);
 		/*NOTREACHED*/
 	    }
@@ -1382,6 +1535,7 @@ __dmtraverse(const char *name, char ***namelist)
     }
     *namelist = list;
 
+    PM_UNLOCK(registered.mutex);
     return sts;
 }
 
@@ -1397,7 +1551,13 @@ __dmchildren(const char *name, char ***offspring, int **statuslist)
     int		start;
     int		len;
 
-    if (need_init) init();
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    initialize_mutex();
+#endif
+#endif
+    PM_LOCK(registered.mutex);
+    init();
 
     for (i = 0; i < registered.nmetric; i++) {
 	/*
@@ -1408,7 +1568,15 @@ __dmchildren(const char *name, char ***offspring, int **statuslist)
 	     (registered.mlist[i].name[matchlen] == '.' ||
 	      registered.mlist[i].name[matchlen] == '\0'))) {
 	    if (registered.mlist[i].name[matchlen] == '\0') {
-		/* leaf node */
+		/*
+		 * leaf node
+		 * assert is for coverity, name uniqueness means we
+		 * should only ever come here after zero passes through
+		 * the block below where sts is incremented and children[]
+		 * and status[] are realloc'd
+		 */
+		assert(sts == 0 && children == NULL && status == NULL);
+		PM_UNLOCK(registered.mutex);
 		return 0;
 	    }
 	    start = matchlen > 0 ? matchlen + 1 : 0;
@@ -1422,12 +1590,14 @@ __dmchildren(const char *name, char ***offspring, int **statuslist)
 		/* first time for this one */
 		sts++;
 		if ((children = (char **)realloc(children, sts*sizeof(children[0]))) == NULL) {
+		    PM_UNLOCK(registered.mutex);
 		    __pmNoMem("__dmchildren: children", sts*sizeof(children[0]), PM_FATAL_ERR);
 		    /*NOTREACHED*/
 		}
 		for (len = 0; registered.mlist[i].name[start+len] != '\0' && registered.mlist[i].name[start+len] != '.'; len++)
 		    ;
 		if ((children[sts-1] = (char *)malloc(len+1)) == NULL) {
+		    PM_UNLOCK(registered.mutex);
 		    __pmNoMem("__dmchildren: name", len+1, PM_FATAL_ERR);
 		    /*NOTREACHED*/
 		}
@@ -1441,6 +1611,7 @@ __dmchildren(const char *name, char ***offspring, int **statuslist)
 
 		if (statuslist != NULL) {
 		    if ((status = (int *)realloc(status, sts*sizeof(status[0]))) == NULL) {
+			PM_UNLOCK(registered.mutex);
 			__pmNoMem("__dmchildren: statrus", sts*sizeof(status[0]), PM_FATAL_ERR);
 			/*NOTREACHED*/
 		    }
@@ -1460,13 +1631,16 @@ __dmchildren(const char *name, char ***offspring, int **statuslist)
 	}
     }
 
-    if (sts == 0)
+    if (sts == 0) {
+	PM_UNLOCK(registered.mutex);
 	return PM_ERR_NAME;
+    }
 
     *offspring = children;
     if (statuslist != NULL)
 	*statuslist = status;
 
+    PM_UNLOCK(registered.mutex);
     return sts;
 }
 
@@ -1475,14 +1649,22 @@ __dmgetpmid(const char *name, pmID *dp)
 {
     int		i;
 
-    if (need_init) init();
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    initialize_mutex();
+#endif
+#endif
+    PM_LOCK(registered.mutex);
+    init();
 
     for (i = 0; i < registered.nmetric; i++) {
 	if (strcmp(name, registered.mlist[i].name) == 0) {
 	    *dp = registered.mlist[i].pmid;
+	    PM_UNLOCK(registered.mutex);
 	    return 0;
 	}
     }
+    PM_UNLOCK(registered.mutex);
     return PM_ERR_NAME;
 }
 
@@ -1491,17 +1673,28 @@ __dmgetname(pmID pmid, char ** name)
 {
     int		i;
 
-    if (need_init) init();
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    initialize_mutex();
+#endif
+#endif
+    PM_LOCK(registered.mutex);
+    init();
 
     for (i = 0; i < registered.nmetric; i++) {
 	if (pmid == registered.mlist[i].pmid) {
 	    *name = strdup(registered.mlist[i].name);
-	    if (*name == NULL)
+	    if (*name == NULL) {
+		PM_UNLOCK(registered.mutex);
 		return -oserror();
-	    else
+	    }
+	    else {
+		PM_UNLOCK(registered.mutex);
 		return 0;
+	    }
 	}
     }
+    PM_UNLOCK(registered.mutex);
     return PM_ERR_PMID;
 }
 
@@ -1512,7 +1705,13 @@ __dmopencontext(__pmContext *ctxp)
     int		sts;
     ctl_t	*cp;
 
-    if (need_init) init();
+#ifdef PM_MULTI_THREAD
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
+    initialize_mutex();
+#endif
+#endif
+    PM_LOCK(registered.mutex);
+    init();
 
 #ifdef PCP_DEBUG
     if ((pmDebug & DBG_TRACE_DERIVE) && (pmDebug & DBG_TRACE_APPL1)) {
@@ -1521,15 +1720,18 @@ __dmopencontext(__pmContext *ctxp)
 #endif
     if (registered.nmetric == 0) {
 	ctxp->c_dm = NULL;
+	PM_UNLOCK(registered.mutex);
 	return;
     }
     if ((cp = (void *)malloc(sizeof(ctl_t))) == NULL) {
+	PM_UNLOCK(registered.mutex);
 	__pmNoMem("pmNewContext: derived metrics (ctl)", sizeof(ctl_t), PM_FATAL_ERR);
 	/* NOTREACHED */
     }
     ctxp->c_dm = (void *)cp;
     cp->nmetric = registered.nmetric;
     if ((cp->mlist = (dm_t *)malloc(cp->nmetric*sizeof(dm_t))) == NULL) {
+	PM_UNLOCK(registered.mutex);
 	__pmNoMem("pmNewContext: derived metrics (mlist)", cp->nmetric*sizeof(dm_t), PM_FATAL_ERR);
 	/* NOTREACHED */
     }
@@ -1559,6 +1761,7 @@ __dmopencontext(__pmContext *ctxp)
 	}
 #endif
     }
+    PM_UNLOCK(registered.mutex);
 }
 
 void
@@ -1571,7 +1774,7 @@ __dmclosecontext(__pmContext *ctxp)
 
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_DERIVE) {
-	fprintf(stderr, "__dmclosecontext(->ctx %d) called dm->%p %d metrics\n", __pmPtrToHandle(ctxp), cp, cp == NULL ? -1 : cp->nmetric);
+	fprintf(stderr, "__dmclosecontext(->ctx %d) called dm->" PRINTF_P_PFX "%p %d metrics\n", __pmPtrToHandle(ctxp), cp, cp == NULL ? -1 : cp->nmetric);
     }
 #endif
     if (cp == NULL) return;
@@ -1604,3 +1807,17 @@ __dmdesc(__pmContext *ctxp, pmID pmid, pmDesc *desc)
     }
     return PM_ERR_PMID;
 }
+
+#ifdef PM_MULTI_THREAD
+#ifdef PM_MULTI_THREAD_DEBUG
+/*
+ * return true if lock == registered.mutex ... no locking here to avoid
+ * recursion ad nauseum
+ */
+int
+__pmIsDeriveLock(void *lock)
+{
+    return lock == (void *)&registered.mutex;
+}
+#endif
+#endif
