@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013 Red Hat.
  * Copyright (c) 2008-2010 Aconex.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -11,6 +12,21 @@
  * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  * License for more details.
  */
+
+/*
+ * For the MinGW headers and library to work correctly, we need
+ * something newer than the default Windows 95 versions of the Win32
+ * APIs.  0x0500 is the magic sauce to select the Windows 2000 version
+ * of the Win32 APIs, which is the minimal version needed for PCP.
+ *
+ * WINVER needs to be set before any of the MinGW headers are processed
+ * and we include <windows.h> from pmapi.h via platform_defs.h.
+ *
+ * Thanks to "Earnie" on the mingw-users@lists.sourceforge.net mailing
+ * list for this tip.
+ */
+#define WINVER 0x0500
+
 #include "pmapi.h"
 #include "impl.h"
 #include <winbase.h>
@@ -86,7 +102,7 @@ __pmSetSignalHandler(int sig, __pmSignalHandler func)
 	return 0;
 
     sts = 0;
-    snprintf(evname, sizeof(evname), "PCP/%d/%s", getpid(), signame);
+    snprintf(evname, sizeof(evname), "PCP/%" FMT_PID "/%s", getpid(), signame);
     if (!(eventhdl = CreateEvent(NULL, FALSE, FALSE, TEXT(evname)))) {
 	sts = GetLastError();
 	fprintf(stderr, "CreateEvent::%s failed (%d)\n", signame, sts);
@@ -111,6 +127,13 @@ sigterm_callback(int sig)
     exit(0);	/* give atexit(3) handlers a look-in */
 }
 
+int
+__pmSetProcessIdentity(const char *username)
+{
+    (void)username;
+    return 0;	/* Not Yet Implemented */
+}
+
 int
 __pmSetProgname(const char *program)
 {
@@ -524,89 +547,13 @@ closelog(void)
     eventlog = NULL;
 }
 
-
-/* loop.c */
-
-int
-pmLoopRegisterInput(
-    int fd,
-    int flags,
-    int (*callback)(int fd, int flags, void *closure),
-    void *closure,
-    int priority)
+const char *
+strerror_r(int errnum, char *buf, size_t buflen)
 {
-    return -1;
-}
-
-void
-pmLoopUnregisterInput(int tag)
-{
-}
-
-int
-pmLoopRegisterSignal(
-    int sig,
-    int (*callback)(int sig, void *closure),
-    void *closure)
-{
-    return -1;
-}
-
-void
-pmLoopUnregisterSignal(int tag)
-{
-}
-
-int
-pmLoopRegisterTimeout(
-    int tout_msec,
-    int (*callback)(void *closure),
-    void *closure)
-{
-    return -1;
-}
-
-void
-pmLoopUnregisterTimeout(int tag)
-{
-}
-
-int
-pmLoopRegisterChild(
-    pid_t pid,
-    int (*callback)(pid_t pid, int status, const struct rusage *, void *closure),
-    void *closure)
-{
-    return -1;
-}
-
-void
-pmLoopUnregisterChild(int tag)
-{
-}
-
-int
-pmLoopRegisterIdle(
-    int (*callback)(void *closure),
-    void *closure)
-{
-    return -1;
-}
-
-void
-pmLoopUnregisterIdle(int tag)
-{
-}
-
-void
-pmLoopStop(void)
-{
-}
-
-int
-pmLoopMain(void)
-{
-    return -1;
+    /* strerror_s is missing from the MinGW string.h */
+    /* we need to wait for it until we can do this:  */
+    /* return strerror_s(buf, buflen, errnum); */
+    return strerror(errnum);
 }
 
 /* Windows socket error codes - what a nightmare! */
@@ -676,7 +623,7 @@ static const struct {
           { 0,"" }
 };
 
-char *
+const char *
 wsastrerror(int code)
 {
     int i;
@@ -686,3 +633,96 @@ wsastrerror(int code)
 	    return wsatab[i].errmess;
     return NULL;
 }
+
+/*
+ * User and group account management using Security IDs (SIDs)
+ */
+int
+__pmValidUserID(__pmUserID sid)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmValidGroupID(__pmGroupID sid)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmEqualUserIDs(__pmUserID sid1, __pmUserID sid2)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmEqualGroupIDs(__pmGroupID sid1, __pmGroupID sid2)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+void
+__pmUserIDFromString(const char *username, __pmUserID *sid)
+{
+    /* NYI */
+}
+
+void
+__pmGroupIDFromString(const char *groupname, __pmGroupID *sid)
+{
+    /* NYI */
+}
+
+char *
+__pmUserIDToString(__pmUserID sid, char *buffer, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+char *
+__pmGroupIDToString(__pmGroupID gid, char *buffer, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+int
+__pmUsernameToID(const char *username, __pmUserID *uidp)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmGroupnameToID(const char *groupname, __pmGroupID *gidp)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+char *
+__pmGroupnameFromID(__pmGroupID gid, char *buf, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+char *
+__pmUsernameFromID(__pmUserID uid, char *buf, size_t size)
+{
+    return NULL;	/* NYI */
+}
+
+int
+__pmUsersGroupIDs(const char *username, __pmGroupID **groupids, unsigned int *ngroups)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmGroupsUserIDs(const char *groupname, __pmUserID **userids, unsigned int *nusers)
+{
+    return -ENOTSUP;	/* NYI */
+}
+
+int
+__pmGetUserIdentity(const char *username, __pmUserID *uid, __pmGroupID *gid, int mode)
+{
+    return -ENOTSUP;	/* NYI */
+}
