@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013 Red Hat.
  * Copyright (c) 1995-2000 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -39,13 +40,16 @@ __pmdaCntInst(pmInDom indom, pmdaExt *pmda)
 		break;
 	    }
 	}
-	if (i == pmda->e_nindoms)
-	    __pmNotifyErr(LOG_WARNING, "cntinst: unknown indom %s", pmInDomStr(indom));
+	if (i == pmda->e_nindoms) {
+	    char	strbuf[20];
+	    __pmNotifyErr(LOG_WARNING, "cntinst: unknown indom %s", pmInDomStr_r(indom, strbuf, sizeof(strbuf)));
+	}
     }
 
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_INDOM) {
-	fprintf(stderr, "__pmdaCntInst(indom=%s) -> %d\n", pmInDomStr(indom), sts);
+	char	strbuf[20];
+	fprintf(stderr, "__pmdaCntInst(indom=%s) -> %d\n", pmInDomStr_r(indom, strbuf, sizeof(strbuf)), sts);
     }
 #endif
 
@@ -111,8 +115,9 @@ __pmdaStartInst(pmInDom indom, pmdaExt *pmda)
 
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_INDOM) {
+	char	strbuf[20];
 	fprintf(stderr, "__pmdaStartInst(indom=%s) e_ordinal=%d\n",
-	    pmInDomStr(indom), pmda->e_ordinal);
+	    pmInDomStr_r(indom, strbuf, sizeof(strbuf)), pmda->e_ordinal);
     }
 #endif
     return;
@@ -144,8 +149,9 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
 		    *inst = myinst;
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_INDOM) {
+			char	strbuf[20];
 			fprintf(stderr, "__pmdaNextInst(indom=%s) -> %d e_ordinal=%d (cache)\n",
-			    pmInDomStr(pmda->e_idp->it_indom), myinst, pmda->e_ordinal);
+			    pmInDomStr_r(pmda->e_idp->it_indom, strbuf, sizeof(strbuf)), myinst, pmda->e_ordinal);
 		    }
 #endif
 		    return 1;
@@ -162,8 +168,9 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
 		    pmda->e_ordinal = j+1;
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_INDOM) {
+			char	strbuf[20];
 			fprintf(stderr, "__pmdaNextInst(indom=%s) -> %d e_ordinal=%d\n",
-			    pmInDomStr(pmda->e_idp->it_indom), *inst, pmda->e_ordinal);
+			    pmInDomStr_r(pmda->e_idp->it_indom, strbuf, sizeof(strbuf)), *inst, pmda->e_ordinal);
 		    }
 #endif
 		    return 1;
@@ -175,6 +182,58 @@ __pmdaNextInst(int *inst, pmdaExt *pmda)
     return 0;
 }
 
+
+/*
+ * Helper routines for performing metric table searches.
+ *
+ * There are currently three ways - using the PMID hash that hangs off
+ * of the e_ext structure, using the direct mapping mechanism (require
+ * PMIDs be allocated one after-the-other, all in one cluster), or via
+ * a linear search of the metric table array.
+ */
+
+static pmdaMetric *
+__pmdaHashedSearch(pmID pmid, __pmHashCtl *hash)
+{
+    __pmHashNode *node;
+
+    if ((node = __pmHashSearch(pmid, hash)) == NULL)
+	return NULL;
+    return (pmdaMetric *)node->data;
+}
+
+static pmdaMetric *
+__pmdaDirectSearch(pmID pmid, pmdaExt *pmda)
+{
+    __pmID_int	*pmidp = (__pmID_int *)&pmid;
+
+    /*
+     * pmidp->domain is correct ... PMCD guarantees this, but
+     * pmda->e_direct only works for a single cluster
+     */
+    if (pmidp->item < pmda->e_nmetrics && 
+	pmidp->cluster == 
+	((__pmID_int *)&pmda->e_metrics[pmidp->item].m_desc.pmid)->cluster) {
+	/* pmidp->item is unsigned, so must be >= 0 */
+	return &pmda->e_metrics[pmidp->item];
+    }
+    return NULL;
+}
+
+static pmdaMetric *
+__pmdaLinearSearch(pmID pmid, pmdaExt *pmda)
+{
+    int		i;
+
+    for (i = 0; i < pmda->e_nmetrics; i++) {
+	if (pmda->e_metrics[i].m_desc.pmid == pmid) {
+	    /* found the hard way */
+	    return &pmda->e_metrics[i];
+	}
+    }
+    return NULL;
+}
+
 /*
  * Save the profile away for use in __pmdaNextInst() during subsequent
  * fetches ... it is the _caller_ of pmdaProfile()'s responsibility to
@@ -391,7 +450,6 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
     int			numval;
     pmValueSet		*vset;
     pmDesc		*dp;
-    __pmID_int		*pmidp;
     pmdaMetric		*metap;
     pmAtomValue		atom;
     int			type;
@@ -415,48 +473,15 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
     extp->res->numpmid = numpmid;
 
     for (i = 0; i < numpmid; i++) {
+	if (pmda->e_flags & PMDA_EXT_FLAG_HASHED)
+	    metap = __pmdaHashedSearch(pmidlist[i], &extp->hashpmids);
+	else if (pmda->e_direct)
+	    metap = __pmdaDirectSearch(pmidlist[i], pmda);
+	else
+	    metap = __pmdaLinearSearch(pmidlist[i], pmda);
 
-    	dp = NULL;
-	metap = NULL;
-	pmidp = (__pmID_int *)&pmidlist[i];
-
-	if (pmda->e_direct) {
-/*
- * pmidp->domain is correct ... PMCD gurantees this, but
- * next to check the cluster
- */
-	    if (pmidp->item < pmda->e_nmetrics &&
-		pmidlist[i] == pmda->e_metrics[pmidp->item].m_desc.pmid) {
-/* 
- * pmidp->item is unsigned, so must be >= 0 
- */
-		metap = &pmda->e_metrics[pmidp->item];
-		dp = &(metap->m_desc);
-	    }
-	}
-	else {
-	    for (j = 0; j < pmda->e_nmetrics; j++) {
-		if (pmidlist[i] == pmda->e_metrics[j].m_desc.pmid) {
-/*
- * found the hard way 
- */
-		    metap = &pmda->e_metrics[j];
-		    dp = &(metap->m_desc);
-		    break;
-		}
-	    }
-	}
-	
-	if (dp == NULL) {
-	    /* dynamic name metrics may often vanish, avoid log spam */
-	    if (extp->pmda_interface < PMDA_INTERFACE_4)
-		__pmNotifyErr(LOG_ERR,
-			"pmdaFetch: Requested metric %s is not defined",
-			 pmIDStr(pmidlist[i]));
-	    numval = PM_ERR_PMID;
-	}
-	else {
-
+	if (metap) {
+	    dp = &(metap->m_desc);
 	    if (dp->indom != PM_INDOM_NULL) {
 		/* count instances in the profile */
 		numval = 0;
@@ -471,12 +496,21 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 		numval = 1;
 	    }
 	}
+	else {
+	    dp = NULL;
+	    /* dynamic name metrics may often vanish, avoid log spam */
+	    if (extp->pmda_interface < PMDA_INTERFACE_4) {
+		char	strbuf[20];
+		__pmNotifyErr(LOG_ERR,
+			"pmdaFetch: Requested metric %s is not defined",
+			 pmIDStr_r(pmidlist[i], strbuf, sizeof(strbuf)));
+	    }
+	    numval = PM_ERR_PMID;
+	}
+
 
 	/* Must use individual malloc()s because of pmFreeResult() */
-	if (numval == 1)
-	    extp->res->vset[i] = vset = (pmValueSet *)
-	    				__pmPoolAlloc(sizeof(pmValueSet));
-	else if (numval > 1)
+	if (numval >= 1)
 	    extp->res->vset[i] = vset = (pmValueSet *)malloc(sizeof(pmValueSet) +
 					    (numval - 1)*sizeof(pmValue));
 	else
@@ -514,35 +548,39 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 	    vset->vlist[j].inst = inst;
 
 	    if ((sts = (*(pmda->e_fetchCallBack))(metap, inst, &atom)) < 0) {
+		char	strbuf[20];
 
-		if (sts == PM_ERR_PMID)
+		pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf));
+		if (sts == PM_ERR_PMID) {
 		    __pmNotifyErr(LOG_ERR, 
 		        "pmdaFetch: PMID %s not handled by fetch callback\n",
-				 pmIDStr(dp->pmid));
+				strbuf);
+		}
 		else if (sts == PM_ERR_INST) {
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_LIBPMDA) {
 			__pmNotifyErr(LOG_ERR,
 			    "pmdaFetch: Instance %d of PMID %s not handled by fetch callback\n",
-				     inst, pmIDStr(dp->pmid));
+				    inst, strbuf);
 		    }
 #endif
 		}
 		else if (sts == PM_ERR_APPVERSION ||
-			(sts == PM_ERR_AGAIN) ||
-			(sts == PM_ERR_PERMISSION)) {
+			 sts == PM_ERR_PERMISSION ||
+			 sts == PM_ERR_AGAIN) {
 #ifdef PCP_DEBUG
 		    if (pmDebug & DBG_TRACE_LIBPMDA) {
 			__pmNotifyErr(LOG_ERR,
-				 "pmdaFetch: Unavailable metric PMID %s\n",
-				 pmIDStr(dp->pmid));
+			     "pmdaFetch: Unavailable metric PMID %s[%d]\n",
+				    strbuf, inst);
 		    }
 #endif
 		}
-		else
+		else {
 		    __pmNotifyErr(LOG_ERR,
-				 "pmdaFetch: Fetch callback error: %s\n",
-				 pmErrStr(sts));
+			"pmdaFetch: Fetch callback error from metric PMID %s[%d]: %s\n",
+				strbuf, inst, pmErrStr(sts));
+		}
 	    }
 	    else {
 		/*
@@ -562,9 +600,12 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 		    int		lsts;
 
 		    if ((lsts = __pmStuffValue(&atom, &vset->vlist[j], type)) == PM_ERR_TYPE) {
+			char	strbuf[20];
+			char	st2buf[20];
 			__pmNotifyErr(LOG_ERR, 
 				     "pmdaFetch: Descriptor type (%s) for metric %s is bad",
-				     pmTypeStr(type), pmIDStr(dp->pmid));
+				     pmTypeStr_r(type, strbuf, sizeof(strbuf)),
+				     pmIDStr_r(dp->pmid, st2buf, sizeof(st2buf)));
 		    }
 		    else if (lsts >= 0) {
 			vset->valfmt = lsts;
@@ -576,9 +617,12 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			else if (type == PM_TYPE_AGGREGATE)
 			    free(atom.vbp);
 			else {
+			    char	strbuf[20];
+			    char	st2buf[20];
 			    __pmNotifyErr(LOG_WARNING,
 					  "pmdaFetch: Attempt to free value for metric %s of wrong type %s\n",
-					  pmIDStr(dp->pmid), pmTypeStr(type));
+					  pmIDStr_r(dp->pmid, strbuf, sizeof(strbuf)),
+					  pmTypeStr_r(type, st2buf, sizeof(st2buf)));
 			}
 		    }
 		    if (lsts < 0)
@@ -612,46 +656,28 @@ pmdaFetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 int
 pmdaDesc(pmID pmid, pmDesc *desc, pmdaExt *pmda)
 {
-    int			j;
-    int			sts = 0;
     e_ext_t		*extp = (e_ext_t *)pmda->e_ext;
+    pmdaMetric		*metric;
+    char		strbuf[32];
 
     if (extp->pmda_interface >= PMDA_INTERFACE_5)
 	__pmdaSetContext(pmda->e_context);
 
-    if (pmda->e_direct) {
-	__pmID_int	*pmidp = (__pmID_int *)&pmid;
-	/*
-	 * pmidp->domain is correct ... PMCD gurantees this, but
-	 * pmda->e_direct only works for a single cluster
-	 */
-	if (pmidp->item < pmda->e_nmetrics && 
-	    pmidp->cluster == 
-	     	((__pmID_int *)&pmda->e_metrics[pmidp->item].m_desc.pmid)->cluster) {
-	    /* pmidp->item is unsigned, so must be >= 0 */
-	    *desc = pmda->e_metrics[pmidp->item].m_desc;
-	}
-	else {
-	    __pmNotifyErr(LOG_ERR, "Requested metric %s is not defined",
-			 pmIDStr(pmid));
-	    sts = PM_ERR_PMID;
-	}
+    if (pmda->e_flags & PMDA_EXT_FLAG_HASHED)
+	metric = __pmdaHashedSearch(pmid, &extp->hashpmids);
+    else if (pmda->e_direct)
+	metric = __pmdaDirectSearch(pmid, pmda);
+    else
+	metric = __pmdaLinearSearch(pmid, pmda);
+
+    if (metric) {
+	*desc = metric->m_desc;
+	return 0;
     }
-    else {
-	for (j = 0; j < pmda->e_nmetrics; j++) {
-	    if (pmda->e_metrics[j].m_desc.pmid == pmid) {
-		/* found the hard way */
-		*desc = pmda->e_metrics[j].m_desc;
-		break;
-	    }
-	}
-	if (j == pmda->e_nmetrics) {
-	    __pmNotifyErr(LOG_ERR, "Requested metric %s is not defined",
-			 pmIDStr(pmid));
-	    sts = PM_ERR_PMID;
-	}
-    }
-    return sts;
+
+    __pmNotifyErr(LOG_ERR, "Requested metric %s is not defined",
+			pmIDStr_r(pmid, strbuf, sizeof(strbuf)));
+    return PM_ERR_PMID;
 }
 
 /*
@@ -685,17 +711,18 @@ pmdaText(int ident, int type, char **buffer, pmdaExt *pmda)
 int
 pmdaStore(pmResult *result, pmdaExt *pmda)
 {
-    return -EACCES;
+    return PM_ERR_PERMISSION;
 }
 
 /*
- * Expect these ones ...
- *	pmdaPMID()
- *	pmdaName()
- *	pmdaChildren()
+ * Expect routines pmdaPMID(), pmdaName() and pmdaChildren() below
  * to be overridden with real routines for any PMDA that is
  * using PMDA_INTERFACE_4 or later and supporting dynamic metrics.
  *
+ * Expect the pmdaAttribute() routine to be overridden with a real
+ * routine for a PMDA using PMDA_INTERFACE_6 or later supporting
+ * metrics whose behaviour depends on clients being authenticated.
+ *
  * These implementations are stubs that return appropriate errors
  * if they are ever called.
  */
@@ -717,3 +744,9 @@ pmdaChildren(const char *name, int traverse, char ***offspring, int **status, pm
 {
     return PM_ERR_NAME;
 }
+
+int
+pmdaAttribute(int context, int attribute, const char *value, int size, pmdaExt *pmda)
+{
+    return 0;	/* simply ignore everything by default */
+}
