@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2012-2013 Red Hat.
  * Copyright (c) 1995-2000,2003,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This library is free software; you can redistribute it and/or modify it
@@ -30,95 +31,135 @@
  * Open an inet port to PMCD
  */
 
-void
-__pmdaOpenInet(char *sockname, int myport, int *infd, int *outfd)
+static void
+__pmdaOpenSocket(char *sockname, int port, int family, int *infd, int *outfd)
 {
     int			sts;
     int			sfd;
-    struct sockaddr_in	myaddr;
-    struct sockaddr_in	from;
     struct servent	*service;
-    mysocklen_t		addrlen;
+    __pmSockAddr	*myaddr;
+    __pmSockLen		addrlen;
+    __pmFdSet		rfds;
     int			one = 1;
 
     if (sockname != NULL) {	/* Translate port name to port num */
 	service = getservbyname(sockname, NULL);
 	if (service == NULL) {
-	    __pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: getservbyname(%s): %s\n", 
+	    __pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: getservbyname(%s): %s\n", 
 		    sockname, netstrerror());
 	    exit(1);
 	}
-	myport = service->s_port;
+	port = service->s_port;
     }
 
-    sfd = __pmCreateSocket();
-    if (sfd < 0) {
-	__pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: inet socket: %s\n",
-			netstrerror());
-	exit(1);
+    if (family != AF_INET6) {
+	sfd = __pmCreateSocket();
+	if (sfd < 0) {
+	    __pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: inet socket: %s\n",
+			  netstrerror());
+	    exit(1);
+	}
+    }
+    else {
+	sfd = __pmCreateIPv6Socket();
+	if (sfd < 0) {
+	    __pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: ipv6 socket: %s\n",
+			  netstrerror());
+	    exit(1);
+	}
     }
 #ifndef IS_MINGW
     /*
      * allow port to be quickly re-used, e.g. when Install and PMDA already
      * installed, this becomes terminate and restart in a hurry ...
      */
-    if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (char *)&one,
-		(mysocklen_t)sizeof(one)) < 0) {
-	__pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: setsockopt(reuseaddr): %s\n",
+    if (__pmSetSockOpt(sfd, SOL_SOCKET, SO_REUSEADDR, (char *)&one,
+		(__pmSockLen)sizeof(one)) < 0) {
+	__pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: __pmSetSockOpt(reuseaddr): %s\n",
 			netstrerror());
 	exit(1);
     }
 #else
     /* see MSDN tech note: "Using SO_REUSEADDR and SO_EXCLUSIVEADDRUSE" */
-    if (setsockopt(sfd, SOL_SOCKET, SO_EXCLUSIVEADDRUSE, (char *)&one,
-		(mysocklen_t)sizeof(one)) < 0) {
-	__pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: setsockopt(excladdruse): %s\n",
+    if (__pmSetSockOpt(sfd, SOL_SOCKET, SO_EXCLUSIVEADDRUSE, (char *)&one,
+		(__pmSockLen)sizeof(one)) < 0) {
+	__pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: __pmSetSockOpt(excladdruse): %s\n",
 			netstrerror());
 	exit(1);
     }
 #endif
 
-    memset(&myaddr, 0, sizeof(myaddr));
-    myaddr.sin_family = AF_INET;
-    myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
-    myaddr.sin_port = htons(myport);
-    sts = bind(sfd, (struct sockaddr*) &myaddr, sizeof(myaddr));
+    if ((myaddr =__pmSockAddrAlloc()) == NULL) {
+	__pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: sock addr alloc failed\n");
+	exit(1);
+    }
+    __pmSockAddrInit(myaddr, family, INADDR_LOOPBACK, port);
+    sts = __pmBind(sfd, (void *)myaddr, __pmSockAddrSize());
     if (sts < 0) {
-	__pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: inet bind: %s\n",
+	__pmSockAddrFree(myaddr);
+	__pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: bind: %s\n",
 			netstrerror());
 	exit(1);
     }
 
-    sts = listen(sfd, 5);	/* Max. of 5 pending connection requests */
+    sts = __pmListen(sfd, 1);	/* Max. 1 pending connection request (pmcd) */
     if (sts == -1) {
-	__pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: inet listen: %s\n",
+        __pmSockAddrFree(myaddr);
+	__pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: listen: %s\n",
 			netstrerror());
 	exit(1);
     }
-    addrlen = sizeof(from);
-    /* block here, waiting for a connection */
-    if ((*infd = accept(sfd, (struct sockaddr *)&from, &addrlen)) < 0) {
-	__pmNotifyErr(LOG_CRIT, "__pmdaOpenInet: inet accept: %s\n",
+
+    /* block here indefinitely, waiting for a connection */
+    __pmFD_ZERO(&rfds);
+    __pmFD_SET(sfd, &rfds);
+    if ((sts = __pmSelectRead(sfd+1, &rfds, NULL)) != 1) {
+        sts = (sts < 0) ? -neterror() : -EINVAL;
+    }
+    if (sts < 0) {
+        __pmSockAddrFree(myaddr);
+        __pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: select: %s\n",
+                        pmErrStr(sts));
+        exit(1);
+    }
+
+    addrlen = __pmSockAddrSize();
+    if ((*infd = __pmAccept(sfd, myaddr, &addrlen)) < 0) {
+        __pmSockAddrFree(myaddr);
+	__pmNotifyErr(LOG_CRIT, "__pmdaOpenSocket: accept: %s\n",
 			netstrerror());
 	exit(1);
     }
     __pmCloseSocket(sfd);
     __pmSetSocketIPC(*infd);
+    __pmSockAddrFree(myaddr);
+
     *outfd = *infd;
 }
 
-#if !defined(IS_MINGW)
+static void
+__pmdaOpenInet(char *sockname, int port, int *infd, int *outfd)
+{
+    __pmdaOpenSocket(sockname, port, AF_INET, infd, outfd);
+}
+
+static void
+__pmdaOpenIPv6(char *sockname, int port, int *infd, int *outfd)
+{
+    __pmdaOpenSocket(sockname, port, AF_INET6, infd, outfd);
+}
+
+#ifdef HAVE_STRUCT_SOCKADDR_UN
 /*
  * Open a unix port to PMCD
  */
-
-void
+static void
 __pmdaOpenUnix(char *sockname, int *infd, int *outfd)
 {
     int			sts;
     int			sfd;
     int			len;
-    mysocklen_t		addrlen;
+    __pmSockLen		addrlen;
     struct sockaddr_un	myaddr;
     struct sockaddr_un	from;
 
@@ -171,12 +212,11 @@ __pmdaOpenUnix(char *sockname, int *infd, int *outfd)
     close(sfd);
     *outfd = *infd;
 }
-
-#else	/* MINGW */
-void
+#else
+static void
 __pmdaOpenUnix(char *sockname, int *infd, int *outfd)
 {
-    __pmNotifyErr(LOG_CRIT, "__pmdaOpenUnix: Not supported on Windows");
+    __pmNotifyErr(LOG_CRIT, "__pmdaOpenUnix: UNIX domain sockets unsupported\n");
     exit(1);
 }
 #endif
@@ -185,11 +225,30 @@ __pmdaOpenUnix(char *sockname, int *infd, int *outfd)
  * capture PMDA args from getopts 
  */
 
+static char
+pmdaIoTypeToOption(pmdaIoType io)
+{
+    switch(io) {
+        case pmdaPipe:
+	    return 'p';
+        case pmdaInet:
+	    return 'i';
+        case pmdaIPv6:
+	    return '6';
+        case pmdaUnix:
+	    return 'u';
+        case pmdaUnknown:
+        default:
+	    break;
+    }
+    return '?';
+}
+
 int 
 pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *dispatch, 
 	   int *err)
 {
-    int 	c;
+    int 	c = EOF;
     int		flag = 0;
     int		sts;
     char	*endnum = NULL;
@@ -197,19 +256,16 @@ pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *di
 
     if (dispatch->status != 0) {
 	(*err)++;
-	return EOF;
+	goto done;
     }
 
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	pmda = dispatch->version.four.ext;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	pmda = dispatch->version.two.ext;
-    else {
+    if (!HAVE_ANY(dispatch->comm.pmda_interface)) {
 	__pmNotifyErr(LOG_CRIT, "pmdaGetOpt: PMDA interface version %d not supported (domain=%d)",
 		     dispatch->comm.pmda_interface, dispatch->domain);
 	(*err)++;
-	return EOF;
+	goto done;
     }
+    pmda = dispatch->version.any.ext;
 
     while (!flag && ((c = getopt(argc, argv, optstring)) != EOF)) {
     	switch (c) {
@@ -241,7 +297,7 @@ pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *di
 	    case 'i':
 		if (pmda->e_io != pmdaUnknown && pmda->e_io != pmdaInet) {
 		    fprintf(stderr, "%s: -i option clashes with -%c option\n",
-			    pmda->e_name, (pmda->e_io == pmdaUnix) ? 'u' : 'p');
+			    pmda->e_name, pmdaIoTypeToOption(pmda->e_io));
 		    (*err)++;
 		    break;
 		}
@@ -259,7 +315,7 @@ pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *di
 	    case 'p':
 		if (pmda->e_io != pmdaUnknown && pmda->e_io != pmdaPipe) {
 		    fprintf(stderr, "%s: -p option clashes with -%c option\n",
-			    pmda->e_name, (pmda->e_io == pmdaUnix) ? 'u' : 'i');
+			    pmda->e_name, pmdaIoTypeToOption(pmda->e_io));
 		    (*err)++;
 		    break;
 		}
@@ -269,7 +325,7 @@ pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *di
 	    case 'u':
 		if (pmda->e_io != pmdaUnknown && pmda->e_io != pmdaUnix) {
 		    fprintf(stderr, "%s: -u option clashes with -%c option\n",
-			    pmda->e_name, (pmda->e_io == pmdaPipe) ? 'p' : 'i');
+			    pmda->e_name, pmdaIoTypeToOption(pmda->e_io));
 		    (*err)++;
 		    break;
 		}
@@ -277,6 +333,19 @@ pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *di
 		pmda->e_sockname = optarg;
 		break;
 
+	    case '6':
+		if (pmda->e_io != pmdaUnknown && pmda->e_io != pmdaIPv6) {
+		    fprintf(stderr, "%s: -6 option clashes with -%c option\n",
+			    pmda->e_name, pmdaIoTypeToOption(pmda->e_io));
+		    (*err)++;
+		    break;
+		}
+		pmda->e_io = pmdaIPv6;
+		pmda->e_port = (int)strtol(optarg, &endnum, 10);
+		if (*endnum != '\0')
+		    pmda->e_sockname = optarg;
+		break;
+
 	    case '?':
 		(*err)++;
 		break;
@@ -286,16 +355,103 @@ pmdaGetOpt(int argc, char *const *argv, const char *optstring, pmdaInterface *di
 	}
     }
 
+done:
+#ifdef HAVE_UNSETENV
+    if (c == EOF) {
+	/* Clear this environment variable PMCD set for us.
+         * This is for any commands that the PMDA may run.
+         */
+	unsetenv("POSIXLY_CORRECT");
+    }
+#endif
     return c;
 }
 
 /*
- * open the help text file and check for direct mapping into the metric table
+ * Recompute the hash table which maps metric PMIDs to metric table
+ * offsets.  Provides an optimised lookup alternative when a direct
+ * mapping is inappropriate or impossible.
+ */
+void
+pmdaRehash(pmdaExt *pmda, pmdaMetric *metrics, int nmetrics)
+{
+    e_ext_t	*extp = (e_ext_t *)pmda->e_ext;
+    __pmHashCtl *hashp = &extp->hashpmids;
+    pmdaMetric	*metric;
+    char	buf[32];
+    int		m, sts;
+
+    pmda->e_direct = 0;
+    pmda->e_metrics = metrics;
+    pmda->e_nmetrics = nmetrics;
+
+    __pmHashClear(hashp);
+    for (m = 0; m < pmda->e_nmetrics; m++) {
+	metric = &pmda->e_metrics[m];
+
+	sts = __pmHashAdd(metric->m_desc.pmid, metric, hashp);
+	if (sts < 0) {
+	    __pmNotifyErr(LOG_WARNING, "pmdaRehash: PMDA %s: "
+			"Hashed mapping for metrics disabled @ metric[%d] %s\n",
+			pmda->e_name, m,
+			pmIDStr_r(metric->m_desc.pmid, buf, sizeof(buf)));
+	    pmda->e_flags &= ~PMDA_EXT_FLAG_HASHED;
+	    __pmHashClear(hashp);
+	    return;
+	}
+    }
+
+    pmda->e_flags |= PMDA_EXT_FLAG_HASHED;
+}
+
+static void
+pmdaDirect(pmdaExt *pmda, pmdaMetric *metrics, int nmetrics)
+{
+    __pmID_int	*pmidp;
+    char	buf[20];
+    int		m;
+
+    pmda->e_direct = 1;
+    for (m = 0; m < pmda->e_nmetrics; m++) {
+	pmidp = (__pmID_int *)&pmda->e_metrics[m].m_desc.pmid;
+
+	if (pmidp->item == m)
+	    continue;
+
+	pmda->e_direct = 0;
+	if ((pmda->e_flags & PMDA_EXT_FLAG_DIRECT) ||
+	    (pmDebug & DBG_TRACE_LIBPMDA))
+	    __pmNotifyErr(LOG_WARNING, "pmdaInit: PMDA %s: "
+		"Direct mapping for metrics disabled @ metrics[%d] %s\n",
+		pmda->e_name, m,
+		pmIDStr_r(pmda->e_metrics[m].m_desc.pmid, buf, sizeof(buf)));
+	break;
+    }
+}
+
+void
+pmdaSetFlags(pmdaInterface *dispatch, int flags)
+{
+    pmdaExt	*pmda;
+
+    if (!HAVE_ANY(dispatch->comm.pmda_interface)) {
+	__pmNotifyErr(LOG_CRIT, "pmdaSetFlags: PMDA interface version %d not supported (domain=%d)",
+		     dispatch->comm.pmda_interface, dispatch->domain);
+	dispatch->status = PM_ERR_GENERIC;
+	return;
+    }
+    pmda = dispatch->version.any.ext;
+    pmda->e_flags |= flags;
+}
+
+/*
+ * Open the help text file, check for direct mapping into the metric table
+ * and whether a hash mapping has been requested.
  */
 
 void
-pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *metrics, 
-	 int nmetrics)
+pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms,
+	 pmdaMetric *metrics, int nmetrics)
 {
     int		        m = 0;
     int                 i = 0;
@@ -303,23 +459,16 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *me
     __pmInDom_int        *mindomp = NULL;
     __pmID_int	        *pmidp = NULL;
     pmdaExt	        *pmda = NULL;
-    int			serial;
 
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	pmda = dispatch->version.four.ext;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	pmda = dispatch->version.two.ext;
-    else {
+    if (!HAVE_ANY(dispatch->comm.pmda_interface)) {
 	__pmNotifyErr(LOG_CRIT, "pmdaInit: PMDA interface version %d not supported (domain=%d)",
 		     dispatch->comm.pmda_interface, dispatch->domain);
 	dispatch->status = PM_ERR_GENERIC;
 	return;
     }
+    pmda = dispatch->version.any.ext;
 
-    if (((HAVE_V_FOUR(dispatch->comm.pmda_interface) &&
-	  dispatch->version.four.fetch == pmdaFetch) ||
-	 (HAVE_V_TWO(dispatch->comm.pmda_interface) &&
-	  dispatch->version.two.fetch == pmdaFetch)) &&
+    if (dispatch->version.any.fetch == pmdaFetch &&
 	pmda->e_fetchCallBack == (pmdaFetchCallBack)0) {
 	__pmNotifyErr(LOG_CRIT, "pmdaInit: PMDA %s: using pmdaFetch() but fetch call back not set", pmda->e_name);
 	dispatch->status = PM_ERR_GENERIC;
@@ -355,24 +504,19 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *me
 	return;
     }
     
-
     /* fix bit fields in indom for all instance domains */
     for (i = 0; i < pmda->e_nindoms; i++) {
-	serial = pmda->e_indoms[i].it_indom;
-	indomp = (__pmInDom_int *)&(pmda->e_indoms[i].it_indom);
-	indomp->serial = serial;
-	indomp->flag = 0;
-	indomp->domain = dispatch->domain;
+	unsigned int domain = dispatch->domain;
+	unsigned int serial = pmda->e_indoms[i].it_indom;
+	pmda->e_indoms[i].it_indom = pmInDom_build(domain, serial);
     }
 
     /* fix bit fields in indom for all metrics */
     for (i = 0; i < pmda->e_nmetrics; i++) {
 	if (pmda->e_metrics[i].m_desc.indom != PM_INDOM_NULL) {
-	    serial = pmda->e_metrics[i].m_desc.indom;
-	    indomp = (__pmInDom_int *)&(pmda->e_metrics[i].m_desc.indom);
-	    indomp->serial = serial;
-	    indomp->flag = 0;
-	    indomp->domain = dispatch->domain;
+	    unsigned int domain = dispatch->domain;
+	    unsigned int serial = pmda->e_metrics[i].m_desc.indom;
+	    pmda->e_metrics[i].m_desc.indom = pmInDom_build(domain, serial);
 	}
     }
 
@@ -388,21 +532,24 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *me
 		    if (indomp->serial == mindomp->serial) {
 #ifdef PCP_DEBUG
 			if (pmDebug & DBG_TRACE_LIBPMDA) {
+			    char	strbuf[20];
+			    char	st2buf[20];
 			    __pmNotifyErr(LOG_DEBUG, 
 				    "pmdaInit: PMDA %s: Metric %s(%d) matched to indom %s(%d)\n",
 				    pmda->e_name,
-				    pmIDStr(pmda->e_metrics[m].m_desc.pmid), m,
-				    pmInDomStr(pmda->e_indoms[i].it_indom), i);
+				    pmIDStr_r(pmda->e_metrics[m].m_desc.pmid, strbuf, sizeof(strbuf)), m,
+				    pmInDomStr_r(pmda->e_indoms[i].it_indom, st2buf, sizeof(st2buf)), i);
 			}
 #endif
 			break;
 		    }
 		}
 		if (i == pmda->e_nindoms) {
+		    char	strbuf[20];
 		    __pmNotifyErr(LOG_CRIT, 
 				 "pmdaInit: PMDA %s: Undefined instance domain serial (%d) specified in metric %s(%d)\n",
 				 pmda->e_name, mindomp->serial, 
-				 pmIDStr(pmda->e_metrics[m].m_desc.pmid), m);
+				 pmIDStr_r(pmda->e_metrics[m].m_desc.pmid, strbuf, sizeof(strbuf)), m);
 		    dispatch->status = PM_ERR_GENERIC;
 		    return;
 		}
@@ -414,7 +561,7 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *me
     if (pmda->e_helptext != NULL) {
 	pmda->e_help = pmdaOpenHelp(pmda->e_helptext);
 	if (pmda->e_help < 0) {
-	    __pmNotifyErr(LOG_WARNING, "pmdaInit: PMDA %s: Unable to open help text file \"%s\": %s\n",
+	    __pmNotifyErr(LOG_WARNING, "pmdaInit: PMDA %s: Unable to open help text file(s) from \"%s\": %s\n",
 		    pmda->e_name, pmda->e_helptext, pmErrStr(pmda->e_help));
 	}
 #ifdef PCP_DEBUG
@@ -433,30 +580,30 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *me
 #endif
     }
 
-    pmda->e_direct = 1;
-
-    for (m = 0; m < pmda->e_nmetrics; m++)
-    {
-    	pmidp = (__pmID_int *)&pmda->e_metrics[m].m_desc.pmid;
+    /*
+     * Stamp the correct domain number in each of the PMIDs
+     */
+    for (m = 0; m < pmda->e_nmetrics; m++) {
+	pmidp = (__pmID_int *)&pmda->e_metrics[m].m_desc.pmid;
 	pmidp->domain = dispatch->domain;
-
-	if (pmda->e_direct && pmidp->item != m) {
-	    pmda->e_direct = 0;
-#ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_LIBPMDA)
-		__pmNotifyErr(LOG_WARNING, "pmdaInit: PMDA %s: Direct mapping for metrics disabled @ metrics[%d] %s\n", pmda->e_name, m, pmIDStr(pmda->e_metrics[m].m_desc.pmid));
-#endif
-	}
     }
 
+    if (pmda->e_flags & PMDA_EXT_FLAG_HASHED)
+	pmdaRehash(pmda, metrics, nmetrics);
+    else
+	pmdaDirect(pmda, metrics, nmetrics);
+
 #ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_LIBPMDA)
-    {
+    if (pmDebug & DBG_TRACE_LIBPMDA) {
     	__pmNotifyErr(LOG_INFO, "name        = %s\n", pmda->e_name);
-	__pmNotifyErr(LOG_INFO, "domain      = %d\n", dispatch->domain);
+        __pmNotifyErr(LOG_INFO, "domain      = %d\n", dispatch->domain);
+        if (dispatch->comm.flags)
+    	    __pmNotifyErr(LOG_INFO, "comm flags  = %x\n", dispatch->comm.flags);
     	__pmNotifyErr(LOG_INFO, "num metrics = %d\n", pmda->e_nmetrics);
     	__pmNotifyErr(LOG_INFO, "num indom   = %d\n", pmda->e_nindoms);
-    	__pmNotifyErr(LOG_INFO, "direct map  = %d\n", pmda->e_direct);
+    	__pmNotifyErr(LOG_INFO, "metric map  = %s\n",
+		(pmda->e_flags & PMDA_EXT_FLAG_HASHED) ? "hashed" :
+		(pmda->e_direct ? "direct" : "linear"));
     }
 #endif
 
@@ -467,25 +614,24 @@ pmdaInit(pmdaInterface *dispatch, pmdaIndom *indoms, int nindoms, pmdaMetric *me
  * version exchange with pmcd via credentials PDU
  */
 
-int
-__pmdaSetupPDU(int infd, int outfd, char *agentname)
+static int
+__pmdaSetupPDU(int infd, int outfd, int flags, char *agentname)
 {
-    __pmCred	handshake[1];
-    __pmCred	*credlist = NULL;
-    __pmPDU	*pb;
-    int		i, sts, vflag = 0;
-    int		version = UNKNOWN_VERSION, credcount = 0, sender = 0;
+    __pmVersionCred	handshake;
+    __pmCred		*credlist = NULL;
+    __pmPDU		*pb;
+    int			i, sts, pinpdu, vflag = 0;
+    int			version = UNKNOWN_VERSION, credcount = 0, sender = 0;
 
-    handshake[0].c_type = CVERSION;
-    handshake[0].c_vala = PDU_VERSION;
-    handshake[0].c_valb = 0;
-    handshake[0].c_valc = 0;
-    if ((sts = __pmSendCreds(outfd, getpid(), 1, handshake)) < 0) {
+    handshake.c_type = CVERSION;
+    handshake.c_version = PDU_VERSION;
+    handshake.c_flags = flags;
+    if ((sts = __pmSendCreds(outfd, (int)getpid(), 1, (__pmCred *)&handshake)) < 0) {
 	__pmNotifyErr(LOG_CRIT, "__pmdaSetupPDU: PMDA %s send creds: %s\n", agentname, pmErrStr(sts));
 	return -1;
     }
 
-    if ((sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_DEFAULT, &pb)) < 0) {
+    if ((pinpdu = sts = __pmGetPDU(infd, ANY_SIZE, TIMEOUT_DEFAULT, &pb)) < 0) {
 	__pmNotifyErr(LOG_CRIT, "__pmdaSetupPDU: PMDA %s getting creds: %s\n", agentname, pmErrStr(sts));
 	return -1;
     }
@@ -493,6 +639,7 @@ __pmdaSetupPDU(int infd, int outfd, char *agentname)
     if (sts == PDU_CREDS) {
 	if ((sts = __pmDecodeCreds(pb, &sender, &credcount, &credlist)) < 0) {
 	    __pmNotifyErr(LOG_CRIT, "__pmdaSetupPDU: PMDA %s decode creds: %s\n", agentname, pmErrStr(sts));
+	    __pmUnpinPDUBuf(pb);
 	    return -1;
 	}
 
@@ -516,6 +663,9 @@ __pmdaSetupPDU(int infd, int outfd, char *agentname)
     else
 	__pmNotifyErr(LOG_CRIT, "__pmdaSetupPDU: PMDA %s: version exchange failure\n", agentname);
 
+    if (pinpdu > 0)
+	__pmUnpinPDUBuf(pb);
+
     return version;
 }
 
@@ -527,18 +677,15 @@ void
 pmdaConnect(pmdaInterface *dispatch)
 {
     pmdaExt	*pmda = NULL;
-    int		sts;
+    int		sts, flags = dispatch->comm.flags;
 
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	pmda = dispatch->version.four.ext;
-    else if (HAVE_V_TWO(dispatch->comm.pmda_interface))
-	pmda = dispatch->version.two.ext;
-    else {
+    if (!HAVE_ANY(dispatch->comm.pmda_interface)) {
 	__pmNotifyErr(LOG_CRIT, "pmdaConnect: PMDA interface version %d not supported (domain=%d)",
 		     dispatch->comm.pmda_interface, dispatch->domain);
 	dispatch->status = PM_ERR_GENERIC;
 	return;
     }
+    pmda = dispatch->version.any.ext;
 
     switch (pmda->e_io) {
 	case pmdaPipe:
@@ -558,7 +705,7 @@ pmdaConnect(pmdaInterface *dispatch)
 	case pmdaInet:
 
 	    __pmdaOpenInet(pmda->e_sockname, pmda->e_port, &(pmda->e_infd), 
-	    		  &(pmda->e_outfd));
+			   &(pmda->e_outfd));
 
 #ifdef PCP_DEBUG
 	    if (pmDebug & DBG_TRACE_LIBPMDA) {
@@ -569,6 +716,20 @@ pmdaConnect(pmdaInterface *dispatch)
 
 	    break;
 
+	case pmdaIPv6:
+
+	    __pmdaOpenIPv6(pmda->e_sockname, pmda->e_port, &(pmda->e_infd), 
+			   &(pmda->e_outfd));
+
+#ifdef PCP_DEBUG
+	    if (pmDebug & DBG_TRACE_LIBPMDA) {
+	    	__pmNotifyErr(LOG_DEBUG, "pmdaConnect: PMDA %s: opened ipv6 connection, infd = %d, outfd = %d\n",
+			     pmda->e_name, pmda->e_infd, pmda->e_outfd);
+	    }
+#endif
+
+	    break;
+
 	case pmdaUnix:
 
 	    __pmdaOpenUnix(pmda->e_sockname, &(pmda->e_infd), &(pmda->e_outfd));
@@ -586,33 +747,31 @@ pmdaConnect(pmdaInterface *dispatch)
 	    exit(1);
     }
 
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface) ||
-        HAVE_V_TWO(dispatch->comm.pmda_interface)) {
-	if ((sts = __pmdaSetupPDU(pmda->e_infd, pmda->e_outfd, pmda->e_name)) < 0)
-	    dispatch->status = sts;
-	else
-	    dispatch->comm.pmapi_version = (unsigned int)sts;
-    }
+    sts = __pmdaSetupPDU(pmda->e_infd, pmda->e_outfd, flags, pmda->e_name);
+    if (sts < 0)
+	dispatch->status = sts;
+    else
+	dispatch->comm.pmapi_version = (unsigned int)sts;
 }
 
 /*
  * initialise the pmdaExt and pmdaInterface structures for a daemon or DSO PMDA.
  */
 
-void
+static void
 __pmdaSetup(pmdaInterface *dispatch, int version, char *name)
 {
     pmdaExt	*pmda = NULL;
     e_ext_t	*extp;
 
-    if (!HAVE_V_FOUR(version) && !HAVE_V_TWO(version)) {
+    if (!HAVE_ANY(version)) {
 	__pmNotifyErr(LOG_CRIT, "__pmdaSetup: %s PMDA: interface version %d not supported (domain=%d)",
 		     name, version, dispatch->domain);
 	dispatch->status = PM_ERR_GENERIC;
 	return;
     }
 
-    pmda = (pmdaExt *)malloc(sizeof(pmdaExt));
+    pmda = (pmdaExt *)calloc(1, sizeof(pmdaExt));
     if (pmda == NULL) {
 	__pmNotifyErr(LOG_ERR, 
 		     "%s: Unable to allocate memory for pmdaExt structure (%d bytes)",
@@ -627,66 +786,44 @@ __pmdaSetup(pmdaInterface *dispatch, int version, char *name)
     dispatch->comm.pmapi_version = PMAPI_VERSION;
     dispatch->comm.flags = 0;
 
+    if (HAVE_V_SIX(version)) {
+	dispatch->version.six.attribute = pmdaAttribute;
+    }
     if (HAVE_V_FOUR(version)) {
-	dispatch->version.four.profile = pmdaProfile;
-	dispatch->version.four.fetch = pmdaFetch;
-	dispatch->version.four.desc = pmdaDesc;
-	dispatch->version.four.instance = pmdaInstance;
-	dispatch->version.four.text = pmdaText;
-	dispatch->version.four.store = pmdaStore;
 	dispatch->version.four.pmid = pmdaPMID;
 	dispatch->version.four.name = pmdaName;
 	dispatch->version.four.children = pmdaChildren;
-	dispatch->version.four.ext = pmda;
-    }
-    else {
-	dispatch->version.two.profile = pmdaProfile;
-	dispatch->version.two.fetch = pmdaFetch;
-	dispatch->version.two.desc = pmdaDesc;
-	dispatch->version.two.instance = pmdaInstance;
-	dispatch->version.two.text = pmdaText;
-	dispatch->version.two.store = pmdaStore;
-	dispatch->version.two.ext = pmda;
     }
+    dispatch->version.any.profile = pmdaProfile;
+    dispatch->version.any.fetch = pmdaFetch;
+    dispatch->version.any.desc = pmdaDesc;
+    dispatch->version.any.instance = pmdaInstance;
+    dispatch->version.any.text = pmdaText;
+    dispatch->version.any.store = pmdaStore;
+    dispatch->version.any.ext = pmda;
 
-    pmda->e_sockname = NULL;
     pmda->e_name = name;
-    pmda->e_logfile = NULL;
-    pmda->e_helptext = NULL;
-    pmda->e_status = 0;
     pmda->e_infd = -1;
     pmda->e_outfd = -1;
     pmda->e_port = -1;
     pmda->e_singular = -1;
     pmda->e_ordinal = -1;
-    pmda->e_direct = 0;
     pmda->e_domain = dispatch->domain;
-    pmda->e_nmetrics = 0;
-    pmda->e_nindoms = 0;
     pmda->e_help = -1;
-    pmda->e_prof = NULL;
     pmda->e_io = pmdaUnknown;
-    pmda->e_indoms = NULL;
-    pmda->e_idp = NULL;
-    pmda->e_metrics = NULL;
     pmda->e_ext = (void *)dispatch;
 
-    extp = (e_ext_t *)malloc(sizeof(*extp));
+    extp = (e_ext_t *)calloc(1, sizeof(*extp));
     if (extp == NULL) {
 	__pmNotifyErr(LOG_ERR, 
 		     "%s: Unable to allocate memory for e_ext_t structure (%d bytes)",
 		     name, (int)sizeof(*extp));
 	free(pmda);
-	if (HAVE_V_FOUR(version))
-	    dispatch->version.four.ext = NULL;
-	else
-	    dispatch->version.two.ext = NULL;
+	dispatch->version.any.ext = NULL;
 	dispatch->status = PM_ERR_GENERIC;
 	return;
     }
     extp->pmda_interface = version;
-    extp->res = NULL;
-    extp->maxnpmids = 0;
     pmda->e_ext = (void *)extp;
 
     pmdaSetResultCallBack(dispatch, __pmFreeResultValues);
@@ -705,7 +842,7 @@ void
 pmdaDaemon(pmdaInterface *dispatch, int version, char *name, int domain, 
 	   char *logfile, char *helptext)
 {
-    pmdaExt	*pmda = NULL;
+    pmdaExt	*pmda;
 
     dispatch->domain = domain;
     __pmdaSetup(dispatch, version, name);
@@ -713,10 +850,7 @@ pmdaDaemon(pmdaInterface *dispatch, int version, char *name, int domain,
     if (dispatch->status < 0)
 	return;
 
-    if (HAVE_V_FOUR(version))
-	pmda = dispatch->version.four.ext;
-    else
-	pmda = dispatch->version.two.ext;
+    pmda = dispatch->version.any.ext;
     pmda->e_logfile = logfile;
     pmda->e_helptext = helptext;
 
@@ -736,10 +870,7 @@ pmdaDSO(pmdaInterface *dispatch, int version, char *name, char *helptext)
     if (dispatch->status < 0)
 	return;
 
-    if (HAVE_V_FOUR(version))
-	dispatch->version.four.ext->e_helptext = helptext;
-    else
-	dispatch->version.two.ext->e_helptext = helptext;
+    dispatch->version.any.ext->e_helptext = helptext;
 }
 
 /*
@@ -751,21 +882,9 @@ pmdaOpenLog(pmdaInterface *dispatch)
 {
     int c;
 
-    if (dispatch->status != 0)
+    if (dispatch->status < 0)
 	return;
 
-    if (!HAVE_V_FOUR(dispatch->comm.pmda_interface) &&
-        !HAVE_V_TWO(dispatch->comm.pmda_interface)) {
-	__pmNotifyErr(LOG_CRIT, "pmdaOpenLog: PMDA interface version %d not supported (domain=%d)",
-		     dispatch->comm.pmda_interface, dispatch->domain);
-	dispatch->status = PM_ERR_GENERIC;
-	return;
-    }
-
-    if (HAVE_V_FOUR(dispatch->comm.pmda_interface))
-	__pmOpenLog(dispatch->version.four.ext->e_name, 
-    	       dispatch->version.four.ext->e_logfile, stderr, &c);
-    else
-	__pmOpenLog(dispatch->version.two.ext->e_name, 
-    	       dispatch->version.two.ext->e_logfile, stderr, &c);
+    __pmOpenLog(dispatch->version.any.ext->e_name, 
+		dispatch->version.any.ext->e_logfile, stderr, &c);
 }
