@@ -3,6 +3,7 @@
 #
 # Copyright (c) 1995-2001,2003 Silicon Graphics, Inc.  All Rights Reserved.
 # Portions Copyright (c) 2008 Aconex.  All Rights Reserved.
+# Portions Copyright (c) 2013 Red Hat.
 # 
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -13,16 +14,13 @@
 # WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 # or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 # for more details.
-# 
-# You should have received a copy of the GNU General Public License along
-# with this program; if not, write to the Free Software Foundation, Inc.,
-# 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 
 # source the PCP configuration environment variables
 . $PCP_DIR/etc/pcp.env
 
-tmp=/var/tmp/$$
-trap "rm -f $tmp; exit" 0 1 2 3 15
+tmp=`mktemp -d $PCP_TMPFILE_DIR/pmdaproc.XXXXXXXXX` || exit 1
+pmdatmp=$tmp
+trap "rm -rf $pmdatmp $pcptmp; exit" 0 1 2 3 15
 
 _setup_platform()
 {
@@ -59,17 +57,17 @@ _setup_localhost()
 	# Larry Wall style hunt for the version of ping that is first
 	# on our path
 	#
-	ping --help >$tmp 2>&1
-	if grep '.-c count' $tmp >/dev/null 2>&1
+	ping --help >$tmp/hlp 2>&1
+	if grep '.-c count' $tmp/hlp >/dev/null 2>&1
 	then
 	    __opt='-c 1 localhost'
-	elif grep '.-n count' $tmp >/dev/null 2>&1
+	elif grep '.-n count' $tmp/hlp >/dev/null 2>&1
 	then
 	    __opt='-n 1 localhost'
-	elif grep 'host .*packetsize .*count' $tmp >/dev/null 2>&1
+	elif grep 'host .*packetsize .*count' $tmp/hlp >/dev/null 2>&1
 	then
 	    __opt='localhost 56 1'
-	elif grep 'host .*data_size.*npackets' $tmp >/dev/null 2>&1
+	elif grep 'host .*data_size.*npackets' $tmp/hlp >/dev/null 2>&1
 	then
 	    __opt='localhost 56 1'
 	fi
@@ -77,7 +75,7 @@ _setup_localhost()
 	then
 	    echo "Warning: can't find a ping(1) that I understand ... pushing on"
 	else
-	    if ping $__opt >$tmp 2>&1
+	    if ping $__opt >$tmp/hlp 2>&1
 	    then
 		:
 	    else
@@ -99,7 +97,7 @@ _setup_localhost()
 
 _setup_localhost
 _setup_platform
-rm -f $tmp
+rm -rf $tmp
 
 # some useful common variables for Install/Remove scripts
 #
@@ -139,6 +137,8 @@ ECHONL="echo -n"
 dso_opt=false
 #	Can install as perl script?
 perl_opt=false
+#	Can install as python script?
+python_opt=false
 #	Can install as daemon?
 daemon_opt=true
 #	If daemon, pipe?
@@ -148,6 +148,8 @@ socket_opt=$default_socket_opt
 socket_inet_def=''
 #	IPC Protocol for daemon (binary only now)
 ipc_prot=binary
+#	Need to force a restart of pmcd?
+forced_restart=true
 #	Delay after install before checking (sec)
 check_delay=3
 #	Additional command line args to go in $PCP_PMCDCONF_PATH
@@ -163,7 +165,8 @@ pmda_interface=1
 # Other variables and constants
 #
 prog=`basename $0`
-tmp=$PCP_TMP_DIR/$$
+tmp=`mktemp -d $PCP_TMPFILE_DIR/pcp.XXXXXXXXX` || exit 1
+pcptmp=$tmp
 do_pmda=true
 do_check=true
 __here=`pwd`
@@ -172,7 +175,7 @@ __echo=false
 __verbose=false
 __ns_opt=''
 
-trap "rm -f $tmp $tmp.*; exit" 0 1 2 3 15
+trap "rm -rf $tmp; exit" 0 1 2 3 15
 
 # Parse command line args
 #
@@ -265,7 +268,7 @@ __wait_for_pmcd()
 #
 __restore_pmcd()
 {
-    if [ -f $tmp.pmcd.conf.save ]
+    if [ -f $tmp/pmcd.conf.save ]
     then
 	__pmcd_is_dead=false
 	echo
@@ -273,10 +276,10 @@ __restore_pmcd()
 	rm -f $PCP_PMCDCONF_PATH.prev
 	mv $PCP_PMCDCONF_PATH $PCP_PMCDCONF_PATH.prev
 	echo "Restoring previous PMCD control file, and trying to restart PMCD ..."
-	cp $tmp.pmcd.conf.save $PCP_PMCDCONF_PATH
+	cp $tmp/pmcd.conf.save $PCP_PMCDCONF_PATH
 	eval $CHOWN root $PCP_PMCDCONF_PATH
 	eval $CHMOD 644 $PCP_PMCDCONF_PATH
-	rm -f $tmp.pmcd.conf.save
+	rm -f $tmp/pmcd.conf.save
 	$PCP_RC_DIR/pcp start
 	__wait_for_pmcd
     fi
@@ -314,7 +317,7 @@ __pmda_cull()
 
     # remove matching entry from $PCP_PMCDCONF_PATH if present
     #
-    $PCP_AWK_PROG <$PCP_PMCDCONF_PATH >/tmp/$$.pmcd.conf '
+    $PCP_AWK_PROG <$PCP_PMCDCONF_PATH >$tmp/pmcd.conf '
 BEGIN					{ status = 0 }
 $1 == "'"$1"'" && $2 == "'"$2"'" 	{ status = 1; next }
 					{ print }
@@ -327,13 +330,13 @@ END					{ exit status }'
 	
 	# log change to the PCP NOTICES file
 	#
-	pmpost "PMDA cull: from $PCP_PMCDCONF_PATH: $1 $2"
+	$PCP_BINADM_DIR/pmpost "PMDA cull: from $PCP_PMCDCONF_PATH: $1 $2"
 
 	# save pmcd.conf in case we encounter a problem, and then
 	# install updated $PCP_PMCDCONF_PATH
 	#
-	cp $PCP_PMCDCONF_PATH $tmp.pmcd.conf.save
-	cp /tmp/$$.pmcd.conf $PCP_PMCDCONF_PATH
+	cp $PCP_PMCDCONF_PATH $tmp/pmcd.conf.save
+	cp $tmp/pmcd.conf $PCP_PMCDCONF_PATH
 	eval $CHOWN root $PCP_PMCDCONF_PATH
 	eval $CHMOD 644 $PCP_PMCDCONF_PATH
 
@@ -354,7 +357,7 @@ END					{ exit status }'
 	    fi
 	fi
     fi
-    rm -f /tmp/$$.pmcd.conf
+    rm -f $tmp/pmcd.conf
 
     # stop any matching PMDA that is still running
     #
@@ -398,34 +401,35 @@ __pmda_add()
 
     # save pmcd.conf in case we encounter a problem
     #
-    cp $PCP_PMCDCONF_PATH $tmp.pmcd.conf.save
+    cp $PCP_PMCDCONF_PATH $tmp/pmcd.conf.save
 
     myname=`echo $1 | $PCP_AWK_PROG '{print $1}'`
     mydomain=`echo $1 | $PCP_AWK_PROG '{print $2}'`
     # add entry to $PCP_PMCDCONF_PATH
     #
-    echo >/tmp/$$.pmcd.access
+    echo >$tmp/pmcd.body
+    echo >$tmp/pmcd.access
     $PCP_AWK_PROG <$PCP_PMCDCONF_PATH '
 NF==0					{ next }
 /^[      ]*\[[   ]*access[       ]*\]/	{ state = 2 }
-state == 2				{ print >"'/tmp/$$.pmcd.access'"; next }
+state == 2				{ print >"'$tmp/pmcd.access'"; next }
 $1=="'$myname'" && $2=="'$mydomain'"	{ next }
-					{ print >"'/tmp/$$.pmcd.body'"; next }'
-    ( cat /tmp/$$.pmcd.body \
+					{ print >"'$tmp/pmcd.body'"; next }'
+    ( cat $tmp/pmcd.body \
       ; echo "$1" \
-      ; cat /tmp/$$.pmcd.access \
+      ; cat $tmp/pmcd.access \
     ) >$PCP_PMCDCONF_PATH
-    rm -f /tmp/$$.pmcd.access /tmp/$$.pmcd.body
+    rm -f $tmp/pmcd.access $tmp/pmcd.body
     eval $CHOWN root $PCP_PMCDCONF_PATH
     eval $CHMOD 644 $PCP_PMCDCONF_PATH
 
     # log change to pcplog/NOTICES
     #
-    pmpost "PMDA add: to $PCP_PMCDCONF_PATH: $1"
+    $PCP_BINADM_DIR/pmpost "PMDA add: to $PCP_PMCDCONF_PATH: $1"
 
-    # signal pmcd if it is running, else start it
+    # signal pmcd if it is running (and ok to do so), else start it
     #
-    if pminfo -v pmcd.version >/dev/null 2>&1
+    if ! $forced_restart && pminfo -v pmcd.version >/dev/null 2>&1
     then
 	pmsignal -a -s HUP pmcd >/dev/null 2>&1
 	# allow signal processing to be done before checking status
@@ -469,6 +473,9 @@ __check_domain()
     elif [ -f domain.h.perl ]
     then
 	__infile=domain.h.perl
+    elif [ -f domain.h.python ]
+    then
+	__infile=domain.h.python
     else
 	echo "Install: cannot find ./domain.h to determine the Performance Metrics Domain"
 	exit 1
@@ -748,10 +755,10 @@ __choose_ipc()
     else
 	while true
 	do
-	    $PCP_ECHO_PROG $PCP_ECHO_N "Use Internet or Unix domain sockets? [Internet] ""$PCP_ECHO_C"
+	    $PCP_ECHO_PROG $PCP_ECHO_N "Use Internet, IPv6 or Unix domain sockets? [Internet] ""$PCP_ECHO_C"
 	    read ans
 	    $__echo && echo "$ans"
-	    if [ "X$ans" = XInternet -o "X$ans" = X ]
+	    if [ "X$ans" = XInternet -o "X$ans" = XIPv6 -o "X$ans" = X ]
 	    then
 		$PCP_ECHO_PROG $PCP_ECHO_N "Internet port number or service name? [$socket_inet_def] ""$PCP_ECHO_C"
 		read port
@@ -770,8 +777,14 @@ __choose_ipc()
 			    fi
 			    ;;
 		esac
-		type="socket	inet $port	$_dir/$pmda_name"
-		args="-i $port $args"
+		if [ "X$ans" = XInternet -o "X$ans" = X ]
+		then
+		    type="socket	inet $port	$_dir/$pmda_name"
+		    args="-i $port $args"
+		else
+		    type="socket	ipv6 $port	$_dir/$pmda_name"
+		    args="-6 $port $args"
+		fi
 		break
 	    elif [ "X$ans" = XUnix ]
 	    then
@@ -793,30 +806,26 @@ __choose_ipc()
     fi
 }
 
-# Function to filter pminfo output
+# filter pmprobe -i output of the format:
+#	postgresql.active.is_in_recovery -12351 Missing metric value(s)
+#	postgresql.statio.sys_sequences.blks_hit 0
+#	disk.partitions.read 13 ?0 ?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9 ?10 ?11 ?12
+# to produce a summary of metrics, values and warnings
 #
 __filter()
 {
     $PCP_AWK_PROG '
-NF==0		{ next }
-/ value /	{ value++; next }
-/^'$1'/		{ metric++
-		  if ($0 !~ /:/) next
-		}
-		{ warn++ }
-END		{ if (warn) printf "%d warnings, ",warn
-		  printf "%d metrics and %d values\n",metric,value
-		}'
-}
-
-__strip_pcp_dir()
-{
-    if [ "X$PCP_DIR" = X ]
-    then
-	echo "$1"
-    else
-	echo "$1" | sed -e "s,^$PCP_DIR,,"
-    fi
+/^'$1'/         { metric++
+                  if ($2 < 0)
+		    warn++
+		  else
+		    value += $2
+		  next
+                }
+                { warn++ }
+END             { if (warn) printf "%d warnings, ",warn
+                  printf "%d metrics and %d values\n",metric,value
+                }'
 }
 
 _setup()
@@ -824,10 +833,10 @@ _setup()
     # some more configuration controls
     pmns_name=${pmns_name-$iam}
     pmda_name=pmda$iam
-    dso_name="${PCP_PMDAS_DIR}/${iam}/pmda_${iam}.${dso_suffix}"
-    dso_name=`__strip_pcp_dir "$dso_name"`
+    pmda_dso_name="${PCP_PMDAS_DIR}/${iam}/pmda_${iam}.${dso_suffix}"
+    dso_name="${dso_name-$pmda_dso_name}"
     dso_entry=${iam}_init
-    pmda_dir=`__strip_pcp_dir "${PCP_PMDAS_DIR}/${iam}"`
+    pmda_dir="${PCP_PMDAS_DIR}/${iam}"
 
     _check_userroot
     _check_directory
@@ -836,10 +845,10 @@ _setup()
     #
     if $perl_opt
     then
-	perl_name="${PCP_PMDAS_DIR}/${iam}/pmda${iam}.pl"
-	perl_name=`__strip_pcp_dir "$perl_name"`
-	perl_pmns="${PCP_PMDAS_DIR}/${iam}/pmns.perl"
-	perl_dom="${PCP_PMDAS_DIR}/${iam}/domain.h.perl"
+	perl_name="${pmda_dir}/pmda${iam}.perl"
+	[ -f "$perl_name" ] || perl_name="${pmda_dir}/pmda${iam}.pl"
+	perl_pmns="${pmda_dir}/pmns.perl"
+	perl_dom="${pmda_dir}/domain.h.perl"
 	perl -e 'use PCP::PMDA' 2>/dev/null
 	if test $? -eq 0
 	then
@@ -854,7 +863,29 @@ _setup()
 	fi
     fi
 
-    # Juggle pmns and domain.h in case perl pmda install was done here
+    # automatically generate files for the Python programmers too
+    #
+    if $python_opt
+    then
+	python_name="${pmda_dir}/pmda${iam}.python"
+	[ -f "$python_name" ] || python_name="${pmda_dir}/pmda${iam}.py"
+	python_pmns="${pmda_dir}/pmns.python"
+	python_dom="${pmda_dir}/domain.h.python"
+	python -c 'from pcp import pmda' 2>/dev/null
+	if test $? -eq 0
+	then
+	    eval PCP_PYTHON_DOMAIN=1 python "$python_name" > "$python_dom"
+	    eval PCP_PYTHON_PMNS=1 python "$python_name" > "$python_pmns"
+	elif $dso_opt || $daemon_opt
+	then
+	    :	# we have an alternative, so continue on
+	else
+	    echo 'Python pcp.pmda module is not installed, install it and try again'
+	    exit 1
+	fi
+    fi
+
+    # Juggle pmns and domain.h in case perl/python pmda install was done here
     # last time
     #
     for file in pmns domain.h
@@ -906,11 +937,11 @@ _install_views()
 #
 # before calling _install,
 # 1. set $iam
-# 2. set one/some/all of $dso_opt, $perl_opt, or $daemon_opt to true
+# 2. set one/some/all of $dso_opt, $perl_opt, $python_opt or $daemon_opt to true
 #    (optional, $daemon_opt is true by default)
 # 3. if $daemon_opt set one or both of $pipe_opt or $socket_opt true
 #    (optional, $pipe_opt is true by default)
-# 4. if $socket_opt and there is an default Internet socket, set
+# 4. if $socket_opt and there is a default Internet socket, set
 #    $socket_inet_def
 
 _install()
@@ -923,11 +954,11 @@ _install()
 
     if $do_pmda
     then
-	if $dso_opt || $perl_opt || $daemon_opt
+	if $dso_opt || $perl_opt || $python_opt || $daemon_opt
 	then
 	    :
 	else
-	    echo 'Botch: must set at least one of $dso_opt, $perl_opt or $daemon_opt to "true"'
+	    echo 'Botch: must set at least one of $dso_opt, $perl_opt, $python_opt or $daemon_opt to "true"'
 	    exit 1
 	fi
 	if $daemon_opt
@@ -941,7 +972,7 @@ _install()
 	    fi
 	fi
 
-	# Select a PMDA style (dso/perl/deamon), and for daemons the
+	# Select a PMDA style (dso/perl/python/deamon), and for daemons the
 	# IPC method for communication between PMCD and the PMDA.
 	#
 	pmda_options=''
@@ -964,6 +995,17 @@ _install()
 		pmda_options="perl"
 	    fi
 	fi
+	if $python_opt
+	then
+	    pmda_default_option="python"
+	    if test -n "$pmda_options"
+	    then
+		pmda_options="python or $pmda_options"
+		pmda_multiple_options=true
+	    else
+		pmda_options="python"
+	    fi
+	fi
 	if $daemon_opt
 	then
 	    pmda_default_option="daemon"
@@ -1000,6 +1042,15 @@ _install()
 		    else
 			break
 		    fi
+		elif [ "X$pmda_type" = Xpython ]
+		then
+		    python -c 'from pcp import pmda' 2>/dev/null
+		    if test $? -ne 0
+		    then
+			echo 'Python pcp pmda module is not installed, install it and try again'
+		    else
+			break
+		    fi
 		else
 		    echo "Must choose one of $pmda_options, please try again"
 		fi
@@ -1013,6 +1064,10 @@ _install()
 	then
 	    type="pipe	binary		perl $perl_name"
 	    args=""
+	elif [ "$pmda_type" = python ]
+	then
+	    type="pipe	binary		python $python_name"
+	    args=""
 	else
 	    type="dso	$dso_entry	$dso_name"
 	    args=""
@@ -1020,9 +1075,15 @@ _install()
 
 	# Install binaries
 	#
-	if [ -f Makefile -o -f makefile -o -f GNUmakefile ]
+	if [ "$pmda_type" = perl -o "$pmda_type" = python ]
 	then
-	    if [ ! -f "$PCP_MAKE_PROG" -o ! -f "$PCP_INC_DIR/pmda.h" ]
+	    :	# we can safely skip building binaries
+	elif [ -f Makefile -o -f makefile -o -f GNUmakefile ]
+	then
+	    # $PCP_MAKE_PROG may contain command line args ... executable
+	    # is first word
+	    #
+	    if [ ! -f "`echo $PCP_MAKE_PROG | sed -e 's/ .*//'`" -o ! -f "$PCP_INC_DIR/pmda.h" ]
 	    then
 		echo "$prog: Arrgh, PCP devel environment required to install this PMDA"
 		exit 1
@@ -1058,46 +1119,44 @@ _install()
 
     if $do_pmda
     then
-	if [ "X$pmda_type" = Xperl ]
+	if [ "X$pmda_type" = Xperl -o "X$pmda_type" = Xpython ]
 	then
 	    # Juggle pmns and domain.h ... save originals and
-	    # use *.perl ones created earlier
+	    # use *.{perl,python} ones created earlier
 	    for file in pmns domain.h
 	    do
-		if [ ! -f $file.perl ]
+		if [ ! -f "$file.$pmda_type" ]
 		then
-		    echo "Botch: $file.perl missing ... giving up"
+		    echo "Botch: $file.$pmda_type missing ... giving up"
 		    exit 1
 		fi
 		if [ -f $file ]
 		then
-		    if diff $file.perl $file >/dev/null
+		    if diff $file.$pmda_type $file >/dev/null
 		    then
 			:
 		    else
 			[ ! -f $file.save ] && mv $file $file.save
-			mv $file.perl $file
+			mv $file.$pmda_type $file
 		    fi
 		else
-		    mv $file.perl $file
+		    mv $file.$pmda_type $file
 		fi
 	    done
 	fi
     else
-	# Maybe PMNS only install, and only implementation may be a
-	# Perl one ... simpler juggling needed here
+	# Maybe PMNS only install, and only implementation may be
+	# Perl or Python ones ... simpler juggling needed here.
 	#
 	for file in pmns domain.h
 	do
-	    if [ ! -f $file -a -f $file.perl ]
-	    then
-		mv $file.perl $file
-	    fi
+	    [ ! -f $file -a -f $file.perl ] && mv $file.perl $file
+	    [ ! -f $file -a -f $file.python ] && mv $file.python $file
 	done
     fi
 
     $PCP_SHARE_DIR/lib/lockpmns $NAMESPACE
-    trap "$PCP_SHARE_DIR/lib/unlockpmns \$NAMESPACE; rm -f $tmp $tmp.*; exit" 0 1 2 3 15
+    trap "$PCP_SHARE_DIR/lib/unlockpmns \$NAMESPACE; rm -rf $pcptmp $pmdatmp; exit" 0 1 2 3 15
 
     echo "Updating the Performance Metrics Name Space (PMNS) ..."
 
@@ -1109,22 +1168,22 @@ _install()
 	if pminfo $__ns_opt $__n >/dev/null 2>&1
 	then
             cd $PMNSDIR
-	    if pmnsdel -n $PMNSROOT $__n >$tmp 2>&1
+	    if pmnsdel -n $PMNSROOT $__n >$tmp/base 2>&1
 	    then
 		pmsignal -a -s HUP pmcd >/dev/null 2>&1
 		# Make sure the PMNS timestamp will be different the next
 		# time the PMNS is updated (for Linux only 1 sec resolution)
 		sleep 2
 	    else
-		if grep 'Non-terminal "'"$__n"'" not found' $tmp >/dev/null
+		if grep 'Non-terminal "'"$__n"'" not found' $tmp/base >/dev/null
 		then
 		    :
-		elif grep 'Error: metricpath "'"$__n"'" not defined' $tmp >/dev/null
+		elif grep 'Error: metricpath "'"$__n"'" not defined' $tmp/base >/dev/null
 		then
 		    :
 		else
 		    echo "$prog: failed to delete \"$__n\" from the PMNS"
-		    cat $tmp
+		    cat $tmp/base
 		    exit 1
 		fi
 	    fi
@@ -1164,7 +1223,7 @@ _install()
         cd $__here
     done
 
-    trap "rm -f $tmp $tmp.*; exit" 0 1 2 3 15
+    trap "rm -rf $pcptmp $pmdatmp; exit" 0 1 2 3 15
     $PCP_SHARE_DIR/lib/unlockpmns $NAMESPACE
 
     _install_views pmchart
@@ -1178,6 +1237,14 @@ _install()
 	echo "Terminate PMDA if already installed ..."
 	__pmda_cull $iam $domain
 
+	# Rotate log files
+	#
+	if [ -f $PCP_LOG_DIR/pmcd/$iam.log ]
+	then
+	    rm -f $PCP_LOG_DIR/pmcd/$iam.log.prev
+	    mv -f $PCP_LOG_DIR/pmcd/$iam.log $PCP_LOG_DIR/pmcd/$iam.log.prev
+	fi
+
 	# Add PMDA to pmcd's configuration file
 	#
 	echo "Updating the PMCD control file, and notifying PMCD ..."
@@ -1192,11 +1259,11 @@ _install()
 	    for __n in $pmns_name
 	    do
 		$PCP_ECHO_PROG $PCP_ECHO_N "Check $__n metrics have appeared ... ""$PCP_ECHO_C"
-		pminfo -f $__ns_opt $__n | tee $tmp.verbose | __filter $__n
+		pmprobe -i $__ns_opt $__n | tee $tmp/verbose | __filter $__n
 		if $__verbose
 		then
 		    echo "pminfo output ..."
-		    cat $tmp.verbose
+		    cat $tmp/verbose
 		fi
 	    done
 	fi
@@ -1211,7 +1278,7 @@ _remove()
     #
 
     $PCP_SHARE_DIR/lib/lockpmns $NAMESPACE
-    trap "$PCP_SHARE_DIR/lib/unlockpmns \$NAMESPACE; rm -f $tmp $tmp.*; exit" 0 1 2 3 15
+    trap "$PCP_SHARE_DIR/lib/unlockpmns \$NAMESPACE; rm -rf $pcptmp $pmdatmp; exit" 0 1 2 3 15
 
     echo "Culling the Performance Metrics Name Space ..."
     cd $PMNSDIR
@@ -1219,22 +1286,22 @@ _remove()
     for __n in $pmns_name
     do
 	$PCP_ECHO_PROG $PCP_ECHO_N "$__n ... ""$PCP_ECHO_C"
-	if pmnsdel -n $PMNSROOT $__n >$tmp 2>&1
+	if pmnsdel -n $PMNSROOT $__n >$tmp/base 2>&1
 	then
 	    rm -f $PMNSDIR/$__n
 	    pmsignal -a -s HUP pmcd >/dev/null 2>&1
 	    sleep 2
 	    echo "done"
 	else
-	    if grep 'Non-terminal "'"$__n"'" not found' $tmp >/dev/null
+	    if grep 'Non-terminal "'"$__n"'" not found' $tmp/base >/dev/null
 	    then
 		echo "not found in Name Space, this is OK"
-	    elif grep 'Error: metricpath "'"$__n"'" not defined' $tmp >/dev/null
+	    elif grep 'Error: metricpath "'"$__n"'" not defined' $tmp/base >/dev/null
 	    then
 		echo "not found in Name Space, this is OK"
 	    else
 		echo "error"
-		cat $tmp
+		cat $tmp/base
 		exit
 	    fi
 	fi
@@ -1276,10 +1343,10 @@ _remove()
 	    for __n in $pmns_name
 	    do
 		$PCP_ECHO_PROG $PCP_ECHO_N "Check $__n metrics have gone away ... ""$PCP_ECHO_C"
-		if pminfo -n $NAMESPACE -f $__n >$tmp 2>&1
+		if pminfo -n $NAMESPACE -f $__n >$tmp/base 2>&1
 		then
 		    echo "Arrgh, something has gone wrong!"
-		    cat $tmp
+		    cat $tmp/base
 		else
 		    echo "OK"
 		fi
@@ -1289,7 +1356,7 @@ _remove()
 	echo "Skipping PMDA removal and PMCD re-configuration"
     fi
 
-    trap "rm -f $tmp $tmp.*; exit" 0 1 2 3 15
+    trap "rm -rf $pcptmp $pmdatmp; exit" 0 1 2 3 15
     $PCP_SHARE_DIR/lib/unlockpmns $NAMESPACE
 }
 
@@ -1297,8 +1364,13 @@ _check_userroot()
 {
     if [ "$uid" -ne 0 ]
     then
-	echo "Error: You must be root (uid 0) to update the PCP collector configuration."
-	exit 1
+	if [ -n "$PCP_DIR" ]
+	then
+	    : running in a non-default installation, do not need to be root
+	else
+	    echo "Error: You must be root (uid 0) to update the PCP collector configuration."
+	    exit 1
+	fi
     fi
 }
 
