@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2012-2013 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -12,8 +13,6 @@
  * for more details.
  */
 
-#include "pmapi.h"
-#include "impl.h"
 #include "pmcd.h"
 
 /* Check returned error from a client.
@@ -65,7 +64,7 @@ DoText(ClientInfo *cp, __pmPDU* pb)
     int		ident;
     int		type;
     AgentInfo	*ap;
-    char	*buffer;
+    char	*buffer = NULL;
 
     if ((sts = __pmDecodeTextReq(pb, &ident, &type)) < 0)
 	return sts;
@@ -78,28 +77,18 @@ DoText(ClientInfo *cp, __pmPDU* pb)
     if (ap->ipcType == AGENT_DSO) {
 	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 	    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
-	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_4)
-	    sts = ap->ipc.dso.dispatch.version.four.text(ident, type, &buffer,
-					  ap->ipc.dso.dispatch.version.four.ext);
-	else if (ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_2 ||
-	         ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_3)
-	    sts = ap->ipc.dso.dispatch.version.two.text(ident, type, &buffer,
-					  ap->ipc.dso.dispatch.version.two.ext);
-	else
-	    sts = ap->ipc.dso.dispatch.version.one.text(ident, type, &buffer);
-	if (sts < 0 &&
-	    ap->ipc.dso.dispatch.comm.pmapi_version == PMAPI_VERSION_1)
-		sts = XLATE_ERR_1TO2(sts);
+	sts = ap->ipc.dso.dispatch.version.any.text(ident, type, &buffer,
+					  ap->ipc.dso.dispatch.version.any.ext);
     }
     else {
 	if (ap->status.notReady)
 	    return PM_ERR_AGAIN;
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_TEXT_REQ, ident);
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_TEXT_REQ, ident);
 	sts = __pmSendTextReq(ap->inFd, cp - client, ident, type);
 	if (sts >= 0) {
-	    sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    int		pinpdu;
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_TEXT)
 		sts = __pmDecodeText(pb, &ident, &buffer);
@@ -115,6 +104,8 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 		pmcd_trace(TR_WRONG_PDU, ap->outFd, PDU_TEXT, sts);
 		sts = PM_ERR_IPC;	/* Wrong PDU type */
 	    }
+	    if (pinpdu > 0)
+		__pmUnpinPDUBuf(pb);
 	}
 	else
 	    pmcd_trace(TR_XMIT_ERR, ap->inFd, PDU_TEXT_REQ, sts);
@@ -125,18 +116,15 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 	CleanupAgent(ap, AT_COMM, ap->inFd);	
 
     if (sts >= 0) {
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_TEXT, ident);
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_TEXT, ident);
 	sts = __pmSendText(cp->fd, FROM_ANON, ident, buffer);
 	if (sts < 0 && ap->ipcType != AGENT_DSO) {
 	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_TEXT, sts);
 	    CleanupClient(cp, sts);
 	}
-	if ((ap->ipcType == AGENT_DSO &&
-	     ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_1) ||
-	    ap->ipcType != AGENT_DSO) {
-		/* daemons and old-style DSOs have a malloc'd buffer */
-		free(buffer);
+	if (ap->ipcType != AGENT_DSO) {
+	    /* daemon PMDAs have a malloc'd buffer */
+	    free(buffer);
 	}
     }
     return sts;
@@ -145,45 +133,35 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 int
 DoProfile(ClientInfo *cp, __pmPDU *pb)
 {
-    int		sts;
     __pmProfile	*newProf;
-    int		ctxnum;
+    int		ctxnum, sts, i;
 
     sts = __pmDecodeProfile(pb, &ctxnum, &newProf);
     if (sts >= 0) {
-	int i;
-
-	if (ctxnum < 0) {
-	    __pmNotifyErr(LOG_ERR, "DoProfile: bad ctxnum = %d\n", ctxnum);
-	    __pmFreeProfile(newProf);
-	    return PM_ERR_NOCONTEXT;
-	}
-
 	/* Allocate more profile pointers if required */
 	if (ctxnum >= cp->szProfile) {
-	    int		oldSize = cp->szProfile;
 	    __pmProfile	**newProfPtrs;
-	    unsigned	n;
+	    int		need, oldSize = cp->szProfile;
 
 	    if (ctxnum - cp->szProfile < 4)
 		cp->szProfile += 4;
 	    else
 		cp->szProfile = ctxnum + 1;
-	    n = cp->szProfile * (int)sizeof(__pmProfile *);
-	    if ((newProfPtrs = (__pmProfile **)malloc(n)) == NULL) {
+	    need = cp->szProfile * (int)sizeof(__pmProfile *);
+	    if ((newProfPtrs = (__pmProfile **)malloc(need)) == NULL) {
 		cp->szProfile = oldSize;
-		__pmNoMem("DoProfile.newProfPtrs", n, PM_RECOV_ERR);
+		__pmNoMem("DoProfile.newProfPtrs", need, PM_RECOV_ERR);
 		__pmFreeProfile(newProf);
 		return -oserror();
 	    }
 
 	    /* Copy any old pointers and zero the newly allocated ones */
-	    if ((n = oldSize * (int)sizeof(__pmProfile *))) {
-		memcpy(newProfPtrs, cp->profile, n);
+	    if ((need = oldSize * (int)sizeof(__pmProfile *))) {
+		memcpy(newProfPtrs, cp->profile, need);
 		free(cp->profile);	/* But not the __pmProfile ptrs! */
 	    }
-	    n = (cp->szProfile - oldSize) * (int)sizeof(__pmProfile *);
-	    memset(&newProfPtrs[oldSize], 0, n);
+	    need = (cp->szProfile - oldSize) * (int)sizeof(__pmProfile *);
+	    memset(&newProfPtrs[oldSize], 0, need);
 	    cp->profile = newProfPtrs;
 	}
 	else				/* cp->profile is big enough */
@@ -210,7 +188,7 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     int		sts, s;
     pmID	pmid;
     AgentInfo	*ap;
-    pmDesc	desc;
+    pmDesc	desc = {0};
     int		fdfail = -1;
 
     if ((sts = __pmDecodeDescReq(pb, &pmid)) < 0)
@@ -224,28 +202,18 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     if (ap->ipcType == AGENT_DSO) {
 	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 	    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
-	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_4)
-	    sts = ap->ipc.dso.dispatch.version.four.desc(pmid, &desc,
-					ap->ipc.dso.dispatch.version.four.ext);
-	else if (ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_2 ||
-	         ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_3)
-	    sts = ap->ipc.dso.dispatch.version.two.desc(pmid, &desc,
-					ap->ipc.dso.dispatch.version.two.ext);
-	else
-	    sts = ap->ipc.dso.dispatch.version.one.desc(pmid, &desc);
-	if (sts < 0 &&
-	    ap->ipc.dso.dispatch.comm.pmapi_version == PMAPI_VERSION_1)
-		sts = XLATE_ERR_1TO2(sts);
+	sts = ap->ipc.dso.dispatch.version.any.desc(pmid, &desc,
+					ap->ipc.dso.dispatch.version.any.ext);
     }
     else {
 	if (ap->status.notReady)
 	    return PM_ERR_AGAIN;
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_DESC_REQ, (int)pmid);
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_DESC_REQ, (int)pmid);
 	sts = __pmSendDescReq(ap->inFd, cp - client, pmid);
 	if (sts >= 0) {
-	    sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    int		pinpdu;
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_DESC)
 		sts = __pmDecodeDesc(pb, &desc);
@@ -262,6 +230,8 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
 		sts = PM_ERR_IPC;	/* Wrong PDU type */
 		fdfail = ap->outFd;
 	    }
+	    if (pinpdu > 0)
+		__pmUnpinPDUBuf(pb);
 	}
 	else {
 	    pmcd_trace(TR_XMIT_ERR, ap->inFd, PDU_DESC_REQ, sts);
@@ -270,8 +240,7 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     }
 
     if (sts >= 0) {
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_DESC, (int)desc.pmid);
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_DESC, (int)desc.pmid);
 	sts = __pmSendDesc(cp->fd, FROM_ANON, &desc);
 	if (sts < 0) {
 	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_DESC, sts);
@@ -290,16 +259,18 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
 int
 DoInstance(ClientInfo *cp, __pmPDU* pb)
 {
-    int			sts = 0, s;
+    int			sts, s;
     __pmTimeval		when;
     pmInDom		indom;
     int			inst;
     char		*name;
-    __pmInResult		*inresult = NULL;
+    __pmInResult	*inresult = NULL;
     AgentInfo		*ap;
     int			fdfail = -1;
 
-    __pmDecodeInstanceReq(pb, &when, &indom, &inst, &name);
+    sts = __pmDecodeInstanceReq(pb, &when, &indom, &inst, &name);
+    if (sts < 0)
+	return sts;
     if (when.tv_sec != 0 || when.tv_usec != 0) {
 	/*
 	 * we have no idea how to do anything but current, yet!
@@ -323,31 +294,21 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     if (ap->ipcType == AGENT_DSO) {
 	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 	    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
-	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_4)
-	    sts = ap->ipc.dso.dispatch.version.four.instance(indom, inst, name,
+	sts = ap->ipc.dso.dispatch.version.any.instance(indom, inst, name,
 					&inresult,
-					ap->ipc.dso.dispatch.version.four.ext);
-	else if (ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_2 ||
-	         ap->ipc.dso.dispatch.comm.pmda_interface == PMDA_INTERFACE_3)
-	    sts = ap->ipc.dso.dispatch.version.two.instance(indom, inst, name,
-					&inresult,
-					ap->ipc.dso.dispatch.version.two.ext);
-	else
-	    sts = ap->ipc.dso.dispatch.version.one.instance(indom, inst, name, 
-							     &inresult);
-	if (sts < 0 &&
-	    ap->ipc.dso.dispatch.comm.pmapi_version == PMAPI_VERSION_1)
-		sts = XLATE_ERR_1TO2(sts);
+					ap->ipc.dso.dispatch.version.any.ext);
     }
     else {
-	if (ap->status.notReady)
+	if (ap->status.notReady) {
+	    if (name != NULL) free(name);
 	    return PM_ERR_AGAIN;
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_INSTANCE_REQ, (int)indom);
+	}
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_INSTANCE_REQ, (int)indom);
 	sts = __pmSendInstanceReq(ap->inFd, cp - client, &when, indom, inst, name);
 	if (sts >= 0) {
-	    sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-	    if (sts > 0 && _pmcd_trace_mask)
+	    int		pinpdu;
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_INSTANCE)
 		sts = __pmDecodeInstance(pb, &inresult);
@@ -365,6 +326,8 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
 		sts = PM_ERR_IPC;	/* Wrong PDU type */
 		fdfail = ap->outFd;
 	    }
+	    if (pinpdu > 0)
+		__pmUnpinPDUBuf(pb);
 	}
 	else {
 	    pmcd_trace(TR_XMIT_ERR, ap->inFd, PDU_INSTANCE_REQ, sts);
@@ -374,15 +337,13 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     if (name != NULL) free(name);
 
     if (sts >= 0) {
-	if (_pmcd_trace_mask)
-	    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_INSTANCE, (int)(inresult->indom));
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_INSTANCE, (int)(inresult->indom));
 	sts = __pmSendInstance(cp->fd, FROM_ANON, inresult);
 	if (sts < 0) {
 	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_INSTANCE, sts);
 	    CleanupClient(cp, sts);
 	}
-	if (inresult != NULL)
-	    __pmFreeInResult(inresult);
+	__pmFreeInResult(inresult);
     }
     else
 	if (ap->ipcType != AGENT_DSO &&
@@ -440,12 +401,12 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 	    /* daemon PMDA ... ship request on */
 	    if (ap->status.notReady)
 		return PM_ERR_AGAIN;
-	    if (_pmcd_trace_mask)
-		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_IDS, 1);
+	    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_IDS, 1);
 	    sts = __pmSendIDList(ap->inFd, cp - client, 1, &idlist[0], 0);
 	    if (sts >= 0) {
-		sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		if (sts > 0 && _pmcd_trace_mask)
+		int		pinpdu;
+		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		if (sts > 0)
 		    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		if (sts == PDU_PMNS_NAMES) {
 		    sts = __pmDecodeNameList(pb, &numnames, &namelist, NULL);
@@ -459,6 +420,8 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 		    sts = PM_ERR_IPC;	/* Wrong PDU type */
 		    fdfail = ap->outFd;
 		}
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 	    }
 	    else {
 		/* __pmSendIDList failed */
@@ -471,8 +434,7 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 
     numnames = sts;
 
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
     if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, namelist, NULL)) < 0){
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
@@ -505,17 +467,17 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
     if ((sts = __pmDecodeNameList(pb, &numids, &namelist, NULL)) < 0)
 	goto done;
 
-    if ((idlist = (pmID*)malloc(sizeof(int)*numids)) == NULL) {
+    if ((idlist = (pmID *)calloc(numids, sizeof(int))) == NULL) {
         sts = -oserror();
 	goto done;
     }
 
     sts = pmLookupName(numids, namelist, idlist);
     for (i = 0; i < numids; i++) {
-	if (idlist[i] < 0) continue;
-	if (((__pmID_int *)&idlist[i])->domain == DYNAMIC_PMID) {
+	if (idlist[i] == PM_ID_NULL) continue;
+	if (pmid_domain(idlist[i]) == DYNAMIC_PMID && pmid_item(idlist[i]) == 0) {
 	    int		lsts;
-	    int		domain = ((__pmID_int *)&idlist[i])->cluster;
+	    int		domain = pmid_cluster(idlist[i]);
 	    /*
 	     * don't return <domain>.*.* ... all return paths from here
 	     * must either set a valid PMID in idlist[i] or indicate
@@ -548,12 +510,12 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		if (ap->status.notReady)
 		    lsts = PM_ERR_AGAIN;
 		else {
-		    if (_pmcd_trace_mask)
-			pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_NAMES, 1);
+		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_NAMES, 1);
 		    lsts = __pmSendNameList(ap->inFd, cp - client, 1, &namelist[i], NULL);
 		    if (lsts >= 0) {
-			lsts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-			if (lsts > 0 && _pmcd_trace_mask)
+			int		pinpdu;
+			pinpdu = lsts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+			if (lsts > 0)
 			    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 			if (lsts == PDU_PMNS_IDS) {
 			    int		xsts;
@@ -570,6 +532,8 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 			    lsts = PM_ERR_IPC;	/* Wrong PDU type */
 			    fdfail = ap->outFd;
 			}
+			if (pinpdu > 0)
+			    __pmUnpinPDUBuf(pb);
 		    }
 		    else {
 			/* __pmSendNameList failed */
@@ -600,8 +564,7 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 	  goto done;
     }
 
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numids);
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numids);
     if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, sts)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_IDS, sts);
 	CleanupClient(cp, sts);
@@ -635,8 +598,8 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 
     namelist[0] = name;
     sts = pmLookupName(1, namelist, idlist);
-    if (sts == 1 && ((__pmID_int *)&idlist[0])->domain == DYNAMIC_PMID) {
-	int		domain = ((__pmID_int *)&idlist[0])->cluster;
+    if (sts == 1 && pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
+	int		domain = pmid_cluster(idlist[0]);
 	AgentInfo	*ap = NULL;
 	if ((ap = FindDomainAgent(domain)) == NULL) {
 	    sts = PM_ERR_NOAGENT;
@@ -671,12 +634,12 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 	    if (ap->status.notReady)
 		sts = PM_ERR_AGAIN;
 	    else {
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_CHILD, 1);
+		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_CHILD, 1);
 		sts = __pmSendChildReq(ap->inFd, cp - client, name, subtype);
 		if (sts >= 0) {
-		    sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		    if (sts > 0 && _pmcd_trace_mask)
+		    int		pinpdu;
+		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == PDU_PMNS_NAMES) {
 			sts = __pmDecodeNameList(pb, &numnames,
@@ -698,6 +661,8 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 			sts = PM_ERR_IPC;	/* Wrong PDU type */
 			fdfail = ap->outFd;
 		    }
+		    if (pinpdu > 0)
+			__pmUnpinPDUBuf(pb);
 		}
 		else {
 		    /* __pmSendChildReq failed */
@@ -723,8 +688,7 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
     }
 
     numnames = sts;
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
     if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, offspring, statuslist)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
@@ -811,8 +775,8 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 	sts = pmLookupName(1, namelist, idlist);
 	if (sts < 1)
 	    continue;
-	if (((__pmID_int *)&idlist[0])->domain == DYNAMIC_PMID) {
-	    int		domain = ((__pmID_int *)&idlist[0])->cluster;
+	if (pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
+	    int		domain = pmid_cluster(idlist[0]);
 	    AgentInfo	*ap;
 	    if ((ap = FindDomainAgent(domain)) == NULL)
 		continue;
@@ -852,14 +816,14 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 		int		fdfail = -1;
 		if (ap->status.notReady)
 		    continue;
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_TRAVERSE, 1);
+		pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_TRAVERSE, 1);
 		sts = __pmSendTraversePMNSReq(ap->inFd, cp - client, namelist[0]);
 		if (sts >= 0) {
 		    int		numnames;
 		    __pmPDU	*pb;
-		    sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		    if (sts > 0 && _pmcd_trace_mask)
+		    int		pinpdu;
+		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == PDU_PMNS_NAMES) {
 			sts = __pmDecodeNameList(pb, &numnames,
@@ -895,6 +859,8 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 			sts = PM_ERR_IPC;	/* Wrong PDU type */
 			fdfail = ap->outFd;
 		    }
+		    if (pinpdu > 0)
+			__pmUnpinPDUBuf(pb);
 		}
 		else {
 		    /* __pmSendChildReq failed */
@@ -1019,10 +985,8 @@ check:
     if (travNL_num < 1)
 	goto done;
 
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, travNL_num);
-    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, 
-                  travNL_num, travNL, NULL)) < 0) {
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, travNL_num);
+    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, travNL_num, travNL, NULL)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
 	goto done;
@@ -1039,49 +1003,55 @@ done:
 int
 DoCreds(ClientInfo *cp, __pmPDU *pb)
 {
-    int			i, sts, credcount = 0;
-    int			version = UNKNOWN_VERSION;
-    int			sender = 0;
-    unsigned int	cookie = 0;
+    int			i, sts, flags = 0, version = 0, sender = 0, credcount = 0;
     __pmCred		*credlist = NULL;
+    __pmVersionCred	*vcp;
 
     if ((sts = __pmDecodeCreds(pb, &sender, &credcount, &credlist)) < 0)
 	return sts;
-
-    if (_pmcd_trace_mask)
-	pmcd_trace(TR_RECV_PDU, cp->fd, PDU_CREDS, credcount);
+    pmcd_trace(TR_RECV_PDU, cp->fd, PDU_CREDS, credcount);
 
     for (i = 0; i < credcount; i++) {
 	switch(credlist[i].c_type) {
 	    case CVERSION:
-		version = credlist[i].c_vala;
-		sts = __pmSetVersionIPC(cp->fd, version);
+		vcp = (__pmVersionCred *)&credlist[i];
+		flags = vcp->c_flags;
+		version = vcp->c_version;
 #ifdef PCP_DEBUG
 		if (pmDebug & DBG_TRACE_CONTEXT)
-		    fprintf(stderr, "pmcd: version cred (%u)\n", version);
+		    fprintf(stderr, "pmcd: version cred (%u) flags=%x\n", vcp->c_version, vcp->c_flags);
 #endif
 		break;
-	    case CAUTH:
-		cookie |= credlist[i].c_valc;
-		cookie <<= 8;
-		cookie |= credlist[i].c_valb;
-		cookie <<= 8;
-		cookie |= credlist[i].c_vala;
 
-		if ((sts = __pmMakeAuthCookie(cp->pduInfo.authorize, sender)) != cookie) {
-		    sts = PM_ERR_PERMISSION;
-		}
+	    default:
 #ifdef PCP_DEBUG
 		if (pmDebug & DBG_TRACE_CONTEXT)
-		    fprintf(stderr, "pmcd: my auth cred cookie=%u - client %s (pid=%d)\n",
-			cookie, (sts==PM_ERR_PERMISSION)?("denied"):("accepted"), sender);
+		    fprintf(stderr, "pmcd: Error: bogus cred type %d\n", credlist[i].c_type);
 #endif
+		sts = PM_ERR_IPC;
 		break;
 	}
     }
-
     if (credlist != NULL)
 	free(credlist);
 
+    if (sts >= 0 && version)
+	sts = __pmSetVersionIPC(cp->fd, version);
+    if (sts >= 0 && flags) {
+	/*
+	 * new client has arrived; may want encryption, authentication, etc
+	 * complete the handshake (depends on features requested), continue
+	 * on to check access is allowed for the authenticated persona, and
+	 * finally notify any interested PMDAs
+	 */
+	if ((sts = __pmSecureServerHandshake(cp->fd, flags, &cp->attrs)) < 0)
+	    return sts;
+    }
+    if ((sts = CheckAccountAccess(cp)) < 0)	/* host access done already */
+	return sts;
+    else if (sts > 0)	/* account authentication successful - inform PMDAs */
+	sts = AgentsAuthentication(cp - client);
+    /* else: no account-based authentication needed, so finish successfully */
+
     return sts;
 }
