@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2012-2013 Red Hat.
  * Copyright (c) 1995-2001,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -15,54 +16,33 @@
 #include "pmcd.h"
 #include "impl.h"
 #include <sys/stat.h>
+#include <assert.h>
 
-extern int  ParseInitAgents(char *);
-extern void ParseRestartAgents(char *);
-extern void PrintAgentInfo(FILE *);
-extern void ResetBadHosts(void);
-extern void StartDaemon(int, char **);
+#define SHUTDOWNWAIT	12	/* < PMDAs wait previously used in rc_pcp */
+#define MAXPENDING	5	/* maximum number of pending connections */
+#define FDNAMELEN	40	/* maximum length of a fd description */
+#define STRINGIFY(s)	#s
+#define TO_STRING(s)	STRINGIFY(s)
 
-#define SHUTDOWNWAIT 12 /* < PMDAs wait previously used in rc_pcp */
+#ifdef PCP_DEBUG
+static char	*FdToString(int);
+#endif
 
 int		AgentDied;		/* for updating mapdom[] */
 static int	timeToDie;		/* For SIGINT handling */
 static int	restart;		/* For SIGHUP restart */
+static int	maxReqPortFd;		/* Largest request port fd */
 static char	configFileName[MAXPATHLEN]; /* path to pmcd.conf */
 static char	*logfile = "pmcd.log";	/* log file name */
 static int	run_daemon = 1;		/* run as a daemon, see -f */
 int		_creds_timeout = 3;	/* Timeout for agents credential PDU */
 static char	*fatalfile = "/dev/tty";/* fatal messages at startup go here */
 static char	*pmnsfile = PM_NS_DEFAULT;
-
-/*
- * Interfaces we're willing to listen for clients on, from -i
- */
-static int		nintf = 0;
-static char		**intflist = NULL;
-
-/*
- * Ports we're willing to listen for clients on, from -p or $PMCD_PORT
- */
-static int		nport = 0;
-static int		*portlist = NULL;
-
-/*
- * For maintaining info about a request port that clients may connect to pmcd on
- */
-typedef struct {
-    int			fd;		/* File descriptor */
-    int			port;		/* Listening port */
-    char*		ipSpec;		/* String used to specify IP addr (or NULL) */
-    __uint32_t		ipAddr;		/* IP address (network byte order) */
-} ReqPortInfo;
-
-/*
- * A list of the ports that pmcd is listening for client connections on
- */
-static unsigned		nReqPorts = 0;	/* number of ports */
-static unsigned		szReqPorts = 0;	/* capacity of ports array */
-static ReqPortInfo	*reqPorts = NULL;	/* ports array */
-int			maxReqPortFd = -1;	/* highest request port file descriptor */
+static char	*username;
+static char	*certdb;		/* certificate database path (NSS) */
+static char	*dbpassfile;		/* certificate database password file */
+static int	dupok;			/* set to 1 for -N pmnsfile */
+static char	sockpath[MAXPATHLEN];	/* local unix domain socket path */
 
 #ifdef HAVE_SA_SIGINFO
 static pid_t	killer_pid;
@@ -75,6 +55,7 @@ DontStart(void)
 {
     FILE	*tty;
     FILE	*log;
+
     __pmNotifyErr(LOG_ERR, "pmcd not started due to errors!\n");
 
     if ((tty = fopen(fatalfile, "w")) != NULL) {
@@ -91,6 +72,14 @@ DontStart(void)
 	    fprintf(tty, "Log file \"%s\" has vanished!\n", logfile);
 	fclose(tty);
     }
+    /*
+     * We are often called after the request ports have been opened. If we don't
+     * explicitely close them, then the unix domain socket file (if any) will be
+     * left in the file system, causing "address already in use" the next time
+     * pmcd starts.
+     */
+    __pmServerCloseRequestPorts();
+
     exit(1);
 }
 
@@ -106,76 +95,14 @@ CreatePIDfile(void)
 	fprintf(stderr, "Error: Cant open PID file %s\n", pidpath);
 	return -1;
     }
-    fprintf(pidfile, "%d", getpid());
+    fprintf(pidfile, "%" FMT_PID, getpid());
     fflush(pidfile);
     fclose(pidfile);
     return 0;
 }
 
-/* Increase the capacity of the reqPorts array (maintain the contents) */
-
 static void
-GrowReqPorts(void)
-{
-    size_t need;
-    szReqPorts += 4;
-    need = szReqPorts * sizeof(ReqPortInfo);
-    reqPorts = (ReqPortInfo*)realloc(reqPorts, need);
-    if (reqPorts == NULL) {
-	__pmNoMem("pmcd: can't grow request port array", need, PM_FATAL_ERR);
-    }
-}
-
-/* Add a request port to the reqPorts array */
-
-static int
-AddRequestPort(char *ipSpec, int port)
-{
-    ReqPortInfo		*rp;
-    u_long		addr = 0;
-
-    if (ipSpec) {
-	int		i;
-	char		*sp = ipSpec;
-	char		*endp;
-	unsigned long	part;
-
-	for (i = 0; i < 4; i++) {
-	    part = strtoul(sp, &endp, 10);
-	    if (*endp != ((i < 3) ? '.' : '\0'))
-		return 0;
-	    if (part > 255)
-		return 0;
-	    addr |= part << (8 * (3 - i));
-	    if (i < 3)
-		sp = endp + 1;
-	}
-    }
-    else {
-	ipSpec = "INADDR_ANY";
-	addr = INADDR_ANY;
-    }
-
-    if (nReqPorts == szReqPorts)
-	GrowReqPorts();
-    rp = &reqPorts[nReqPorts];
-    rp->fd = -1;
-    rp->ipSpec = strdup(ipSpec);
-    rp->ipAddr = (__uint32_t)htonl(addr);
-    rp->port = port;
-    nReqPorts++;
-
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_APPL0)
-	fprintf(stderr, "AddRequestPort: %s -> 0x%08lx -> 0x%08x port %d\n",
-		rp->ipSpec, addr, rp->ipAddr, rp->port);
-#endif
-
-    return 1;	/* success */
-}
-
-static void
-ParseOptions(int argc, char *argv[])
+ParseOptions(int argc, char *argv[], int *nports)
 {
     int		c;
     int		sts;
@@ -183,12 +110,9 @@ ParseOptions(int argc, char *argv[])
     int		usage = 0;
     char	*endptr;
     int		val;
-    int		port;
-    char	*p;
 
-    __pmSetProgname(argv[0]);
-
-    strcpy(configFileName, pmGetConfig("PCP_PMCDCONF_PATH"));
+    endptr = pmGetConfig("PCP_PMCDCONF_PATH");
+    strncpy(configFileName, endptr, sizeof(configFileName)-1);
 
 #ifdef HAVE_GETOPT_NEEDS_POSIXLY_CORRECT
     /*
@@ -199,9 +123,21 @@ ParseOptions(int argc, char *argv[])
     putenv("POSIXLY_CORRECT=");
 #endif
 
-    while ((c = getopt(argc, argv, "D:fi:l:L:n:p:q:t:T:x:?")) != EOF)
+    while ((c = getopt(argc, argv, "Ac:C:D:fH:i:l:L:N:n:p:P:q:s:St:T:U:x:?")) != EOF)
 	switch (c) {
 
+	    case 'A':	/* disable pmcd service advertising */
+		__pmServerClearFeature(PM_SERVER_FEATURE_DISCOVERY);
+		break;
+
+	    case 'c':	/* configuration file */
+		strncpy(configFileName, optarg, sizeof(configFileName)-1);
+		break;
+
+	    case 'C':	/* path to NSS certificate database */
+		certdb = optarg;
+		break;
+
 	    case 'D':	/* debug flag */
 		sts = __pmParseDebug(optarg);
 		if (sts < 0) {
@@ -218,12 +154,13 @@ ParseOptions(int argc, char *argv[])
 		break;
 
 	    case 'i':
-		/* one (of possibly several) IP addresses for client requests */
-		nintf++;
-		if ((intflist = (char **)realloc(intflist, nintf * sizeof(char *))) == NULL) {
-		    __pmNoMem("pmcd: can't grow interface list", nintf * sizeof(char *), PM_FATAL_ERR);
-		}
-		intflist[nintf-1] = optarg;
+		/* one (of possibly several) interfaces for client requests */
+		__pmServerAddInterface(optarg);
+		break;
+
+	    case 'H':
+		/* use the given name as the pmcd.hostname for this host */
+		_pmcd_hostname = optarg;
 		break;
 
 	    case 'l':
@@ -234,44 +171,35 @@ ParseOptions(int argc, char *argv[])
 	    case 'L': /* Maximum size for PDUs from clients */
 		val = (int)strtol (optarg, NULL, 0);
 		if ( val <= 0 ) {
-		    fputs ("pmcd: -L require a posivite value\n", stderr);
+		    fputs("pmcd: -L requires a positive value\n", stderr);
 		    errflag++;
 		} else {
 		    __pmSetPDUCeiling (val);
 		}
 		break;
 
+	    case 'N':
+		dupok = 1;
+		/*FALLTHROUGH*/
 	    case 'n':
 	    	/* name space file name */
 		pmnsfile = optarg;
 		break;
 
 	    case 'p':
-		/*
-		 * one (of possibly several) ports for client requests
-		 * ... accept a comma separated list of ports here
-		 */
-		p = optarg;
-		for ( ; ; ) {
-		    port = (int)strtol(p, &endptr, 0);
-		    if ((*endptr != '\0' && *endptr != ',') || port < 0) {
-			fprintf(stderr,
-				"pmcd: -p requires a positive numeric argument (%s)\n", optarg);
-			errflag++;
-			break;
-		    }
-		    else {
-			nport++;
-			if ((portlist = (int *)realloc(portlist, nport * sizeof(int))) == NULL) {
-			    __pmNoMem("pmcd: can't grow port list", nport * sizeof(int), PM_FATAL_ERR);
-			}
-			portlist[nport-1] = port;
-		    }
-		    if (*endptr == '\0')
-			break;
-		    p = &endptr[1];
+		if (__pmServerAddPorts(optarg) < 0) {
+		    fprintf(stderr,
+			"pmcd: -p requires a positive numeric argument (%s)\n",
+			optarg);
+		    errflag++;
+		} else {
+		    *nports += 1;
 		}
 		break;
+		    
+	    case 'P':	/* password file for certificate database access */
+		dbpassfile = optarg;
+		break;
 
 	    case 'q':
 		val = (int)strtol(optarg, &endptr, 10);
@@ -284,6 +212,14 @@ ParseOptions(int argc, char *argv[])
 		    _creds_timeout = val;
 		break;
 
+	    case 's':	/* path to local unix domain socket */
+		snprintf(sockpath, sizeof(sockpath), "%s", optarg);
+		break;
+
+	    case 'S':	/* only allow authenticated clients */
+		__pmServerSetFeature(PM_SERVER_FEATURE_CREDS_REQD);
+		break;
+
 	    case 't':
 		val = (int)strtol(optarg, &endptr, 10);
 		if (*endptr != '\0' || val < 0.0) {
@@ -306,6 +242,10 @@ ParseOptions(int argc, char *argv[])
 		    _pmcd_trace_mask = val;
 		break;
 
+	    case 'U':
+		username = optarg;
+		break;
+
 	    case 'x':
 		fatalfile = optarg;
 		break;
@@ -323,15 +263,23 @@ ParseOptions(int argc, char *argv[])
 	fprintf(stderr,
 "Usage: %s [options]\n\n"
 "Options:\n"
+"  -A              disable service advertisement\n" 
+"  -c config       path to configuration file\n"
+"  -C dirname      path to NSS certificate database\n"
 "  -f              run in the foreground\n" 
+"  -H hostname     set the hostname to be used for pmcd.hostname metric\n"
 "  -i ipaddress    accept connections on this IP address\n"
 "  -l logfile      redirect diagnostics and trace output\n"
 "  -L bytes        maximum size for PDUs from clients [default 65536]\n"
 "  -n pmnsfile     use an alternative PMNS\n"
+"  -N pmnsfile     use an alternative PMNS (duplicate PMIDs are allowed)\n"
 "  -p port         accept connections on this port\n"
+"  -P passfile     password file for certificate database access\n"
 "  -q timeout      PMDA initial negotiation timeout (seconds) [default 3]\n"
+"  -s sockname     Unix domain socket file [default $PCP_RUN_DIR/pmcd.socket]\n"
 "  -T traceflag    Event trace control\n"
 "  -t timeout      PMDA response timeout (seconds) [default 5]\n"
+"  -U username     in daemon mode, run as named user [default pcp]\n"
 "  -x file         fatal messages at startup sent to file [default /dev/tty]\n",
 			pmProgname);
 	if (usage)
@@ -341,103 +289,12 @@ ParseOptions(int argc, char *argv[])
     }
 }
 
-/* Create socket for incoming connections and bind to it an address for
- * clients to use.  Returns -1 on failure.
- * ipAddr is the IP address that the port is advertised for (in network byte
- * order, see htonl(3N)).  To allow connections to all this host's IP addresses
- * from clients use ipAddr = htonl(INADDR_ANY).
- */
-static int
-OpenRequestSocket(int port, __uint32_t ipAddr)
-{
-    int			fd;
-    int			one, sts;
-    struct sockaddr_in	myAddr;
-
-    fd = __pmCreateSocket();
-    if (fd < 0) {
-	__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, 0x%x) socket: %s\n",
-		port, ipAddr, netstrerror());
-	return -1;
-    }
-    if (fd > maxClientFd)
-	maxClientFd = fd;
-    FD_SET(fd, &clientFds);
-
-    /* Ignore dead client connections */
-    one = 1;
-#ifndef IS_MINGW
-    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&one,
-		(mysocklen_t)sizeof(one)) < 0) {
-	__pmNotifyErr(LOG_ERR,
-		"OpenRequestSocket(%d, 0x%x) setsockopt(SO_REUSEADDR): %s\n",
-		port, ipAddr, netstrerror());
-	goto fail;
-    }
-#else
-    if (setsockopt(fd, SOL_SOCKET, SO_EXCLUSIVEADDRUSE, (char *)&one,
-		(mysocklen_t)sizeof(one)) < 0) {
-	__pmNotifyErr(LOG_ERR,
-		"OpenRequestSocket(%d,0x%x) setsockopt(EXCLUSIVEADDRUSE): %s\n",
-		port, ipAddr, netstrerror());
-	goto fail;
-    }
-#endif
-
-    /* and keep alive please - pv 916354 bad networks eat fds */
-    if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&one,
-		(mysocklen_t)sizeof(one)) < 0) {
-	__pmNotifyErr(LOG_ERR,
-		"OpenRequestSocket(%d, 0x%x) setsockopt(SO_KEEPALIVE): %s\n",
-		port, ipAddr, netstrerror());
-	goto fail;
-    }
-
-    memset(&myAddr, 0, sizeof(myAddr));
-    myAddr.sin_family = AF_INET;
-    myAddr.sin_addr.s_addr = ipAddr;
-    myAddr.sin_port = htons(port);
-    sts = bind(fd, (struct sockaddr*)&myAddr, sizeof(myAddr));
-    if (sts < 0) {
-	sts = neterror();
-	__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, 0x%x) bind: %s\n",
-		port, ipAddr, netstrerror());
-	if (sts == EADDRINUSE)
-	    __pmNotifyErr(LOG_ERR, "pmcd may already be running\n");
-	goto fail;
-    }
-
-    sts = listen(fd, 5);	/* Max. of 5 pending connection requests */
-    if (sts == -1) {
-	__pmNotifyErr(LOG_ERR, "OpenRequestSocket(%d, 0x%x) listen: %s\n",
-		port, ipAddr, netstrerror());
-	goto fail;
-    }
-    return fd;
-
-fail:
-    __pmCloseSocket(fd);
-    return -1;
-}
-
-extern int DoFetch(ClientInfo *, __pmPDU *);
-extern int DoProfile(ClientInfo *, __pmPDU *);
-extern int DoDesc(ClientInfo *, __pmPDU *);
-extern int DoInstance(ClientInfo *, __pmPDU *);
-extern int DoText(ClientInfo *, __pmPDU *);
-extern int DoStore(ClientInfo *, __pmPDU *);
-extern int DoCreds(ClientInfo *, __pmPDU *);
-extern int DoPMNSIDs(ClientInfo *, __pmPDU *);
-extern int DoPMNSNames(ClientInfo *, __pmPDU *);
-extern int DoPMNSChild(ClientInfo *, __pmPDU *);
-extern int DoPMNSTraverse(ClientInfo *, __pmPDU *);
-
-/* Determine which clients (if any) have sent data to the server and handle it
+/*
+ * Determine which clients (if any) have sent data to the server and handle it
  * as required.
  */
-
 void
-HandleClientInput(fd_set *fdsPtr)
+HandleClientInput(__pmFdSet *fdsPtr)
 {
     int		sts;
     int		i;
@@ -446,32 +303,39 @@ HandleClientInput(fd_set *fdsPtr)
     ClientInfo	*cp;
 
     for (i = 0; i < nClients; i++) {
-	if (!client[i].status.connected || !FD_ISSET(client[i].fd, fdsPtr))
+	int		pinpdu;
+	if (!client[i].status.connected || !__pmFD_ISSET(client[i].fd, fdsPtr))
 	    continue;
 
 	cp = &client[i];
 	this_client_id = i;
 
-	sts = __pmGetPDU(cp->fd, LIMIT_SIZE, _pmcd_timeout, &pb);
-	if (sts > 0 && _pmcd_trace_mask)
+	pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, _pmcd_timeout, &pb);
+	if (sts > 0) {
 	    pmcd_trace(TR_RECV_PDU, cp->fd, sts, (int)((__psint_t)pb & 0xffffffff));
-	if (sts <= 0) {
+	} else {
 	    CleanupClient(cp, sts);
 	    continue;
 	}
 
 	php = (__pmPDUHdr *)pb;
-	if (__pmVersionIPC(cp->fd) == UNKNOWN_VERSION && php->type != PDU_CREDS)
-	    __pmSetVersionIPC(cp->fd, PDU_VERSION1);
+	if (__pmVersionIPC(cp->fd) == UNKNOWN_VERSION && php->type != PDU_CREDS) {
+	    /* old V1 client protocol, no longer supported */
+	    sts = PM_ERR_IPC;
+	    CleanupClient(cp, sts);
+	    __pmUnpinPDUBuf(pb);
+	    continue;
+	}
 
 #ifdef PCP_DEBUG
-	    if (pmDebug & DBG_TRACE_APPL0)
-		ShowClients(stderr);
+	if (pmDebug & DBG_TRACE_APPL0)
+	    ShowClients(stderr);
 #endif
 
 	switch (php->type) {
 	    case PDU_PROFILE:
-		sts = DoProfile(cp, pb);
+		sts = (cp->denyOps & PMCD_OP_FETCH) ?
+		      PM_ERR_PERMISSION : DoProfile(cp, pb);
 		break;
 
 	    case PDU_FETCH:
@@ -520,14 +384,7 @@ HandleClientInput(fd_set *fdsPtr)
 		break;
 
 	    case PDU_CREDS:
-		if ((sts = __pmVersionIPC(cp->fd)) < 0)
-		    break;
-		if (sts == PDU_VERSION1) {
-		    __pmNotifyErr(LOG_ERR, "pmcd: protocol version error on fd=%d\n", cp->fd);
-		    sts = PM_ERR_V1(PM_ERR_IPC);
-		}
-		else
-		    sts = DoCreds(cp, pb);
+		sts = DoCreds(cp, pb);
 		break;
 
 	    default:
@@ -542,14 +399,15 @@ HandleClientInput(fd_set *fdsPtr)
 #endif
 	    /* Make sure client still alive before sending. */
 	    if (cp->status.connected) {
-		if (_pmcd_trace_mask)
-		    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
+		pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
 		sts = __pmSendError(cp->fd, FROM_ANON, sts);
 		if (sts < 0)
 		    __pmNotifyErr(LOG_ERR, "HandleClientInput: "
 			"error sending Error PDU to client[%d] %s\n", i, pmErrStr(sts));
 	    }
 	}
+	if (pinpdu > 0)
+	    __pmUnpinPDUBuf(pb);
     }
 }
 
@@ -559,7 +417,6 @@ void
 Shutdown(void)
 {
     int	i;
-    int	fd;
 
     for (i = 0; i < nAgents; i++) {
 	AgentInfo *ap = &agent[i];
@@ -597,9 +454,8 @@ Shutdown(void)
     for (i = 0; i < nClients; i++)
 	if (client[i].status.connected)
 	    __pmCloseSocket(client[i].fd);
-    for (i = 0; i < nReqPorts; i++)
-	if ((fd = reqPorts[i].fd) != -1)
-	    __pmCloseSocket(fd);
+    __pmServerCloseRequestPorts();
+    __pmSecureServerShutdown();
     __pmNotifyErr(LOG_INFO, "pmcd Shutdown\n");
     fflush(stderr);
 }
@@ -612,11 +468,11 @@ SignalShutdown(void)
     char	buf[256];
 #endif
     if (killer_pid != 0) {
-	__pmNotifyErr(LOG_INFO, "pmcd caught %s from pid=%d uid=%d\n",
+	__pmNotifyErr(LOG_INFO, "pmcd caught %s from pid=%" FMT_PID " uid=%d\n",
 	    killer_sig == SIGINT ? "SIGINT" : "SIGTERM", killer_pid, killer_uid);
 #if DESPERATE
 	__pmNotifyErr(LOG_INFO, "Try to find process in ps output ...\n");
-	sprintf(buf, "sh -c \". \\$PCP_DIR/etc/pcp.env; ( \\$PCP_PS_PROG \\$PCP_PS_ALL_FLAGS | \\$PCP_AWK_PROG 'NR==1 {print} \\$2==%d {print}' )\"", killer_pid);
+	sprintf(buf, "sh -c \". \\$PCP_DIR/etc/pcp.env; ( \\$PCP_PS_PROG \\$PCP_PS_ALL_FLAGS | \\$PCP_AWK_PROG 'NR==1 {print} \\$2==%" FMT_PID " {print}' )\"", killer_pid);
 	system(buf);
 #endif
     }
@@ -662,7 +518,11 @@ SignalReloadPMNS(void)
 	__pmNotifyErr(LOG_INFO, "Reloading PMNS \"%s\"",
 	   (pmnsfile==PM_NS_DEFAULT)?"DEFAULT":pmnsfile);
 	pmUnloadNameSpace();
-	if ((sts = pmLoadNameSpace(pmnsfile)) < 0) {
+	if (dupok)
+	    sts = pmLoadASCIINameSpace(pmnsfile, 1);
+	else
+	    sts = pmLoadNameSpace(pmnsfile);
+	if (sts < 0) {
 	    __pmNotifyErr(LOG_ERR, "PMNS \"%s\" load failed: %s",
 		(pmnsfile == PM_NS_DEFAULT) ? "DEFAULT" : pmnsfile,
 		pmErrStr(sts));
@@ -678,7 +538,7 @@ SignalReloadPMNS(void)
  * to handle PDUs.
  */
 static int
-HandleReadyAgents(fd_set *readyFds)
+HandleReadyAgents(__pmFdSet *readyFds)
 {
     int		i, s, sts;
     int		fd;
@@ -691,12 +551,13 @@ HandleReadyAgents(fd_set *readyFds)
 	ap = &agent[i];
 	if (ap->status.notReady) {
 	    fd = ap->outFd;
-	    if (FD_ISSET(fd, readyFds)) {
+	    if (__pmFD_ISSET(fd, readyFds)) {
+		int		pinpdu;
 
 		/* Expect an error PDU containing PM_ERR_PMDAREADY */
 		reason = AT_COMM;	/* most errors are protocol failures */
-		sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
-		if (sts > 0 && _pmcd_trace_mask)
+		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		if (sts > 0)
 		    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		if (sts == PDU_ERROR) {
 		    s = __pmDecodeError(pb, &sts);
@@ -730,40 +591,95 @@ HandleReadyAgents(fd_set *readyFds)
 			pmcd_trace(TR_RECV_ERR, ap->outFd, PDU_RESULT, sts);
 		    else
 			pmcd_trace(TR_WRONG_PDU, ap->outFd, PDU_ERROR, sts);
-		    sts = PM_ERR_IPC; /* Wrong PDU type */
+ 		    sts = PM_ERR_IPC; /* Wrong PDU type */
 		}
+		if (pinpdu > 0)
+		    __pmUnpinPDUBuf(pb);
 
 		if (ap->ipcType != AGENT_DSO && sts <= 0)
 		    CleanupAgent(ap, reason, fd);
-
 	    }
 	}
     }
-
     return ready;
 }
 
+static void
+CheckNewClient(__pmFdSet * fdset, int rfd, int family)
+{
+    int		s, sts, accepted = 1;
+    __uint32_t	challenge;
+    ClientInfo	*cp;
+
+    if (__pmFD_ISSET(rfd, fdset)) {
+	if ((cp = AcceptNewClient(rfd)) == NULL)
+	    return;	/* Accept failed and no client added */
+
+	sts = __pmAccAddClient(cp->addr, &cp->denyOps);
+#if defined(HAVE_STRUCT_SOCKADDR_UN)
+	if (sts >= 0 && family == AF_UNIX) {
+	    if ((sts = __pmServerSetLocalCreds(cp->fd, &cp->attrs)) < 0) {
+		__pmNotifyErr(LOG_ERR,
+			"ClientLoop: error extracting local credentials: %s",
+			pmErrStr(sts));
+	    }
+	}
+#endif
+	if (sts >= 0) {
+	    memset(&cp->pduInfo, 0, sizeof(cp->pduInfo));
+	    cp->pduInfo.version = PDU_VERSION;
+	    cp->pduInfo.licensed = 1;
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_SECURE))
+		cp->pduInfo.features |= PDU_FLAG_SECURE;
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_COMPRESS))
+		cp->pduInfo.features |= PDU_FLAG_COMPRESS;
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_AUTH))     /* optionally */
+		cp->pduInfo.features |= PDU_FLAG_AUTH;
+	    if (__pmServerHasFeature(PM_SERVER_FEATURE_CREDS_REQD)) /* required */
+		cp->pduInfo.features |= PDU_FLAG_CREDS_REQD;
+	    challenge = *(__uint32_t *)(&cp->pduInfo);
+	    sts = 0;
+	}
+	else {
+	    challenge = 0;
+	    accepted = 0;
+	}
+
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
+
+	/* reset (no meaning, use fd table to version) */
+	cp->pduInfo.version = UNKNOWN_VERSION;
+
+	s = __pmSendXtendError(cp->fd, FROM_ANON, sts, htonl(challenge));
+	if (s < 0) {
+	    __pmNotifyErr(LOG_ERR,
+		"ClientLoop: error sending Conn ACK PDU to new client %s\n",
+		pmErrStr(s));
+	    if (sts >= 0)
+	        /*
+		 * prefer earlier failure status if any, else
+		 * use the one from __pmSendXtendError()
+		 */
+	        sts = s;
+	    accepted = 0;
+	}
+	if (!accepted)
+	    CleanupClient(cp, sts);
+    }
+}
+
 /* Loop, synchronously processing requests from clients. */
 
 static void
 ClientLoop(void)
 {
-    int		i, sts;
-    int		challenge;
+    int		i, fd, sts;
     int		maxFd;
     int		checkAgents;
     int		reload_ns = 0;
-    fd_set	readableFds;
-    int		CheckClientAccess(ClientInfo *);
-    ClientInfo	*cp;
-    __pmPDUInfo	xchallenge;
+    __pmFdSet	readableFds;
 
     for (;;) {
-	if (_pmcd_done) {
-	    __pmNotifyErr(LOG_INFO,
-		"pmcd terminated via pmcd pmda and pmcd.control.debug");
-	    break;
-	}
 
 	/* Figure out which file descriptors to wait for input on.  Keep
 	 * track of the highest numbered descriptor for the select call.
@@ -777,11 +693,10 @@ ClientLoop(void)
 	checkAgents = 0;
 	for (i = 0; i < nAgents; i++) {
 	    AgentInfo	*ap = &agent[i];
-	    int		fd;
 
 	    if (ap->status.notReady) {
 		fd = ap->outFd;
-		FD_SET(fd, &readableFds);
+		__pmFD_SET(fd, &readableFds);
 		if (fd > maxFd)
 		    maxFd = fd + 1;
 		checkAgents = 1;
@@ -794,76 +709,15 @@ ClientLoop(void)
 	    }
 	}
 
-	sts = select(maxFd, &readableFds, NULL, NULL, NULL);
-
+	sts = __pmSelectRead(maxFd, &readableFds, NULL);
 	if (sts > 0) {
 #ifdef PCP_DEBUG
 	    if (pmDebug & DBG_TRACE_APPL0)
 		for (i = 0; i <= maxClientFd; i++)
-		    if (FD_ISSET(i, &readableFds))
+		    if (__pmFD_ISSET(i, &readableFds))
 			fprintf(stderr, "DATA: from %s (fd %d)\n", FdToString(i), i);
 #endif
-	    /* Accept any new client connections */
-	    for (i = 0; i < nReqPorts; i++) {
-		int rfd = reqPorts[i].fd;
-		if (rfd == -1)
-		    continue;
-		if (FD_ISSET(rfd, &readableFds)) {
-		    int	sts, s;
-		    int	accepted = 1;
-
-		    cp = AcceptNewClient(rfd);
-
-		    /* Accept failed and no client added */
-		    if (cp == NULL)
-		    	continue;
-
-		    sts = __pmAccAddClient(&cp->addr.sin_addr, &cp->denyOps);
-		    if (sts >= 0) {
-			cp->pduInfo.zero = 0;
-			cp->pduInfo.version = PDU_VERSION;
-			cp->pduInfo.licensed = 1;
-			cp->pduInfo.authorize = 0;
-			challenge = *(int*)(&cp->pduInfo);
-			sts = 0;
-			/* reset (no meaning, use fd table to version) */
-			cp->pduInfo.version = UNKNOWN_VERSION;
-		    }
-		    else {
-			/* __pmAccAddClient failed, this is grim! */
-			challenge = 0;
-			accepted = 0;
-		    }
-
-		    if (_pmcd_trace_mask)
-			pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_ERROR, sts);
-		    /*
-		     * Note. in the event of an error being detected,
-		     *       sts is converted from 2.0 to 1.x in
-		     *       __pmSendXtendError() ... special case code
-		     *	     on the client side knows about this for 2.0
-		     *       clients.
-		     */
-		    xchallenge = *(__pmPDUInfo *)&challenge;
-		    xchallenge = __htonpmPDUInfo(xchallenge);
-		    s = __pmSendXtendError(cp->fd, FROM_ANON, sts, *(unsigned int *)&xchallenge);
-		    if (s < 0) {
-			__pmNotifyErr(LOG_ERR,
-				"ClientLoop: error sending Conn ACK PDU to new client %s\n",
-				pmErrStr(s));
-			if (sts >= 0)
-			    /*
-			     * prefer earlier failure status if any, else
-			     * use the one from __pmSendXtendError()
-			     */
-			    sts = s;
-			accepted = 0;
-		    }
-		    if (!accepted)
-			CleanupClient(cp, sts);
-		}
-	    }
-
+	    __pmServerAddNewClients(&readableFds, CheckNewClient);
 	    if (checkAgents)
 		reload_ns = HandleReadyAgents(&readableFds);
 	    HandleClientInput(&readableFds);
@@ -961,15 +815,55 @@ do_traceback(FILE *f)
 #endif
     return;
 }
-#endif
+#endif /* HAVE_TRACE_BACK_STACK */
+
+#if HAVE_BACKTRACE
+#include <execinfo.h>
+
+#define MAX_DEPTH 30
+
+static void
+do_traceback(FILE *f)
+{
+    int		nframe;
+    void	*buf[MAX_DEPTH];
+    char	**symbols;
+    int		i;
+
+    nframe = backtrace(buf, MAX_DEPTH);
+    if (nframe < 1) {
+	fprintf(f, "backtrace -> %d frames?\n", nframe);
+	return;
+    }
+    symbols = backtrace_symbols(buf, nframe);
+    if (symbols == NULL) {
+	fprintf(f, "backtrace_symbols failed!\n");
+	return;
+
+    }
+    for (i = 1; i < nframe; i++)
+	fprintf(f, "  " PRINTF_P_PFX "%p [%s]\n", buf[i], symbols[i]);
+
+    return;
+}
+#endif /* HAVE_BACKTRACE */
 
 void SigBad(int sig)
 {
     __pmNotifyErr(LOG_ERR, "Unexpected signal %d ...\n", sig);
-#if HAVE_TRACE_BACK_STACK
-    fprintf(stderr, "\nProcedure call traceback ...\n");
-    do_traceback(stderr);
+#ifdef PCP_DEBUG
+    if (pmDebug & DBG_TRACE_DESPERATE) {
+	/* -D desperate on the command line to enable traceback,
+	 * if we have platform support for it
+	 */
+#if defined(HAVE_TRACE_BACK_STACK) || defined(HAVE_BACKTRACE)
+	fprintf(stderr, "\nProcedure call traceback ...\n");
+	do_traceback(stderr);
+#else
+	fprintf(stderr, "\nSorry, no procedure call traceback support ...\n");
 #endif
+    }
+#endif /* PCP_DEBUG */
     fprintf(stderr, "\nDumping to core ...\n");
     fflush(stderr);
     abort();
@@ -978,49 +872,36 @@ void SigBad(int sig)
 int
 main(int argc, char *argv[])
 {
-    int		i;
-    int		n;
     int		sts;
-    int		status;
+    int		nport = 0;
     char	*envstr;
-    unsigned	nReqPortsOK = 0;
 #ifdef HAVE_SA_SIGINFO
     static struct sigaction act;
 #endif
 
     umask(022);
+    __pmSetProgname(argv[0]);
     __pmProcessDataSize(NULL);
+    __pmGetUsername(&username);
     __pmSetInternalState(PM_STATE_PMCS);
+    __pmServerSetFeature(PM_SERVER_FEATURE_DISCOVERY);
 
-    /*
-     * get optional stuff from environment ... PMCD_PORT ...
-     * same code is in connect.c of libpcp
+    if ((envstr = getenv("PMCD_PORT")) != NULL)
+	nport = __pmServerAddPorts(envstr);
+    ParseOptions(argc, argv, &nport);
+    if (nport == 0)
+	__pmServerAddPorts(TO_STRING(SERVER_PORT));
+
+    /* Set the local socket path. A message will be generated into the log
+     * if this fails, but it is not fatal, since other connection options
+     * may exist. 
      */
-    if ((envstr = getenv("PMCD_PORT")) != NULL) {
-	char	*p = envstr;
-	char	*endptr;
-	int	port;
+    __pmServerSetLocalSocket(sockpath);
 
-	for ( ; ; ) {
-	    port = (int)strtol(p, &endptr, 0);
-	    if ((*endptr != '\0' && *endptr != ',') || port < 0) {
-		__pmNotifyErr(LOG_WARNING,
-			 "pmcd: ignored bad PMCD_PORT = '%s'", p);
-	    }
-	    else {
-		nport++;
-		if ((portlist = (int *)realloc(portlist, nport * sizeof(int))) == NULL) {
-		    __pmNoMem("pmcd: can't grow port list", nport * sizeof(int), PM_FATAL_ERR);
-		}
-		portlist[nport-1] = port;
-	    }
-	    if (*endptr == '\0')
-		break;
-	    p = &endptr[1];
-	}
-    }
-
-    ParseOptions(argc, argv);
+    /* Set the service spec. This will cause our service to be advertised on
+     * the network if that is supported.
+     */
+    __pmServerSetServiceSpec(PM_SERVER_SERVICE_SPEC);
 
     if (run_daemon) {
 	fflush(stderr);
@@ -1040,84 +921,23 @@ main(int argc, char *argv[])
     __pmSetSignalHandler(SIGBUS, SigBad);
     __pmSetSignalHandler(SIGSEGV, SigBad);
 
-    /* seed random numbers for authorisation */
-    srand48((long)time(0));
-
-    if (nport == 0) {
-	/*
-	 * no ports from $PMCD_PORT, nor from -p, set up defaults
-	 * for compatibility this used to be SERVER_PORT,4321 but
-	 * has now transitioned to just SERVER_PORT
-	 */
-	nport = 1;
-	if ((portlist = (int *)realloc(portlist, nport * sizeof(int))) == NULL) {
-	    __pmNoMem("pmcd: can't grow port list", nport * sizeof(int), PM_FATAL_ERR);
-	}
-	portlist[0] = SERVER_PORT;
-    }
-
-    /*
-     * check for duplicate ports, warn and remove duplicates
-     */
-    for (i = 0; i < nport; i++) {
-	for (n = i+1; n < nport; n++) {
-	    if (portlist[i] == portlist[n])
-		break;
-	}
-	if (n < nport) {
-	    __pmNotifyErr(LOG_WARNING,
-		     "pmcd: duplicate client request port (%d) will be ignored\n",
-		     portlist[n]);
-	    portlist[n] = -1;
-	}
-    }
-
-    if (nintf == 0) {
-	/*
-	 * no -i IP_ADDR options specified, allow connections on any
-	 * IP addr
-	 */
-	for (n = 0; n < nport; n++) {
-	    if (portlist[n] != -1)
-		AddRequestPort(NULL, portlist[n]);
-	}
-    }
-    else {
-	for (i = 0; i < nintf; i++) {
-	    for (n = 0; n < nport; n++) {
-		if (portlist[n] == -1)
-		    continue;
-		if (!AddRequestPort(intflist[i], portlist[n])) {
-		    fprintf(stderr, "pmcd: bad IP spec: -i %s\n", intflist[i]);
-		    exit(1);
-		}
-	    }
-	}
-    }
-
-    /* Open request ports for client connections */
-    for (i = 0; i < nReqPorts; i++) {
-	reqPorts[i].fd = OpenRequestSocket(reqPorts[i].port, reqPorts[i].ipAddr);
-	if (reqPorts[i].fd != -1) {
-	    if (reqPorts[i].fd > maxReqPortFd)
-		maxReqPortFd = reqPorts[i].fd;
-	    nReqPortsOK++;
-	}
-    }
-    if (nReqPortsOK == 0) {
-	__pmNotifyErr(LOG_ERR, "pmcd: can't open any request ports, exiting\n");
+    if ((sts = __pmServerOpenRequestPorts(&clientFds, MAXPENDING)) < 0)
 	DontStart();
-    }	
+    maxReqPortFd = maxClientFd = sts;
 
-    __pmOpenLog("pmcd", logfile, stderr, &status);
+    __pmOpenLog(pmProgname, logfile, stderr, &sts);
     /* close old stdout, and force stdout into same stream as stderr */
     fflush(stdout);
     close(fileno(stdout));
-    if (dup(fileno(stderr)) < 0)
-	/* tough luck, do nothing ... but makes gcc warning go away! */
-    	;
+    sts = dup(fileno(stderr));
+    /* if this fails beware of the sky falling in */
+    assert(sts >= 0);
 
-    if ((sts = pmLoadNameSpace(pmnsfile)) < 0) {
+    if (dupok)
+	sts = pmLoadASCIINameSpace(pmnsfile, 1);
+    else
+	sts = pmLoadNameSpace(pmnsfile);
+    if (sts < 0) {
 	fprintf(stderr, "Error: pmLoadNameSpace: %s\n", pmErrStr(sts));
 	DontStart();
     }
@@ -1133,24 +953,21 @@ main(int argc, char *argv[])
 	DontStart();
     }
 
-    if (run_daemon && CreatePIDfile() < 0)
+    if (run_daemon) {
+	if (CreatePIDfile() < 0)
+	    DontStart();
+	if (__pmSetProcessIdentity(username) < 0)
+	    DontStart();
+    }
+
+    if (__pmSecureServerSetup(certdb, dbpassfile) < 0)
 	DontStart();
 
     PrintAgentInfo(stderr);
-    __pmAccDumpHosts(stderr);
-    fprintf(stderr, "\npmcd: PID = %u", (int)getpid());
-    fprintf(stderr, ", PDU version = %u", PDU_VERSION);
-    fputc('\n', stderr);
-    fputs("pmcd request port(s):\n"
-	  "  sts fd  port  IP addr\n"
-	  "  === === ===== ==========\n", stderr);
-    for (i = 0; i < nReqPorts; i++) {
-	ReqPortInfo *rp = &reqPorts[i];
-	fprintf(stderr, "  %s %3d %5d 0x%08x %s\n",
-		(rp->fd != -1) ? "ok " : "err",
-		rp->fd, rp->port, rp->ipAddr,
-		rp->ipSpec ? rp->ipSpec : "(any address)");
-    }
+    __pmAccDumpLists(stderr);
+    fprintf(stderr, "\npmcd: PID = %" FMT_PID, getpid());
+    fprintf(stderr, ", PDU version = %u\n", PDU_VERSION);
+    __pmServerDumpRequestPorts(stderr);
     fflush(stderr);
 
     /* all the work is done here */
@@ -1168,17 +985,17 @@ main(int argc, char *argv[])
  * The list is cleared when PMCD is reconfigured.
  */
 
-static int		 nBadHosts = 0;
-static int		 szBadHosts = 0;
-static struct in_addr	*badHost = NULL;
+static int		 nBadHosts;
+static int		 szBadHosts;
+static __pmSockAddr	**badHost;
 
 static int
-AddBadHost(struct in_addr *hostId)
+AddBadHost(struct __pmSockAddr *hostId)
 {
     int		i, need;
 
     for (i = 0; i < nBadHosts; i++)
-	if (hostId->s_addr == badHost[i].s_addr)
+        if (__pmSockAddrCompare(hostId, badHost[i]) == 0)
 	    /* already there */
 	    return 0;
 
@@ -1186,26 +1003,33 @@ AddBadHost(struct in_addr *hostId)
     if (nBadHosts == szBadHosts) {
 	szBadHosts += 8;
 	need = szBadHosts * (int)sizeof(badHost[0]);
-	if ((badHost = (struct in_addr *)realloc(badHost, need)) == NULL) {
+	if ((badHost = (__pmSockAddr **)realloc(badHost, need)) == NULL) {
 	    __pmNoMem("pmcd.AddBadHost", need, PM_FATAL_ERR);
 	}
     }
-    badHost[nBadHosts++].s_addr = hostId->s_addr;
+    badHost[nBadHosts++] = __pmSockAddrDup(hostId);
     return 1;
 }
 
 void
 ResetBadHosts(void)
 {
-    if (szBadHosts)
+    if (szBadHosts) {
+        while (nBadHosts > 0) {
+	    --nBadHosts;
+	    free (badHost[nBadHosts]);
+	}
 	free(badHost);
+    }
     nBadHosts = 0;
     szBadHosts = 0;
+    badHost = NULL;
 }
 
 void
 CleanupClient(ClientInfo *cp, int sts)
 {
+    char	*caddr;
     int		i, msg;
     int		force;
 
@@ -1220,9 +1044,10 @@ CleanupClient(ClientInfo *cp, int sts)
 	 * been dinged for an access violation since startup or reconfiguration
 	 */
 	if (sts == PM_ERR_PERMISSION || sts == PM_ERR_CONNLIMIT) {
-	    if ( (msg = AddBadHost(&cp->addr.sin_addr)) ) {
-		fprintf(stderr, "access violation from host %s:\n",
-				inet_ntoa(cp->addr.sin_addr));
+	    if ( (msg = AddBadHost(cp->addr)) ) {
+		caddr = __pmSockAddrToString(cp->addr);
+		fprintf(stderr, "access violation from host %s\n", caddr);
+		free(caddr);
 	    }
 	}
 	else
@@ -1243,7 +1068,7 @@ CleanupClient(ClientInfo *cp, int sts)
      * count
      */
     if (sts != PM_ERR_PERMISSION && sts != PM_ERR_CONNLIMIT)
-	__pmAccDelClient(&cp->addr.sin_addr);
+        __pmAccDelClient(cp->addr);
 
     pmcd_trace(TR_DEL_CLIENT, cp->fd, sts, 0);
     DeleteClient(cp);
@@ -1258,26 +1083,23 @@ CleanupClient(ClientInfo *cp, int sts)
 
 #ifdef PCP_DEBUG
 /* Convert a file descriptor to a string describing what it is for. */
-char*
+static char *
 FdToString(int fd)
 {
-#define FDNAMELEN 40
     static char fdStr[FDNAMELEN];
     static char *stdFds[4] = {"*UNKNOWN FD*", "stdin", "stdout", "stderr"};
     int		i;
 
     if (fd >= -1 && fd < 3)
 	return stdFds[fd + 1];
-    for (i = 0; i < nReqPorts; i++) {
-	if (fd == reqPorts[i].fd) {
-	    sprintf(fdStr, "pmcd request socket %s", reqPorts[i].ipSpec);
-	    return fdStr;
-	}
-    }
+    if (__pmServerRequestPortString(fd, fdStr, FDNAMELEN) != NULL)
+	return fdStr;
     for (i = 0; i < nClients; i++)
-	if (client[i].status.connected && fd == client[i].fd) {
-	    sprintf(fdStr, "client[%d] input socket", i);
-	    return fdStr;
+        if (client[i].status.connected) {
+	    if (fd == client[i].fd) {
+	        sprintf(fdStr, "client[%d] input socket", i);
+		return fdStr;
+	    }
 	}
     for (i = 0; i < nAgents; i++)
 	if (agent[i].status.connected) {
