@@ -1,4 +1,5 @@
 /*
+ * Copyright (c) 2013 Red Hat.
  * Copyright (c) 1995-2001,2003,2004 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -15,7 +16,7 @@
 #include "pmapi.h"
 #include "impl.h"
 #include "pmda.h"
-#include "pmiestats.h"
+#include "stats.h"
 #include "pmcd/src/pmcd.h"
 #include "pmcd/src/client.h"
 #include <sys/stat.h>
@@ -66,8 +67,7 @@ static pmDesc	desctab[] = {
     { PMDA_PMID(0,14), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(0,0,0,0,0,0) },
 /* control.sighup -- push-button, pmStore to SIGHUP pmcd */
     { PMDA_PMID(0,15), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(0,0,0,0,0,0) },
-/* license -- bit-vector of license capabilities */
-    { PMDA_PMID(0,16), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* license -- bit-vector of license capabilities -- removed PMDA_PMID(0,16) */
 /* openfds -- number of open file descriptors */
     { PMDA_PMID(0,17), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
 /* buf.alloc */
@@ -76,6 +76,8 @@ static pmDesc	desctab[] = {
     { PMDA_PMID(0,19), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* build -- pcp build number */
     { PMDA_PMID(0,20), PM_TYPE_STRING, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* hostname -- local hostname -- for pmlogger */
+    { PMDA_PMID(0,21), PM_TYPE_STRING, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(0,0,0,0,0,0) },
 
 /* pdu_in.error */
     { PMDA_PMID(1,0), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
@@ -85,34 +87,34 @@ static pmDesc	desctab[] = {
     { PMDA_PMID(1,2), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_in.fetch */
     { PMDA_PMID(1,3), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.desc-req */
+/* pdu_in.desc_req */
     { PMDA_PMID(1,4), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_in.desc */
     { PMDA_PMID(1,5), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.instance-req */
+/* pdu_in.instance_req */
     { PMDA_PMID(1,6), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_in.instance */
     { PMDA_PMID(1,7), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.text-req */
+/* pdu_in.text_req */
     { PMDA_PMID(1,8), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_in.text */
     { PMDA_PMID(1,9), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.control-req */
+/* pdu_in.control_req */
     { PMDA_PMID(1,10), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.datax */
-    { PMDA_PMID(1,11), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_in.creds */
     { PMDA_PMID(1,12), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.pmns-ids */
+/* pdu_in.pmns_ids */
     { PMDA_PMID(1,13), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.pmns-names */
+/* pdu_in.pmns_names */
     { PMDA_PMID(1,14), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.pmns-child */
+/* pdu_in.pmns_child */
     { PMDA_PMID(1,15), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_in.total */
     { PMDA_PMID(1,_TOTAL), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_in.pmns-traverse */
+/* pdu_in.pmns_traverse */
     { PMDA_PMID(1,17), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
+/* pdu_in.auth */
+    { PMDA_PMID(1,18), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 
 /* pdu_out.error */
     { PMDA_PMID(2,0), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
@@ -122,34 +124,34 @@ static pmDesc	desctab[] = {
     { PMDA_PMID(2,2), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_out.fetch */
     { PMDA_PMID(2,3), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.desc-req */
+/* pdu_out.desc_req */
     { PMDA_PMID(2,4), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_out.desc */
     { PMDA_PMID(2,5), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.instance-req */
+/* pdu_out.instance_req */
     { PMDA_PMID(2,6), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_out.instance */
     { PMDA_PMID(2,7), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.text-req */
+/* pdu_out.text_req */
     { PMDA_PMID(2,8), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_out.text */
     { PMDA_PMID(2,9), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.control-req */
+/* pdu_out.control_req */
     { PMDA_PMID(2,10), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.datax */
-    { PMDA_PMID(2,11), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_out.creds */
     { PMDA_PMID(2,12), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.pmns-ids */
+/* pdu_out.pmns_ids */
     { PMDA_PMID(2,13), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.pmns-names */
+/* pdu_out.pmns_names */
     { PMDA_PMID(2,14), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.pmns-child */
+/* pdu_out.pmns_child */
     { PMDA_PMID(2,15), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 /* pdu_out.total */
     { PMDA_PMID(2,_TOTAL), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
-/* pdu_out.pmns-traverse */
+/* pdu_out.pmns_traverse */
     { PMDA_PMID(2,17), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
+/* pdu_out.auth */
+    { PMDA_PMID(2,18), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_COUNTER, PMDA_PMUNITS(0,0,1,0,0,PM_COUNT_ONE) },
 
 /* pmlogger.port */
     { PMDA_PMID(3,0), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(0,0,0,0,0,0) },
@@ -196,6 +198,21 @@ static pmDesc	desctab[] = {
 /* pmcd.cputime.per_pdu_in */
     { PMDA_PMID(7,1), PM_TYPE_DOUBLE, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,1,-1,0,PM_TIME_USEC,PM_COUNT_ONE) },
 
+/* pmcd.feature.secure */
+    { PMDA_PMID(8,0), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* pmcd.feature.compress */
+    { PMDA_PMID(8,1), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* pmcd.feature.ipv6 */
+    { PMDA_PMID(8,2), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* pmcd.feature.authentication */
+    { PMDA_PMID(8,3), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* pmcd.feature.creds_required */
+    { PMDA_PMID(8,4), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* pmcd.feature.unix_domain_sockets */
+    { PMDA_PMID(8,5), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+/* pmcd.feature.service_discovery */
+    { PMDA_PMID(8,6), PM_TYPE_U32, PM_INDOM_NULL, PM_SEM_INSTANT, PMDA_PMUNITS(0,0,0,0,0,0) },
+
 /* End-of-List */
     { PM_ID_NULL, 0, 0, 0, PMDA_PMUNITS(0, 0, 0, 0, 0, 0) }
 };
@@ -362,7 +379,7 @@ remove_pmie_indom(void)
     npmies = 0;
 }
 
-/* use a static timestamp, stat PMIE_DIR, if changed update "pmies" */
+/* use a static timestamp, stat PMIE_SUBDIR, if changed update "pmies" */
 static unsigned int
 refresh_pmie_indom(void)
 {
@@ -378,7 +395,9 @@ refresh_pmie_indom(void)
     int			fd;
     int			sep = __pmPathSeparator();
 
-    if (stat(PMIE_DIR, &statbuf) == 0) {
+    snprintf(fullpath, sizeof(fullpath), "%s%c%s",
+	     pmGetConfig("PCP_TMP_DIR"), sep, PMIE_SUBDIR);
+    if (stat(fullpath, &statbuf) == 0) {
 #if defined(HAVE_ST_MTIME_WITH_E) && defined(HAVE_STAT_TIME_T)
 	if (statbuf.st_mtime != lastsbuf.st_mtime)
 #elif defined(HAVE_ST_MTIME_WITH_SPEC)
@@ -398,9 +417,9 @@ refresh_pmie_indom(void)
 		remove_pmie_indom();
 
 	    /* open the directory iterate through mmaping as we go */
-	    if ((pmiedir = opendir(PMIE_DIR)) == NULL) {
+	    if ((pmiedir = opendir(fullpath)) == NULL) {
 		__pmNotifyErr(LOG_ERR, "pmcd pmda cannot open %s: %s",
-				PMIE_DIR, osstrerror());
+				fullpath, osstrerror());
 		return 0;
 	    }
 	    /* NOTE:  all valid files are already mmapped by pmie */
@@ -411,7 +430,9 @@ refresh_pmie_indom(void)
 		    continue;
 		if (!__pmProcessExists(pmiepid))
 		    continue;
-		snprintf(fullpath, sizeof(fullpath), "%s%c%s", PMIE_DIR, sep, dp->d_name);
+		snprintf(fullpath, sizeof(fullpath), "%s%c%s%c%s",
+			 pmGetConfig("PCP_TMP_DIR"), sep, PMIE_SUBDIR, sep,
+			 dp->d_name);
 		if (stat(fullpath, &statbuf) < 0) {
 		    __pmNotifyErr(LOG_WARNING, "pmcd pmda cannot stat %s: %s",
 				fullpath, osstrerror());
@@ -916,19 +937,13 @@ vset_resize(pmResult *rp, int i, int onumval, int numval)
     int		expect = numval;
 
     if (rp->vset[i] != NULL) {
-	if (onumval == 1)
-	    __pmPoolFree(rp->vset[i], sizeof(pmValueSet));
-	else
-	    free(rp->vset[i]);
+	free(rp->vset[i]);
     }
 
     if (numval < 0)
 	expect = 0;
 
-    if (expect == 1)
-	rp->vset[i] = (pmValueSet *)__pmPoolAlloc(sizeof(pmValueSet));
-    else
-	rp->vset[i] = (pmValueSet *)malloc(sizeof(pmValueSet) + (expect-1)*sizeof(pmValue));
+    rp->vset[i] = (pmValueSet *)malloc(sizeof(pmValueSet) + (expect-1)*sizeof(pmValue));
 
     if (rp->vset[i] == NULL) {
 	if (i) {
@@ -961,10 +976,10 @@ simabi()
     } else {
 	return abi;
     }
-#elif defined(IS_FREEBSD)
+#elif defined(IS_FREEBSD) || defined(IS_NETBSD)
     return "elf";
 #elif defined(IS_DARWIN)
-    return "Mach-O";
+    return "Mach-O " SIM_ABI;
 #elif defined(IS_MINGW)
     return "i386";	// TODO: need to handle x86_64 too
 #elif defined(IS_AIX)
@@ -990,6 +1005,27 @@ tzinfo(void)
     return __pmTimezone();
 }
 
+static char *
+hostnameinfo(void)
+{
+    static char	host[MAXHOSTNAMELEN];
+    char	*name;
+
+    (void)gethostname(host, MAXHOSTNAMELEN);
+    name = host;
+
+    return name;
+}
+
+static int
+fetch_feature(int item, pmAtomValue *avp)
+{
+    if (item < 0 || item >= PM_SERVER_FEATURES)
+	return PM_ERR_PMID;
+    avp->ul = __pmServerHasFeature((__pmServerFeature)item);
+    return 0;
+}
+
 static int
 fetch_cputime(int item, int ctx, pmAtomValue *avp)
 {
@@ -1054,7 +1090,7 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
     unsigned long	datasize;
     static pmResult	*res = NULL;
     static int		maxnpmids = 0;
-    static char		*hostname = NULL;
+    char		*host = NULL;	/* refresh max once per fetch */
     pmiestats_t		*pmie;
     pmValueSet		*vset;
     pmDesc		*dp = NULL;	/* initialize to pander to gcc */
@@ -1186,9 +1222,6 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			case 15:	/* sighup ... always 0 */
 				atom.l = 0;
 				break;
-			case 16:	/* license - hysterical raisins */
-				atom.l = 0;
-				break;
 			case 17:	/* openfds */
 				atom.ul = (unsigned int)pmcd_hi_openfds;
 				break;
@@ -1208,28 +1241,22 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 				for (j = numval = 0; j < nbufsz; j++) {
 				    int		alloced;
 				    int		free;
+				    int		xtra_alloced;
+				    int		xtra_free;
 				    if (!__pmInProfile(bufindom, _profile, bufinst[j].inst))
 					continue;
 				    vset->vlist[numval].inst = bufinst[j].inst;
-				    if (bufinst[j].inst < 1024) {
-					/* PoolAlloc pool */
-					__pmPoolCount(bufinst[j].inst, &alloced, &free);
-				    }
-				    else {
-					/* PDUBuf pool */
-					int	xtra_alloced;
-					int	xtra_free;
-					__pmCountPDUBuf(bufinst[j].inst, &alloced, &free);
-					/*
-					 * the 2K buffer count also includes
-					 * the 3K, 4K, ... buffers, so sub
-					 * these ... which are reported as
-					 * the 3K buffer count
-					 */
-					__pmCountPDUBuf(bufinst[j].inst + 1024, &xtra_alloced, &xtra_free);
-					alloced -= xtra_alloced;
-					free -= xtra_free;
-				    }
+				    /* PDUBuf pool */
+				    __pmCountPDUBuf(bufinst[j].inst, &alloced, &free);
+				    /*
+				     * the 2K buffer count also includes
+				     * the 3K, 4K, ... buffers, so sub
+				     * these ... which are reported as
+				     * the 3K buffer count
+				     */
+				    __pmCountPDUBuf(bufinst[j].inst + 1024, &xtra_alloced, &xtra_free);
+				    alloced -= xtra_alloced;
+				    free -= xtra_free;
 				    if (pmidp->item == 18)
 					atom.l = alloced;
 				    else
@@ -1245,6 +1272,19 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			case 20:	/* build */
 				atom.cp = BUILD;
 				break;
+
+			case 21:	/* hostname */
+                                if (_pmcd_hostname) {
+				    atom.cp = _pmcd_hostname;
+				} else {
+                                    if (!host)
+					host = hostnameinfo();
+                                    atom.cp = host;
+				}
+				break;
+			default:
+				sts = atom.l = PM_ERR_PMID;
+				break;
 		    }
 		    break;
 
@@ -1255,6 +1295,8 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			for (j = 0; j <= PDU_MAX; j++)
 			    atom.ul += __pmPDUCntIn[j];
 		    }
+		    else if (pmidp->item > PDU_MAX+1)
+			sts = atom.l = PM_ERR_PMID;
 		    else if (pmidp->item < _TOTAL)
 			atom.ul = __pmPDUCntIn[pmidp->item];
 		    else
@@ -1268,6 +1310,8 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			for (j = 0; j <= PDU_MAX; j++)
 			    atom.ul += __pmPDUCntOut[j];
 		    }
+		    else if (pmidp->item > PDU_MAX+1)
+			sts = atom.l = PM_ERR_PMID;
 		    else if (pmidp->item < _TOTAL)
 			atom.ul = __pmPDUCntOut[pmidp->item];
 		    else
@@ -1308,19 +1352,16 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 				atom.cp = lpp[j].archive ? lpp[j].archive : "";
 				break;
 			    case 3:		/* pmlogger.host */
-				if (hostname == NULL) {
-				    char		hbuf[MAXHOSTNAMELEN];
-				    struct hostent	*hep;
-				    (void)gethostname(hbuf, MAXHOSTNAMELEN);
-				    hbuf[MAXHOSTNAMELEN-1] = '\0';
-				    hep = gethostbyname(hbuf);
-				    if (hep != NULL)
-					hostname = strdup(hep->h_name);
-				}
-				atom.cp = (hostname != NULL) ? hostname : "";
+                                if (!host)
+				    host = hostnameinfo();
+                                atom.cp = host;
+				break;
+			    default:
+				sts = atom.l = PM_ERR_PMID;
 				break;
 			}
-			sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type);
+			if (sts >= 0)
+			    sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type);
 			if (sts < 0)
 			    break;
 			valfmt = sts;
@@ -1356,8 +1397,12 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			    else
 				atom.l = agent[j].reason;
 			    break;
+			default:
+			    sts = atom.l = PM_ERR_PMID;
+			    break;
 		    }
-		    sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type);
+		    if (sts >= 0)
+			sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type);
 		    if (sts < 0)
 			break;
 		    valfmt = sts;
@@ -1421,8 +1466,13 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			case 9:		/* pmie.eval.actual */
 			    atom.ul = pmie->eval_actual;
 			    break;
+			default:
+			    sts = atom.l = PM_ERR_PMID;
+			    break;
 		    }
-		    if ((sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type)) < 0)
+		    if (sts >= 0)
+			sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type);
+		    if (sts < 0)
 			break;
 		    valfmt = sts;
 		    numval++;
@@ -1475,8 +1525,13 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 			    k = strlen(atom.cp);
 			    atom.cp[k-1] = '\0';
 			    break;
+			default:
+			    sts = atom.l = PM_ERR_PMID;
+			    break;
 		    }
-		    if ((sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type)) < 0)
+		    if (sts >= 0)
+			sts = __pmStuffValue(&atom, &vset->vlist[numval], dp->type);
+		    if (sts < 0)
 			break;
 		    valfmt = sts;
 		    numval++;
@@ -1492,6 +1547,10 @@ pmcd_fetch(int numpmid, pmID pmidlist[], pmResult **resp, pmdaExt *pmda)
 	    case 7:	/* cputime metrics */
 		sts = fetch_cputime(pmidp->item, pmda->e_context, &atom);
 		break;
+
+	    case 8:	/* feature metrics */
+		sts = fetch_feature(pmidp->item, &atom);
+		break;
 	}
 
 	if (sts == 0 && valfmt == -1 && vset->numval == 1)
@@ -1538,9 +1597,6 @@ pmcd_store(pmResult *result, pmdaExt *pmda)
 	if (pmidp->cluster == 0) {
 	    if (pmidp->item == 0) {	/* pmcd.control.debug */
 		pmDebug = vsp->vlist[0].value.lval;
-		if (pmDebug == -1)
-		    /* terminate pmcd */
-		    _pmcd_done = 1;
 	    }
 	    else if (pmidp->item == 4) { /* pmcd.control.timeout */
 		int	val = vsp->vlist[0].value.lval;
@@ -1695,6 +1751,7 @@ pmcd_store(pmResult *result, pmdaExt *pmda)
 }
 
 void
+__PMDA_INIT_CALL
 pmcd_init(pmdaInterface *dp)
 {
     char helppath[MAXPATHLEN];
