@@ -21,7 +21,16 @@
 #include <sys/stat.h>
 #include "../domain.h"
 #ifdef HAVE_SYSINFO
+/*
+ * On Solaris, need <sys/systeminfo.h> and sysinfo() is different.
+ * Other platforms need <sys/sysinfo.h>
+ */
+#ifdef IS_SOLARIS
+#include <sys/systeminfo.h>
+#define MAX_SYSNAME	257
+#else
 #include <sys/sysinfo.h>
+#endif
 #else
 static struct sysinfo {
     char	dummy[64];
@@ -353,28 +362,17 @@ static int	ndesc = sizeof(desctab)/sizeof(desctab[0]);
 static pmDesc magic = 
     { PMDA_PMID(0,86), PM_TYPE_32, PM_INDOM_NULL, PM_SEM_DISCRETE, PMDA_PMUNITS(1,-1,0,PM_SPACE_BYTE,PM_TIME_SEC,0) };
 
-typedef struct {
-    int		i_inst;
-    char	*i_name;
-} instid_t;
-
-typedef struct {
-    pmInDom	it_indom;
-    int		it_numinst;
-    instid_t	*it_set;
-} indom_t;
-
-static instid_t	_colour[] = {
+static pmdaInstid _colour[] = {
     { 0, "red" }, { 1, "green" }, { 2, "blue" }
 };
 
-static instid_t	_bin[] = {
+static pmdaInstid _bin[] = {
     { 100, "bin-100" }, { 200, "bin-200" }, { 300, "bin-300" },
     { 400, "bin-400" }, { 500, "bin-500" }, { 600, "bin-600" },
     { 700, "bin-700" }, { 800, "bin-800" }, { 900, "bin-900" }
 };
 
-static instid_t	_scramble[] = {
+static pmdaInstid _scramble[] = {
     { 100, "bin-100" }, { 200, "bin-200" }, { 300, "bin-300" },
     { 400, "bin-400" }, { 500, "bin-500" }, { 600, "bin-600" },
     { 700, "bin-700" }, { 800, "bin-800" }, { 900, "bin-900" }
@@ -382,15 +380,15 @@ static instid_t	_scramble[] = {
 
 static long scramble_ver = 0;
 
-static instid_t	_family[] = {
+static pmdaInstid _family[] = {
     { 0, "colleen" }, { 1, "terry" }, { 2, "emma" }, { 3, "cathy" }, { 4, "fat bald bastard" }
 };
 
-static instid_t	_dodgey[] = {
+static pmdaInstid _dodgey[] = {
     { 1, NULL}, { 2, NULL }, { 3, NULL }, { 4, NULL }, { 5, NULL }
 };
 
-static instid_t _hordes[] = {
+static pmdaInstid _hordes[] = {
     {  0, "0" }, {  1, "1" }, {  2, "2" }, {  3, "3" }, {  4, "4" },
     {  5, "5" }, {  6, "6" }, {  7, "7" }, {  8, "8" }, {  9, "9" },
     { 10, "10" }, { 11, "11" }, { 12, "12" }, { 13, "13" }, { 14, "14" },
@@ -493,12 +491,12 @@ static instid_t _hordes[] = {
     {495, "495" }, {496, "496" }, {497, "497" }, {498, "498" }, {499, "499" }
 };
 
-static instid_t	_events[] = {
+static pmdaInstid _events[] = {
     { 0, "fungus" }, { 1, "bogus" }
 };
 
 /* all domains supported in this PMDA - one entry each */
-static indom_t	indomtab[] = {
+static pmdaIndom indomtab[] = {
 #define COLOUR_INDOM	0
     { 0, 3, _colour },
 #define BIN_INDOM	1
@@ -527,7 +525,7 @@ static struct timeval	_then;		/* time we started */
 static time_t		_start;		/* ditto */
 static __pmProfile	*_profile;	/* last received profile */
 static int		_x;
-static indom_t		*_idp;
+static pmdaIndom	*_idp;
 static int		_singular = -1;	/* =0 for singular values */
 static int		_ordinal = -1;	/* >=0 for non-singular values */
 static int		_control;	/* the control variable */
@@ -611,11 +609,12 @@ static int	numdyn = sizeof(dynamic_ones)/sizeof(dynamic_ones[0]);
 static int
 redo_dynamic(void)
 {
+    int			err;
     int			i;
     int			sep = __pmPathSeparator();
     static struct stat	lastsbuf;
     struct stat		statbuf;
-    indom_t		*idp = &indomtab[DYNAMIC_INDOM];
+    pmdaIndom		*idp = &indomtab[DYNAMIC_INDOM];
     char		mypath[MAXPATHLEN];
 
     snprintf(mypath, sizeof(mypath), "%s%c" "sample" "%c" "dynamic.indom",
@@ -655,19 +654,26 @@ redo_dynamic(void)
 		    if (fscanf(fspec, "%d %s", &newinst, newname) != 2)
 			break;
 		    numinst++;
-		    if ((idp->it_set = (instid_t *)realloc(idp->it_set, numinst * sizeof(instid_t))) == NULL)
-			return -oserror();
+		    if ((idp->it_set = (pmdaInstid *)realloc(idp->it_set, numinst * sizeof(pmdaInstid))) == NULL) {
+			err = -oserror();
+			fclose(fspec);
+			return err;
+		    }
 		    idp->it_set[numinst-1].i_inst = newinst;
 		    if ((idp->it_set[numinst-1].i_name = strdup(newname)) == NULL) {
+			err = -oserror();
 			free(idp->it_set);
 			idp->it_set = NULL;
-			return -oserror();
+			fclose(fspec);
+			return err;
 		    }
 		    if (newinst > _dyn_max) {
 			if ((_dyn_ctr = (int *)realloc(_dyn_ctr, (newinst+1)*sizeof(_dyn_ctr[0]))) == NULL) {
+			    err = -oserror();
 			    free(idp->it_set);
 			    idp->it_set = NULL;
-			    return -oserror();
+			    fclose(fspec);
+			    return err;
 			}
 			for (i = _dyn_max+1; i <= newinst; i++)
 			    _dyn_ctr[i] = 0;
@@ -726,7 +732,7 @@ redo_dynamic(void)
 static int
 redo_many(void)
 {
-    indom_t     	*idp;
+    pmdaIndom   	*idp;
     int			a;
     static char		*tags=NULL;
     char		*tag;
@@ -740,7 +746,7 @@ redo_many(void)
 
     /* realloc instances buffer */
 
-    idp->it_set=realloc(idp->it_set,many_count*sizeof(instid_t));
+    idp->it_set = realloc(idp->it_set, many_count*sizeof(pmdaInstid));
     if (!idp->it_set) {
 	idp->it_numinst=0;
 	many_count=0;
@@ -749,7 +755,7 @@ redo_many(void)
 
     /* realloc string buffer */
 
-    tags=realloc(tags,many_count*MANY_MAX_LEN);
+    tags = realloc(tags, many_count*MANY_MAX_LEN);
     if (!idp->it_set) {
 	idp->it_numinst=0;
 	many_count=0;
@@ -780,7 +786,7 @@ redo_mirage(void)
     int			i;
     int			j;
     static int		newinst = 0;
-    indom_t		*idp;
+    pmdaIndom		*idp;
 
     now = time(NULL);
     if (now < doit)
@@ -789,7 +795,7 @@ redo_mirage(void)
     idp = &indomtab[MIRAGE_INDOM];
     if (idp->it_set == NULL) {
 	/* first time */
-	if ((idp->it_set = (instid_t *)malloc(sizeof(instid_t))) == NULL)
+	if ((idp->it_set = (pmdaInstid *)malloc(sizeof(pmdaInstid))) == NULL)
 	    return -oserror();
 	if ((idp->it_set[0].i_name = (char *)malloc(5)) == NULL) {
 	    idp->it_set = NULL;
@@ -814,7 +820,7 @@ redo_mirage(void)
 	    idp->it_set[numinst++] = idp->it_set[i];
 	}
 	if (numinst != idp->it_numinst) {
-	    if ((idp->it_set = (instid_t *)realloc(idp->it_set, numinst * sizeof(instid_t))) == NULL) {
+	    if ((idp->it_set = (pmdaInstid *)realloc(idp->it_set, numinst * sizeof(pmdaInstid))) == NULL) {
 		idp->it_set = NULL;
 		idp->it_numinst = 0;
 		return -oserror();
@@ -825,7 +831,7 @@ redo_mirage(void)
 	    if (lrand48() % 1000 < 500) {
 		/* add a new one */
 		numinst++;
-		if ((idp->it_set = (instid_t *)realloc(idp->it_set, numinst * sizeof(instid_t))) == NULL) {
+		if ((idp->it_set = (pmdaInstid *)realloc(idp->it_set, numinst * sizeof(pmdaInstid))) == NULL) {
 		    idp->it_set = NULL;
 		    idp->it_numinst = 0;
 		    return -oserror();
@@ -913,7 +919,7 @@ redo_dodgey(void)
 static int
 cntinst(pmInDom indom)
 {
-    indom_t	*idp;
+    pmdaIndom	*idp;
 
     if (indom == PM_INDOM_NULL)
 	return 1;
@@ -1174,8 +1180,8 @@ static int
 sample_instance(pmInDom indom, int inst, char *name, __pmInResult **result, pmdaExt *ep)
 {
     int		i;
-    __pmInResult  *res;
-    indom_t	*idp;
+    __pmInResult *res;
+    pmdaIndom	*idp;
     int		err = 0;
 
     sample_inc_recv(ep->e_context);
@@ -1293,7 +1299,7 @@ sample_instance(pmInDom indom, int inst, char *name, __pmInResult **result, pmda
     return 0;
 }
 
-int
+static int
 sample_pmid(const char *name, pmID *pmid, pmdaExt *pmda)
 {
     int		i;
@@ -1314,7 +1320,7 @@ sample_pmid(const char *name, pmID *pmid, pmdaExt *pmda)
     return PM_ERR_NAME;
 }
 
-int
+static int
 sample_name(pmID pmid, char ***nameset, pmdaExt *pmda)
 {
     size_t	len = 0;
@@ -1361,7 +1367,7 @@ sample_name(pmID pmid, char ***nameset, pmdaExt *pmda)
     return nmatch;
 }
 
-int
+static int
 sample_children(const char *name, int traverse, char ***offspring, int **status, pmdaExt *pmda)
 {
     int		i;
@@ -1499,6 +1505,27 @@ fail:
      return j;
 }
 
+static int
+sample_attribute(int ctx, int attr, const char *value, int length, pmdaExt *pmda)
+{
+    /*
+     * We have no special security or other requirements, so we're just
+     * going to log any connection attribute messages we happen to get
+     * from pmcd (handy for demo and testing purposes).
+     */
+    if (pmDebug & DBG_TRACE_AUTH) {
+	char buffer[256];
+
+	if (!__pmAttrStr_r(attr, value, buffer, sizeof(buffer))) {
+	    __pmNotifyErr(LOG_ERR, "Bad Attribute: ctx=%d, attr=%d\n", ctx, attr);
+	} else {
+	    buffer[sizeof(buffer)-1] = '\0';
+	    __pmNotifyErr(LOG_INFO, "Attribute: ctx=%d %s", ctx, buffer);
+	}
+    }
+    return 0;
+}
+
 /*
  * high precision counter
  */
@@ -1671,9 +1698,7 @@ doit:
 	    numval = 0;
 
 	/* Must use individual malloc()s because of pmFreeResult() */
-	if (numval == 1)
-	    res->vset[i] = vset = (pmValueSet *)__pmPoolAlloc(sizeof(pmValueSet));
-	else if (numval > 1)
+	if (numval >= 1)
 	    res->vset[i] = vset = (pmValueSet *)malloc(sizeof(pmValueSet) + 
 					    (numval - 1)*sizeof(pmValue));
 	else
@@ -1729,7 +1754,7 @@ doit:
 			atom.l = _control;
 			break;
 		    case 1:
-			if (_mypid == 0) _mypid = getpid();
+			if (_mypid == 0) _mypid = (int)getpid();
 			atom.ul = _mypid;
 			break;
 		    case 2:
@@ -1849,8 +1874,13 @@ doit:
 			    /* malloc and init the pmValueBlock for
 			     * sysinfo first type around */
 
-			    int size = sizeof(pmValueBlock) - sizeof(int) + 
-				sizeof (struct sysinfo);
+			    int size = sizeof(pmValueBlock) - sizeof(int);
+
+#ifdef IS_SOLARIS
+			    size += MAX_SYSNAME;
+#else
+			    size += sizeof (struct sysinfo);
+#endif
 
 			    if ((sivb = calloc(1, size)) == NULL ) 
 				return PM_ERR_GENERIC;
@@ -1860,7 +1890,11 @@ doit:
 			}
 
 #ifdef HAVE_SYSINFO
+#ifdef IS_SOLARIS
+			sysinfo(SI_SYSNAME, sivb->vbuf, MAX_SYSNAME);
+#else
 			sysinfo((struct sysinfo *)sivb->vbuf);
+#endif
 #else
 			strncpy((char *)sivb->vbuf, si.dummy, sizeof(struct sysinfo));
 #endif
@@ -2335,7 +2369,7 @@ done:
 	vset->numval = j;
     }
     *resp = res;
-    return 1;
+    return PMDA_FETCH_STATIC;
 }
 
 static int
@@ -2500,7 +2534,7 @@ sample_store(pmResult *result, pmdaExt *ep)
 		break;
 
 	    default:
-		sts = -EACCES;
+		sts = PM_ERR_PERMISSION;
 		break;
 
 	}
@@ -2600,6 +2634,7 @@ sample_store(pmResult *result, pmdaExt *ep)
 	    case 90:	/* dynamic.meta.pmdesc.units */
 		lp = (__int32_t *)&magic.units;
 		*lp = av.l;
+		break;
 	    case 97:	/* ulong.write_me */
 		_ulong = av.ul;
 		break;
@@ -2618,7 +2653,7 @@ sample_store(pmResult *result, pmdaExt *ep)
 		event_set_fetch_count(av.l);
 		break;
 	    default:
-		sts = -EACCES;
+		sts = PM_ERR_PERMISSION;
 		break;
 	}
     }
@@ -2626,7 +2661,9 @@ sample_store(pmResult *result, pmdaExt *ep)
     return sts;
 }
 
-void sample_init(pmdaInterface *dp)
+void
+__PMDA_INIT_CALL
+sample_init(pmdaInterface *dp)
 {
     char	helppath[MAXPATHLEN];
     int		i;
@@ -2643,16 +2680,18 @@ void sample_init(pmdaInterface *dp)
 
     if (dp->status != 0)
 	return;
+    dp->comm.flags |= PDU_FLAG_AUTH;
 
-    dp->version.four.fetch = sample_fetch;
-    dp->version.four.desc = sample_desc;
-    dp->version.four.instance = sample_instance;
-    dp->version.four.text = sample_text;
-    dp->version.four.store = sample_store;
-    dp->version.four.profile = sample_profile;
+    dp->version.any.fetch = sample_fetch;
+    dp->version.any.desc = sample_desc;
+    dp->version.any.instance = sample_instance;
+    dp->version.any.text = sample_text;
+    dp->version.any.store = sample_store;
+    dp->version.any.profile = sample_profile;
     dp->version.four.pmid = sample_pmid;
     dp->version.four.name = sample_name;
     dp->version.four.children = sample_children;
+    dp->version.six.attribute = sample_attribute;
     pmdaSetEndContextCallBack(dp, sample_ctx_end);
 
     pmdaInit(dp, NULL, 0, NULL, 0);	/* don't use indomtab or metrictab */
@@ -2686,6 +2725,4 @@ void sample_init(pmdaInterface *dp)
     _dodgey[2].i_name = strdup("d3");
     _dodgey[3].i_name = strdup("d4");
     _dodgey[4].i_name = strdup("d5");
-
-    return;
 }
