@@ -2,6 +2,7 @@
  * dstruct.c - central data structures and associated operations
  ***********************************************************************
  *
+ * Copyright (c) 2013 Red Hat.
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -35,15 +36,11 @@
 #if defined(HAVE_IEEEFP_H)
 #include <ieeefp.h>
 #endif
-#ifndef HAVE_MEMALIGN
-#define memalign(a,b) malloc(b)
-#endif
 
 /***********************************************************************
  * constants
  ***********************************************************************/
 
-char	localHost[MAXHOSTNAMELEN+1];	/* "official" name of localhost */
 double	mynan;				/* not-a-number run time initialized */
 
 
@@ -55,7 +52,8 @@ char		*pmnsfile = PM_NS_DEFAULT;	/* alternate name space */
 Archive		*archives;			/* list of open archives */
 RealTime	first = -1;			/* archive starting point */
 RealTime	last = 0.0;			/* archive end point */
-char		*dfltHost;			/* default host */
+char		*dfltHostConn;			/* default host connection string */
+char		*dfltHostName;			/* default host name */
 RealTime	dfltDelta = DELTA_DFLT;		/* default sample interval */
 char		*startFlag;			/* start time specified? */
 char		*stopFlag;			/* end time specified? */
@@ -118,7 +116,6 @@ static double	weekday;		/* days since Sunday 0..6 */
  ***********************************************************************/
 int		need_wait;
 
-
 /* return real time */
 RealTime
 getReal(void)
@@ -209,7 +206,7 @@ sleepTight(Task *t, int type)
 	while ((pid = waitpid(-1, &sts, WNOHANG)) > (pid_t)0) {
 #if PCP_DEBUG
 	    if (pmDebug & DBG_TRACE_APPL2) {
-		fprintf(stderr, "sleepTight: wait: pid=%d done status=0x%x", (int)pid, sts);
+		fprintf(stderr, "sleepTight: wait: pid=%" FMT_PID " done status=0x%x", pid, sts);
 		if (WIFEXITED(sts))
 		    fprintf(stderr, " exit=%d", WEXITSTATUS(sts));
 		if (WIFSIGNALED(sts))
@@ -377,9 +374,20 @@ zalloc(size_t size)
 void *
 aalloc(size_t align, size_t size)
 {
-    void *p;
-
-    if ((p = memalign(align, size)) == NULL) {
+    void	*p;
+    int		sts = 0;
+#ifdef HAVE_POSIX_MEMALIGN
+    sts = posix_memalign(&p, align, size);
+#else
+#ifdef HAVE_MEMALIGN
+    p = memalign(align, size);
+    if (p == NULL) sts = -1;
+#else
+    p = malloc(size);
+    if (p == NULL) sts = -1;
+#endif
+#endif
+    if (sts != 0) {
 	__pmNoMem("pmie.aalloc", size, PM_FATAL_ERR);
     }
     return p;
@@ -437,7 +445,13 @@ newExpr(int op, Expr *arg1, Expr *arg2,
 	arg = primary(arg1, arg2);
 	x->metrics = arg->metrics;
     }
-    x->units = noUnits;
+    if (sem == SEM_NUMVAR || sem == SEM_NUMCONST || sem == SEM_TRUTH ||
+        sem == SEM_CHAR || sem == SEM_REGEX)
+	x->units = noUnits;
+    else {
+	x->units = noUnits;
+	SET_UNITS_UNKNOWN(x->units);
+    }
     x->sem = sem;
     return x;
 }
@@ -630,7 +644,6 @@ freeMetric(Metric *m)
     }
     if (numinst && m->iids) free(m->iids);
     if (m->vals) free(m->vals);
-    free(m);
 }
 
 
@@ -646,17 +659,13 @@ freeExpr(Expr *x)
 	if (x->arg2 && x->arg2->parent == x)
 	    freeExpr(x->arg2);
 	if (x->metrics && x->op == CND_FETCH) {
-	    if (x->hdom < 0) {
-		/*
-		 * no trips through the loop to call freeMetric(),
-		 * so free partially allocated structure
-		 */
-		free(x->metrics);
-	    }
-	    else {
-		for (m = x->metrics, i = 0; i < x->hdom; m++, i++)
-		    freeMetric(m);
-	    }
+	    for (m = x->metrics, i = 0; i < x->hdom; m++, i++)
+		freeMetric(m);
+	    /*
+	     * x->metrics allocated in a block, one element per host, so
+	     * free as one after all other freeing has been done.
+	     */
+	    free(x->metrics);
 	}
 	if (x->ring) free(x->ring);
 	free(x);
@@ -745,43 +754,54 @@ instExpr(Expr *x)
     Expr    *arg1 = x->arg1;
     Expr    *arg2 = x->arg2;
     Expr    *arg = primary(arg1, arg2);
-    pmUnits u;
 
-    /* semantics and units */
+    /* semantics ... */
     if (x->sem == SEM_UNKNOWN) {
-
-	/* unary expression */
 	if (arg2 == NULL) {
-	    up = 1;
-	    x->sem = arg1->sem;
-	    x->units = arg1->units;
+	    /* unary expression */
+	    if (arg1->sem != SEM_UNKNOWN) {
+		up = 1;
+		x->sem = arg1->sem;
+	    }
 	}
-
-	/* bianry expression with known args */
 	else if ((arg1->sem != SEM_UNKNOWN) &&
 		 (arg2->sem != SEM_UNKNOWN)) {
+	    /* binary expression with known args */
 	    up = 1;
 	    x->sem = arg->sem;
-	    if (x->op == CND_MUL) {
-		u.dimSpace = arg1->units.dimSpace + arg2->units.dimSpace;
-		u.dimTime = arg1->units.dimTime + arg2->units.dimTime;
-		u.dimCount = arg1->units.dimCount + arg2->units.dimCount;
-	    }
-	    else if (x->op == CND_DIV) {
-		u.dimSpace = arg1->units.dimSpace - arg2->units.dimSpace;
-		u.dimTime = arg1->units.dimTime - arg2->units.dimTime;
-		u.dimCount = arg1->units.dimCount - arg2->units.dimCount;
-	    }
-	    else
-		u = arg->units;
-	    x->units = u;
 	}
-
 	/* binary expression with unknown arg */
 	else
 	    return;
     }
 
+    /* units ... */
+    if (UNITS_UNKNOWN(x->units)) {
+	if (arg2 == NULL) {
+	    /* unary expression */
+	    if (!UNITS_UNKNOWN(arg1->units)) {
+		up = 1;
+		x->units = arg1->units;
+	    }
+	}
+	else if (!UNITS_UNKNOWN(arg1->units) &&
+		 !UNITS_UNKNOWN(arg2->units)) {
+	    /* binary expression with known args */
+	    up = 1;
+	    x->units = arg->units;
+	    if (x->op == CND_MUL) {
+		x->units.dimSpace = arg1->units.dimSpace + arg2->units.dimSpace;
+		x->units.dimTime = arg1->units.dimTime + arg2->units.dimTime;
+		x->units.dimCount = arg1->units.dimCount + arg2->units.dimCount;
+	    }
+	    else if (x->op == CND_DIV) {
+		x->units.dimSpace = arg1->units.dimSpace - arg2->units.dimSpace;
+		x->units.dimTime = arg1->units.dimTime - arg2->units.dimTime;
+		x->units.dimCount = arg1->units.dimCount - arg2->units.dimCount;
+	    }
+	}
+    }
+
     /* instance domain */
     if ((x->e_idom != -1) && (x->e_idom != arg->e_idom)) {
 	up = 1;
@@ -848,6 +868,7 @@ instFetchExpr(Expr *x)
 	m++;
     }
     if (x->e_idom != ninst) {
+	/* number of instances is different */
 	x->e_idom = ninst;
 	x->tspan = (x->e_idom >= 0) ? x->e_idom : abs(x->hdom);
 	x->nvals = x->nsmpls * x->tspan;
@@ -855,8 +876,13 @@ instFetchExpr(Expr *x)
 	newRingBfr(x);
 	up = 1;
     }
-    if (up && x->parent)
-	instExpr(x->parent);
+    if (x->parent) {
+	/* do we need to propagate changes? */
+	if (up ||
+	    (UNITS_UNKNOWN(x->parent->units) && !UNITS_UNKNOWN(x->units))) {
+	    instExpr(x->parent);
+	}
+    }
 }
 
 
@@ -872,10 +898,7 @@ void dstructInit(void)
     /* not-a-number initialization */
     mynan = zero / zero;
 
-    /* initialize default host */
-    gethostname(localHost, MAXHOSTNAMELEN);
-    localHost[MAXHOSTNAMELEN-1] = '\0';
-    dfltHost = localHost;
+    /* don't initialize dfltHost*; let pmie.c do it after getopt. */
 
     /* set up symbol tables */
     symSetTable(&hosts);
@@ -1109,7 +1132,10 @@ __dumpExpr(int level, Expr *x)
     }
     if (sem_map[j].name == NULL)
 	fprintf(stderr, "%d", x->sem);
-    fprintf(stderr, " units=%s\n", pmUnitsStr(&x->units));
+    if (UNITS_UNKNOWN(x->units))
+	fprintf(stderr, " units=UNKNOWN\n");
+    else
+	fprintf(stderr, " units=%s\n", pmUnitsStr(&x->units));
     if (x->valid > 0) {
 	if (x->sem == SEM_TRUTH || x->sem == SEM_CHAR ||
 	    x->sem == SEM_NUMVAR || x->sem == SEM_NUMCONST ||
@@ -1141,7 +1167,15 @@ __dumpExpr(int level, Expr *x)
 		    }
 		    else {
 			double	v = *((double *)x->smpls[j].ptr+k);
-			if (isnand(v))
+			int		fp_bad = 0;
+#ifdef HAVE_FPCLASSIFY
+			fp_bad = fpclassify(v) == FP_NAN;
+#else
+#ifdef HAVE_ISNAN
+			fp_bad = isnan(v);
+#endif
+#endif
+			if (fp_bad)
 			    fputc('?', stderr);
 			else
 			    fprintf(stderr, "%g", v);
@@ -1254,7 +1288,10 @@ dumpTask(Task *t)
     fprintf(stderr, "  retry time: ");
     showFullTime(stderr, t->retry);
     fputc('\n', stderr);
-    fprintf(stderr, "  host=%s (%s)\n", symName(t->hosts->name), t->hosts->down ? "down" : "up");
+    if (t->hosts == NULL)
+	fprintf(stderr, "  host=<null>\n");
+    else
+	fprintf(stderr, "  host=%s (%s)\n", symName(t->hosts->name), t->hosts->down ? "down" : "up");
     fprintf(stderr, "  rules:\n");
     for (i = 0; i < t->nrules; i++) {
 	fprintf(stderr, "    %s\n", symName(t->rules[i]));
