@@ -2,6 +2,7 @@
  * pmie.c - performance inference engine
  ***********************************************************************
  *
+ * Copyright (c) 2013 Red Hat, Inc.
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -27,6 +28,7 @@
 #include "pmapi.h"
 #include "impl.h"
 #include <sys/stat.h>
+#include <assert.h>
 
 #include "dstruct.h"
 #include "stomp.h"
@@ -48,7 +50,6 @@
 
 #define LINE_LENGTH	255		/* max length of command token */
 #define PROC_FNAMESIZE  20              /* from proc pmda - proc.h */
-#define PMIE_PATHSIZE   (sizeof(PMIE_DIR)+PROC_FNAMESIZE)
 
 static char *prompt = "pmie> ";
 static char *intro  = "Performance Co-Pilot Inference Engine (pmie), "
@@ -57,7 +58,8 @@ char	*clientid;
 
 static FILE *logfp;
 static char logfile[MAXPATHLEN+1];
-static char perffile[PMIE_PATHSIZE];	/* /var/tmp/<pid> file name */
+static char perffile[MAXPATHLEN+1];	/* /var/tmp/<pid> file name */
+static char *username;
 
 static char menu[] =
 "pmie debugger commands\n\n"
@@ -91,6 +93,7 @@ static char usage[] =
     "  -S starttime start of the time window\n"
     "  -T endtime   end of the time window\n"
     "  -t interval  sample interval [default 10 seconds]\n"
+    "  -U username  in daemon mode, run as named user [default pcp]\n"
     "  -V           verbose mode, annotated expression values printed\n"
     "  -v           verbose mode, expression values printed\n"
     "  -W           verbose mode, satisfying expression values printed\n"
@@ -206,8 +209,8 @@ load(char *fname)
 	    fprintf(stderr, "load: cannot access config file %s: %s\n", fname, strerror(sts));
 	}
 #endif
-	snprintf(config, sizeof(config), "%s%c" "config" "%c" "pmie" "%c%s",
-		pmGetConfig("PCP_VAR_DIR"), sep, sep, sep, fname);
+	snprintf(config, sizeof(config)-1, "%s%c" "pmie" "%c%s",
+		pmGetConfig("PCP_SYSCONF_DIR"), sep, sep, fname);
 	if (access(config, F_OK) != 0) {
 	    fprintf(stderr, "%s: cannot access config file as either %s or %s: %s\n",
 		    pmProgname, fname, config, strerror(sts));
@@ -318,30 +321,31 @@ stopmonitor(void)
 static void
 startmonitor(void)
 {
-    struct hostent	*hep = noDnsFlag ? NULL : gethostbyname(dfltHost);
     void		*ptr;
+    char		*path;
     int			fd;
     char		zero = '\0';
+    char		pmie_dir[MAXPATHLEN];
 
     /* try to create the port file directory. OK if it already exists */
-    if ( (mkdir2(PMIE_DIR, S_IRWXU | S_IRWXG | S_IRWXO) < 0) &&
-	 (oserror() != EEXIST) ) {
-	fprintf(stderr, "%s: error creating stats file dir %s: %s\n",
-		pmProgname, PMIE_DIR, osstrerror());
-	exit(1);
+    snprintf(pmie_dir, sizeof(pmie_dir), "%s%c%s",
+	     pmGetConfig("PCP_TMP_DIR"), __pmPathSeparator(), PMIE_SUBDIR);
+    if (mkdir2(pmie_dir, S_IRWXU | S_IRWXG | S_IRWXO) < 0) {
+	if (oserror() != EEXIST) {
+	    fprintf(stderr, "%s: warning cannot create stats file dir %s: %s\n",
+		    pmProgname, pmie_dir, osstrerror());
+	}
     }
-
-    chmod(PMIE_DIR, S_IRWXU | S_IRWXG | S_IRWXO | S_ISVTX);
     atexit(stopmonitor);
 
     /* create and initialize memory mapped performance data file */
-    sprintf(perffile, "%s%c%u", PMIE_DIR, __pmPathSeparator(), (int)getpid());
+    sprintf(perffile, "%s%c%" FMT_PID, pmie_dir, __pmPathSeparator(), getpid());
     unlink(perffile);
     if ((fd = open(perffile, O_RDWR | O_CREAT | O_EXCL | O_TRUNC,
 			     S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
-	fprintf(stderr, "%s: cannot create stats file %s: %s\n",
-		pmProgname, perffile, osstrerror());
-	exit(1);
+	/* cannot create stats file; too bad, so sad, continue on without it */
+	perf = &instrument;
+	return;
     }
     /* seek to struct size and write one zero */
     lseek(fd, sizeof(pmiestats_t)-1, SEEK_SET);
@@ -354,13 +358,17 @@ startmonitor(void)
     if ((ptr = __pmMemoryMap(fd, sizeof(pmiestats_t), 1)) == NULL) {
 	fprintf(stderr, "%s: memory map failed for stats file %s: %s\n",
 		pmProgname, perffile, osstrerror());
-	exit(1);
+	perf = &instrument;
+    } else {
+	perf = (pmiestats_t *)ptr;
     }
     close(fd);
 
-    perf = (pmiestats_t *)ptr;
-    strcpy(perf->logfile, logfile[0] == '\0'? "<none>" : logfile);
-    strcpy(perf->defaultfqdn, hep == NULL? dfltHost : hep->h_name);
+    path = (logfile[0] == '\0') ? "<none>" : logfile;
+    strncpy(perf->logfile, path, sizeof(perf->logfile));
+    perf->logfile[sizeof(perf->logfile)-1] = '\0';
+    strncpy(perf->defaultfqdn, dfltHostName, sizeof(perf->defaultfqdn));
+    perf->defaultfqdn[sizeof(perf->defaultfqdn)-1] = '\0';
     perf->version = 1;
 }
 
@@ -409,8 +417,8 @@ sighupproc(int sig)
 
     fp = __pmRotateLog(pmProgname, logfile, logfp, &sts);
     if (sts != 0) {
-	fprintf(stderr, "pmie: PID = %d, default host = %s\n\n",
-			(int)getpid(), dfltHost);
+	fprintf(stderr, "pmie: PID = %" FMT_PID ", default host = %s via %s\n\n",
+                getpid(), dfltHostName, dfltHostConn);
 	remap_stdout_stderr();
 	logfp = fp;
     } else {
@@ -469,8 +477,10 @@ getargs(int argc, char *argv[])
     int			sts;
     int			c;
     int			bflag = 0;
+    int			dfltConn = 0;	/* default context type */
     Archive		*a;
     struct timeval	tv, tv1, tv2;
+
     extern int		showTimeFlag;
     extern int		errs;		/* syntax errors from syntax.c */
 
@@ -479,11 +489,12 @@ getargs(int argc, char *argv[])
     memset(&tv2, 0, sizeof(tv2));
     dstructInit();
 
-    while ((c=getopt(argc, argv, "a:A:bc:CdD:efHh:j:l:n:O:S:t:T:vVWXxzZ:?")) != EOF) {
+    while ((c=getopt(argc, argv, "a:A:bc:CdD:efHh:j:l:n:O:S:t:T:U:vVWXxzZ:?")) != EOF) {
         switch (c) {
 
 	case 'a':			/* archives */
 	    if (dfltConn && dfltConn != PM_CONTEXT_ARCHIVE) {
+                /* (technically, multiple -a's are allowed.) */
 		fprintf(stderr, "%s: at most one of -a or -h allowed\n",
 			pmProgname);
 		err++;
@@ -497,7 +508,7 @@ getargs(int argc, char *argv[])
 		if (subopts != NULL) {
 		    *subopts++ = '\0';
 		}
-		a = (Archive *) zalloc(sizeof(Archive));
+		a = (Archive *)zalloc(sizeof(Archive));
 		a->fname = subopt;
 		if (!initArchive(a)) {
 		    exit(1);
@@ -570,7 +581,8 @@ getargs(int argc, char *argv[])
 		break;
 	    }
 	    dfltConn = PM_CONTEXT_HOST;
-	    dfltHost = optarg;
+	    dfltHostConn = optarg;
+            dfltHostName = ""; /* unknown until newContext */
 	    break;
 
         case 'j':			/* stomp protocol (JMS) config */
@@ -615,6 +627,11 @@ getargs(int argc, char *argv[])
 	    stopFlag = optarg;
 	    break;
 
+	case 'U': 			/* run as named user */
+	    username = optarg;
+	    isdaemon = 1;
+	    break;
+
 	case 'v': 			/* print values */
 	    verbose = 1;
 	    break;
@@ -682,6 +699,9 @@ getargs(int argc, char *argv[])
 	perf = &instrument;
 
     if (isdaemon) {			/* daemon mode */
+	/* done before opening log to get permissions right */
+	__pmSetProcessIdentity(username);
+
 #if defined(HAVE_TERMIO_SIGNALS)
 	signal(SIGTTOU, SIG_IGN);
 	signal(SIGTTIN, SIG_IGN);
@@ -716,24 +736,41 @@ getargs(int argc, char *argv[])
     if ((bflag || isdaemon) && !agent)
 	remap_stdout_stderr();
 
-    if (dfltConn == 0) {
-	/* default case, no -a or -h */
-	dfltConn = PM_CONTEXT_HOST;
-	dfltHost = localHost;
-    }
-
-    if (!archives && !interactive) {
-	if (commandlog != NULL)
-	    fprintf(stderr, "pmie: PID = %d, default host = %s\n\n", (int)getpid(), dfltHost);
-	startmonitor();
-    }
-
-    /* default host from leftmost archive on command line */
-    if (archives && dfltHost == localHost) {
+    /* default host from leftmost archive on command line, or from
+       discovery after a brief connection */
+    if (archives) {
 	a = archives;
 	while (a->next)
 	    a = a->next;
-	dfltHost = a->hname;
+	dfltHostName = a->hname; /* already filled in during initArchive() */
+    } else if (!dfltConn || dfltConn == PM_CONTEXT_HOST) {
+	if (dfltConn == 0)	/* default case, no -a or -h */
+	    dfltHostConn = "local:";
+	sts = pmNewContext(PM_CONTEXT_HOST, dfltHostConn);
+	/* pmcd down locally, try to extract hostname manually */
+	if (sts < 0 && (!dfltConn ||
+			!strcmp(dfltHostConn, "localhost") ||
+			!strcmp(dfltHostConn, "local:") ||
+			!strcmp(dfltHostConn, "unix:")))
+	    sts = pmNewContext(PM_CONTEXT_LOCAL, NULL);
+	if (sts < 0) {
+	    __pmNotifyErr(LOG_ERR, "%s: cannot find host name for %s\n"
+			"pmNewContext failed: %s\n",
+			pmProgname, dfltHostConn, pmErrStr(sts));
+	    dfltHostName = "?";
+	} else {
+	    if ((dfltHostName = strdup(pmGetContextHostName(sts))) == NULL)
+		__pmNoMem("host name copy", 0, PM_FATAL_ERR);
+	    pmDestroyContext(sts);
+        }
+    }
+    assert (dfltHostName != NULL);
+
+    if (!archives && !interactive) {
+	if (commandlog != NULL)
+            fprintf(stderr, "pmie: PID = %" FMT_PID ", default host = %s via %s\n\n",
+                    getpid(), dfltHostName, dfltHostConn);
+	startmonitor();
     }
 
     /* initialize time */
@@ -743,10 +780,12 @@ getargs(int argc, char *argv[])
 
     /* parse time window - just to check argument syntax */
     unrealize(now, &tv1);
-    if (archives)
+    if (archives) {
 	unrealize(last, &tv2);
-    else
+    } else {
 	tv2.tv_sec = INT_MAX;		/* sizeof(time_t) == sizeof(int) */
+	tv2.tv_usec = 0;
+    }
     if (pmParseTimeWindow(startFlag, stopFlag, alignFlag, offsetFlag,
                           &tv1, &tv2,
                           &tv, &tv2, &tv1,
@@ -813,10 +852,12 @@ getargs(int argc, char *argv[])
 	reflectTime(dfltDelta);
     }
     unrealize(now, &tv1);
-    if (archives)
+    if (archives) {
 	unrealize(last, &tv2);
-    else
+    } else {
 	tv2.tv_sec = INT_MAX;
+	tv2.tv_usec = 0;
+    }
     if (pmParseTimeWindow(startFlag, stopFlag, alignFlag, offsetFlag,
 		          &tv1, &tv2,
                           &tv, &tv2, &tv1,
@@ -839,7 +880,7 @@ static void
 interact(void)
 {
     int			quit = 0;
-    char		*line = (char *) zalloc(LINE_LENGTH + 2);
+    char		*line = (char *)zalloc(LINE_LENGTH + 2);
     char		*finger;
     char		*token;
     char		*msg;
@@ -857,6 +898,10 @@ interact(void)
 	    case 'f':
 		token = scanArg(finger);
 		load(token);
+#if PCP_DEBUG
+		if (pmDebug & DBG_TRACE_APPL1)
+		    dumpRules();
+#endif
 		break;
 
 	    case 'l':
@@ -893,10 +938,12 @@ interact(void)
 		    break;
 		}
 		unrealize(start, &tv1);
-		if (archives)
+		if (archives) {
 		    unrealize(last, &tv2);
-		else
+		} else {
 		    tv2.tv_sec = INT_MAX;
+		    tv2.tv_usec = 0;
+		}
 		if (__pmParseTime(token, &tv1, &tv2, &tv1, &msg) < 0) {
 		    fputs(msg, stderr);
 		    free(msg);
@@ -939,6 +986,7 @@ interact(void)
 	    fflush(stdout);
 	}
     }
+    free(line);
 }
 
 
@@ -950,6 +998,7 @@ int
 main(int argc, char **argv)
 {
     __pmSetProgname(argv[0]);
+    __pmGetUsername(&username);
     setlinebuf(stdout);
 
     /* PCP_COUNTER_WRAP in environment enables "counter wrap" logic */
