@@ -2,6 +2,7 @@
  * pragmatics.c - inference engine pragmatics analysis
  * 
  * Copyright (c) 1995-2003 Silicon Graphics, Inc.  All Rights Reserved.
+ * Copyright (c) 2013 Red Hat, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -15,7 +16,7 @@
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program; if not, write to the Free Software Foundation, Inc.,
- * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  * 
  * The analysis of how to organize the fetching of metrics (pragmatics),
  * and any other parts of the inference engine sensitive to details of
@@ -35,11 +36,9 @@
 
 extern char	*clientid;
 
-int	dfltConn;	/* default context type */
-
 /* for initialization of pmUnits struct */
-pmUnits noUnits    = {0, 0, 0, 0, 0, 0};
-pmUnits countUnits = {0, 0, 1, 0, 0, 0};
+pmUnits noUnits;
+pmUnits countUnits = { .dimCount = 1 };
 
 char *
 findsource(char *host)
@@ -362,10 +361,25 @@ findProfile(Fetch *f, Metric *m)
 		pmErrStr(sts));
 	exit(1);
     }
-    if (m->specinst == 0)
+
+    /*
+     * If any rule requires all instances, then ignore restricted
+     * instance lists from all other rules
+     */
+    if (m->specinst == 0 && p->need_all == 0) {
+	sts = pmDelProfile(p->indom, 0, (int *)0);
+	if (sts < 0) {
+	    fprintf(stderr, "%s: pmDelProfile failed: %s\n", pmProgname,
+		    pmErrStr(sts));
+	    exit(1);
+	}
 	sts = pmAddProfile(p->indom, 0, (int *)0);
-    else
+	p->need_all = 1;
+    }
+    else if (m->specinst > 0 && p->need_all == 0)
 	sts = pmAddProfile(p->indom, m->m_idom, m->iids);
+    else
+	sts = 0;
 
     if (sts < 0) {
 	fprintf(stderr, "%s: pmAddProfile failed: %s\n", pmProgname,
@@ -432,15 +446,36 @@ sendDesc(Expr *x, pmValueSet *vset)
 
     d.pmid = vset->pmid;
     d.indom = PM_INDOM_NULL;
-    if (x->sem != SEM_UNKNOWN) {
-	d.type = PM_TYPE_DOUBLE;
-	d.sem = x->sem;
-	d.units = x->units;
-    }
-    else {
-	d.type = PM_TYPE_NOSUPPORT;
-	d.sem = PM_SEM_INSTANT;
-	d.units = noUnits;
+    switch (x->sem) {
+	case PM_SEM_COUNTER:
+	case PM_SEM_INSTANT:
+	case PM_SEM_DISCRETE:
+	    /* these map directly to PMAPI semantics */
+	    d.type = PM_TYPE_DOUBLE;
+	    d.sem = x->sem;
+	    d.units = x->units;
+	    break;
+
+	case SEM_NUMVAR:
+	case SEM_NUMCONST:
+	case SEM_TRUTH:
+	    /* map to a numeric value */
+	    d.type = PM_TYPE_DOUBLE;
+	    d.sem = PM_SEM_INSTANT;
+	    d.units = x->units;
+	    break;
+
+	default:
+	    fprintf(stderr, "sendDesc(%s): botch sem=%d?\n", pmIDStr(d.pmid), x->sem);
+	    /* FALLTHROUGH */
+	case SEM_UNKNOWN:
+	case SEM_CHAR:
+	case SEM_REGEX:
+	    /* no mapping is possible */
+	    d.type = PM_TYPE_NOSUPPORT;
+	    d.sem = PM_SEM_INSTANT;
+	    d.units = noUnits;
+	    break;
     }
     __pmSendDesc(STDOUT_FILENO, pmWhichContext(), &d);
 }
@@ -458,7 +493,6 @@ initArchive(Archive *a)
     struct timeval  tv;
     int		    sts;
     int		    handle;
-    int		    n;
     Archive	    *b;
 
     /* setup temorary context for the archive */
@@ -470,6 +504,8 @@ initArchive(Archive *a)
     }
     handle = sts;
 
+    a->hname = strdup (pmGetContextHostName (handle));
+
     /* get the goodies from archive label */
     if ((sts = pmGetArchiveLabel(&label)) < 0) {
 	fprintf(stderr, "%s: cannot read label from archive %s\n"
@@ -478,9 +514,6 @@ initArchive(Archive *a)
 	pmDestroyContext(handle);
 	return 0;
     }
-    n = (int) strlen(label.ll_hostname);
-    a->hname = (char *) alloc(n + 1);
-    strcpy(a->hname, label.ll_hostname);
     a->first = realize(label.ll_start);
     if ((sts = pmGetArchiveEnd(&tv)) < 0) {
 	fprintf(stderr, "%s: archive %s is corrupted\n"
@@ -533,40 +566,40 @@ zoneInit(void)
 		    pmErrStr(sts));
     }
     else if (! archives && hostZone) {	/* TZ from live host */
-	if ((handle = pmNewContext(PM_CONTEXT_HOST, dfltHost)) < 0)
+	if ((handle = pmNewContext(PM_CONTEXT_HOST, dfltHostConn)) < 0)
 	    fprintf(stderr, "%s: cannot set timezone from %s\n"
 		    "pmNewContext failed: %s\n", pmProgname,
-		    findsource(dfltHost), pmErrStr(handle));
+		    findsource(dfltHostConn), pmErrStr(handle));
 	else if ((sts = pmNewContextZone()) < 0)
 	    fprintf(stderr, "%s: cannot set timezone from %s\n"
 		    "pmNewContextZone failed: %s\n", pmProgname,
-		    findsource(dfltHost), pmErrStr(sts));
+		    findsource(dfltHostConn), pmErrStr(sts));
 	else
 	    fprintf(stdout, "%s: timezone set to local timezone of host %s\n",
-		    pmProgname, dfltHost);
+		    pmProgname, dfltHostConn);
 	if (handle >= 0)
 	    pmDestroyContext(handle);
     }
     else if (hostZone) {		/* TZ from an archive */
 	a = archives;
 	while (a) {
-	    if (strcmp(dfltHost, a->hname) == 0)
+	    if (strcmp(dfltHostName, a->hname) == 0)
 		break;
 	    a = a->next;
 	}
 	if (! a)
 	    fprintf(stderr, "%s: no archive supplied for host %s\n",
-		    pmProgname, dfltHost);
+		    pmProgname, dfltHostName);
 	else if ((handle = pmNewContext(PM_CONTEXT_ARCHIVE, a->fname)) < 0)
 	    fprintf(stderr, "%s: cannot set timezone from %s\npmNewContext failed: %s\n",
-		    pmProgname, findsource(dfltHost), pmErrStr(handle));
+		    pmProgname, findsource(dfltHostName), pmErrStr(handle));
 	else if ((sts = pmNewContextZone()) < 0)
 	    fprintf(stderr, "%s: cannot set timezone from %s\n"
 		    "pmNewContextZone failed: %s\n",
-		    pmProgname, findsource(dfltHost), pmErrStr(sts));
+		    pmProgname, findsource(dfltHostName), pmErrStr(sts));
 	else
 	    fprintf(stdout, "%s: timezone set to local timezone of host %s\n",
-		    pmProgname, dfltHost);
+		    pmProgname, dfltHostName);
 	if (handle >= 0)
 	    pmDestroyContext(handle);
     }
@@ -586,7 +619,6 @@ canon(pmUnits in)
     return out;
 }
 
-
 /* scale factor to canonical pmUnits */
 double
 scale(pmUnits in)
