@@ -17,6 +17,7 @@
 #include <math.h>
 #include <ctype.h>
 #include <sys/stat.h>
+#include <assert.h>
 #include "pmapi.h"
 #include "impl.h"
 #include "logger.h"
@@ -62,7 +63,6 @@ Options:\n\
   -d             desperate, save output after fatal error\n\
   -f             choose timezone from first archive [default is to use\n\
                  timezone from last archive]\n\
-  -n pmnsfile    use an alternative PMNS\n\
   -S starttime   start of the time window\n\
   -s samples     terminate after this many log records have been written\n\
   -T endtime     end of the time window\n\
@@ -100,11 +100,6 @@ metricname(pmID pmid)
 #define MARK_FOR_WRITE		1
 #define WRITTEN			2
 
-#define BREAK_EOF		1
-#define BREAK_LOGEND		2
-#define BREAK_WINEND		3
-
-
 /*
  *  reclist_t is in logger.h
  *	(list of pdu's to write out at start of time window)
@@ -193,14 +188,13 @@ static struct timeval logend_tval = {0,0};	/* extracted log end */
 static struct timeval winstart_tval = {0,0};	/* window start tval*/
 static struct timeval winend_tval = {0,0};	/* window end tval*/
 
-static double 	winstart_time = -1.0;		/* window start time */
-static double	winend_time = -1.0;		/* window end time */
-static double	logend_time = -1.0;		/* log end time */
+static __pmTimeval 	winstart = {-1,0};	/* window start time */
+static __pmTimeval	winend = {-1,0};	/* window end time */
+static __pmTimeval	logend = {-1,0};	/* log end time */
 
 /* command line args */
 char	*configfile = NULL;		/* -c arg - name of config file */
 int	farg = 0;			/* -f arg - use first timezone */
-char	*pmnsfile = PM_NS_DEFAULT;	/* -n arg - alternate namespace */
 int	sarg = -1;			/* -s arg - finish after X samples */
 char	*Sarg = NULL;			/* -S arg - window start */
 char	*Targ = NULL;			/* -T arg - window end */
@@ -215,18 +209,22 @@ char	*Oarg = NULL;			/* -O arg - non-existent */
 
 /*--- START FUNCTIONS -------------------------------------------------------*/
 
-extern int _pmLogGet(__pmLogCtl *, int, __pmPDU **);
-extern int _pmLogPut(FILE *, __pmPDU *);
-extern void insertresult(rlist_t **, pmResult *);
-extern pmResult *searchmlist(pmResult *);
-
 /*
- *  convert timeval to double
+ * return -1, 0 or 1 as the __pmTimeval's compare
+ * a < b, a == b or a > b
  */
-double
-tv2double(struct timeval *tv)
+static int
+tvcmp(__pmTimeval a, __pmTimeval b)
 {
-    return tv->tv_sec + (double)tv->tv_usec / 1000000.0;
+    if (a.tv_sec < b.tv_sec)
+	return -1;
+    if (a.tv_sec > b.tv_sec)
+	return 1;
+    if (a.tv_usec < b.tv_usec)
+	return -1;
+    if (a.tv_usec > b.tv_usec)
+	return 1;
+    return 0;
 }
 
 static void
@@ -319,7 +317,7 @@ newlabel(void)
     inarchvers = iap->label.ll_magic & 0xff;
     outarchvers = inarchvers;
 
-    if (inarchvers != PM_LOG_VERS01 && inarchvers != PM_LOG_VERS02) {
+    if (inarchvers != PM_LOG_VERS02) {
 	fprintf(stderr,"%s: Error: illegal version number %d in archive (%s)\n",
 		pmProgname, inarchvers, iap->name);
 	abandon();
@@ -327,8 +325,9 @@ newlabel(void)
 
     /* copy magic number, pid, host and timezone */
     lp->ill_magic = iap->label.ll_magic;
-    lp->ill_pid = getpid();
+    lp->ill_pid = (int)getpid();
     strncpy(lp->ill_hostname, iap->label.ll_hostname, PM_LOG_MAXHOSTLEN);
+    lp->ill_hostname[PM_LOG_MAXHOSTLEN-1] = '\0';
     if (farg) {
 	/*
 	 * use timezone from first archive ... this is the OLD default
@@ -435,7 +434,7 @@ mk_reclist_t(void)
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_APPL0) {
         totalmalloc += sizeof(reclist_t);
-        printf ("mk_reclist_t: allocated %d\n", (int)sizeof(reclist_t));
+        fprintf(stderr, "mk_reclist_t: allocated %d\n", (int)sizeof(reclist_t));
     }
 #endif
     rec->pdu = NULL;
@@ -553,30 +552,30 @@ update_descreclist(int i)
     else {
 	curr = rdesc;
 	/* find matching record or last record */
-	while (curr->next != NULL && curr->desc.pmid != __ntohpmID(iap->pb[META][2]))
+	while (curr->next != NULL && curr->desc.pmid != ntoh_pmID(iap->pb[META][2]))
 	    curr = curr->next;
 
 #ifdef PCP_DEBUG
 	if (pmDebug & DBG_TRACE_APPL1) {
 	    fprintf(stderr, "update_descreclist: pmid: last/match %s", metricname(curr->desc.pmid));
-	    fprintf(stderr, " new %s", metricname(__ntohpmID(iap->pb[META][2])));
+	    fprintf(stderr, " new %s", metricname(ntoh_pmID(iap->pb[META][2])));
 	    fputc('\n', stderr);
 	}
 #endif
     }
 
-    if (curr->desc.pmid == __ntohpmID(iap->pb[META][2])) {
+    if (curr->desc.pmid == ntoh_pmID(iap->pb[META][2])) {
 #ifdef PCP_DEBUG
 	if (pmDebug & DBG_TRACE_APPL1) {
 	    fprintf(stderr, " type: old %s", pmTypeStr(curr->desc.type));
 	    fprintf(stderr, " new %s", pmTypeStr(ntohl(iap->pb[META][3])));
 	    fprintf(stderr, " indom: old %s", pmInDomStr(curr->desc.indom));
-	    fprintf(stderr, " new %s", pmInDomStr(__ntohpmInDom(iap->pb[META][4])));
+	    fprintf(stderr, " new %s", pmInDomStr(ntoh_pmInDom(iap->pb[META][4])));
 	    fprintf(stderr, " sem: old %d", curr->desc.sem);
 	    fprintf(stderr, " new %d", (int)ntohl(iap->pb[META][5]));
 	    fprintf(stderr, " units: old %s", pmUnitsStr(&curr->desc.units));
 	    pmup = (pmUnits *)&iap->pb[META][6];
-	    pmu = __ntohpmUnits(*pmup);
+	    pmu = ntoh_pmUnits(*pmup);
 	    fprintf(stderr, " new %s", pmUnitsStr(&pmu));
 	    fputc('\n', stderr);
 	}
@@ -589,12 +588,12 @@ update_descreclist(int i)
 	    fprintf(stderr, " to %s!\n", pmTypeStr(ntohl(iap->pb[META][3])));
 	    abandon();
 	}
-	if (curr->desc.indom != __ntohpmInDom(iap->pb[META][4])) {
+	if (curr->desc.indom != ntoh_pmInDom(iap->pb[META][4])) {
 	    fprintf(stderr,
 		"%s: Error: metric %s: indom changed from",
 		pmProgname, metricname(curr->desc.pmid));
 	    fprintf(stderr, " %s", pmInDomStr(curr->desc.indom));
-	    fprintf(stderr, " to %s!\n", pmInDomStr(__ntohpmInDom(iap->pb[META][4])));
+	    fprintf(stderr, " to %s!\n", pmInDomStr(ntoh_pmInDom(iap->pb[META][4])));
 	    abandon();
 	}
 	if (curr->desc.sem != ntohl(iap->pb[META][5])) {
@@ -606,7 +605,7 @@ update_descreclist(int i)
 	    abandon();
 	}
 	pmup = (pmUnits *)&iap->pb[META][6];
-	pmu = __ntohpmUnits(*pmup);
+	pmu = ntoh_pmUnits(*pmup);
 	if (curr->desc.units.dimSpace != pmu.dimSpace ||
 	    curr->desc.units.dimTime != pmu.dimTime ||
 	    curr->desc.units.dimCount != pmu.dimCount ||
@@ -629,12 +628,12 @@ update_descreclist(int i)
 	curr->next = mk_reclist_t();
 	curr = curr->next;
 	curr->pdu = iap->pb[META];
-	curr->desc.pmid = __ntohpmID(iap->pb[META][2]);
+	curr->desc.pmid = ntoh_pmID(iap->pb[META][2]);
 	curr->desc.type = ntohl(iap->pb[META][3]);
-	curr->desc.indom = __ntohpmInDom(iap->pb[META][4]);
+	curr->desc.indom = ntoh_pmInDom(iap->pb[META][4]);
 	curr->desc.sem = ntohl(iap->pb[META][5]);
 	pmup =(pmUnits *)&iap->pb[META][6];
-	curr->desc.units = __ntohpmUnits(*pmup);
+	curr->desc.units = ntoh_pmUnits(*pmup);
 	curr->ptr = findnadd_indomreclist(curr->desc.indom);
 	iap->pb[META] = NULL;
     }
@@ -655,9 +654,11 @@ append_indomreclist(int i)
     if (rindom == NULL) {
 	rindom = mk_reclist_t();
 	rindom->pdu = iap->pb[META];
+	rindom->stamp.tv_sec = ntohl(rindom->pdu[2]);
+	rindom->stamp.tv_usec = ntohl(rindom->pdu[3]);
 	rindom->desc.pmid = PM_ID_NULL;
 	rindom->desc.type = PM_TYPE_NOSUPPORT;
-	rindom->desc.indom = __ntohpmInDom(iap->pb[META][4]);
+	rindom->desc.indom = ntoh_pmInDom(iap->pb[META][4]);
 	rindom->desc.sem = 0;
 	rindom->desc.units = nullunits;	/* struct assignment */
     }
@@ -665,22 +666,26 @@ append_indomreclist(int i)
 	curr = rindom;
 
 	/* find matching record or last record */
-	while (curr->next != NULL && curr->desc.indom != __ntohpmInDom(iap->pb[META][4])) {
+	while (curr->next != NULL && curr->desc.indom != ntoh_pmInDom(iap->pb[META][4])) {
 	    curr = curr->next;
 	}
 
-	if (curr->desc.indom == __ntohpmInDom(iap->pb[META][4])) {
+	if (curr->desc.indom == ntoh_pmInDom(iap->pb[META][4])) {
 	    if (curr->pdu == NULL) {
 		/* insert new record */
 		curr->pdu = iap->pb[META];
+		curr->stamp.tv_sec = ntohl(curr->pdu[2]);
+		curr->stamp.tv_usec = ntohl(curr->pdu[3]);
 	    }
 	    else {
 		/* do NOT discard old record; insert new record */
 		rec = mk_reclist_t();
 		rec->pdu = iap->pb[META];
+		rec->stamp.tv_sec = ntohl(rec->pdu[2]);
+		rec->stamp.tv_usec = ntohl(rec->pdu[3]);
 		rec->desc.pmid = PM_ID_NULL;
 		rec->desc.type = PM_TYPE_NOSUPPORT;
-		rec->desc.indom = __ntohpmInDom(iap->pb[META][4]);
+		rec->desc.indom = ntoh_pmInDom(iap->pb[META][4]);
 		rec->desc.sem = 0;
 		rec->desc.units = nullunits;	/* struct assignment */
 		rec->next = curr->next;
@@ -692,9 +697,11 @@ append_indomreclist(int i)
 	    curr->next = mk_reclist_t();
 	    curr = curr->next;
 	    curr->pdu = iap->pb[META];
+	    curr->stamp.tv_sec = ntohl(curr->pdu[2]);
+	    curr->stamp.tv_usec = ntohl(curr->pdu[3]);
 	    curr->desc.pmid = PM_ID_NULL;
 	    curr->desc.type = PM_TYPE_NOSUPPORT;
-	    curr->desc.indom = __ntohpmInDom(iap->pb[META][4]);
+	    curr->desc.indom = ntoh_pmInDom(iap->pb[META][4]);
 	    curr->desc.sem = 0;
 	    curr->desc.units = nullunits;	/* struct assignment */
 	}
@@ -718,6 +725,67 @@ write_rec(reclist_t *rec)
 	    abandon();
 	}
 
+#ifdef PCP_DEBUG
+	if (pmDebug & DBG_TRACE_LOGMETA) {
+	    __pmLogHdr	*h;
+	    int		len;
+	    int		type;
+	    h = (__pmLogHdr *)rec->pdu;
+	    len = ntohl(h->len);
+	    type = ntohl(h->type);
+	    fprintf(stderr, "write_rec: record len=%d, type=%d @ offset=%d\n",
+	    	len, type, (int)(ftell(logctl.l_mdfp) - sizeof(__pmLogHdr)));
+	    if (type == TYPE_DESC) {
+		pmDesc	*dp;
+		pmDesc	desc;
+		int	*namelen;
+		char	*name;	/* just first name for diag */
+		dp = (pmDesc *)((void *)rec->pdu + sizeof(__pmLogHdr));
+		desc.type = ntohl(dp->type);
+		desc.sem = ntohl(dp->sem);
+		desc.indom = ntoh_pmInDom(dp->indom);
+		desc.units = ntoh_pmUnits(dp->units);
+		desc.pmid = ntoh_pmID(dp->pmid);
+		namelen = (int *)((void *)rec->pdu + sizeof(__pmLogHdr) + sizeof(pmDesc) + sizeof(int));
+		len = ntohl(*namelen);
+		name = (char *)((void *)rec->pdu + sizeof(__pmLogHdr) + sizeof(pmDesc) + sizeof(int) + sizeof(int));
+		fprintf(stderr, "PMID: %s name: %*.*s\n", pmIDStr(desc.pmid), len, len, name);
+		__pmPrintDesc(stderr, &desc);
+	    }
+	    else if (type == TYPE_INDOM) {
+		__pmTimeval	*tvp;
+		__pmTimeval	when;
+		int		k = 2;
+		pmInDom		indom;
+		int		numinst;
+		int		*instlist;
+		int		inst;
+
+		tvp = (__pmTimeval *)&rec->pdu[k];
+		when.tv_sec = ntohl(tvp->tv_sec);
+		when.tv_usec = ntohl(tvp->tv_usec);
+		k += sizeof(__pmTimeval)/sizeof(rec->pdu[0]);
+		indom = ntoh_pmInDom((unsigned int)rec->pdu[k++]);
+		fprintf(stderr, "INDOM: %s when: ", pmInDomStr(indom));
+		__pmPrintTimeval(stderr, &when);
+		numinst = ntohl(rec->pdu[k++]);
+		fprintf(stderr, " numinst: %d", numinst);
+		if (numinst > 0) {
+		    int		i;
+		    instlist = (int *)&rec->pdu[k];
+		    for (i = 0; i < numinst; i++) {
+			inst = ntohl(instlist[i]);
+			fprintf(stderr, " [%d] %d", i, inst);
+		    }
+		}
+		fputc('\n', stderr);
+	    }
+	    else {
+		fprintf(stderr, "Botch: bad type\n");
+	    }
+	}
+#endif
+
 	/* write out the pdu ; exit if write failed */
 	if ((sts = _pmLogPut(logctl.l_mdfp, rec->pdu)) < 0) {
 	    fprintf(stderr, "%s: Error: _pmLogPut: meta data : %s\n",
@@ -805,42 +873,33 @@ write_metareclist(pmResult *result, int *needti)
 	     * to traverse the entire list
 	     *	- we can safely ignore all indoms after the current timestamp
 	     *	- we want the latest indom at, or before the current timestamp
-	     *  - all others before the current timestamp can be discarded(?)
 	     */
 	    othr_indom = NULL;
 	    while (curr_indom != NULL && curr_indom->desc.indom == indom) {
-		if (curr_indom->written != WRITTEN) {
-		    if (curr_indom->pdu == NULL) {
-			/* indom is in list, has not been written,
-			 * but has no pdu - this is possible and acceptable
-			 * behaviour; do nothing
-			 */
-	    	    }
-		    else if (ntohl(curr_indom->pdu[2]) < this->tv_sec ||
-			(ntohl(curr_indom->pdu[2]) == this->tv_sec &&
-			ntohl(curr_indom->pdu[3]) <= this->tv_usec))
+		if (curr_indom->stamp.tv_sec < this->tv_sec ||
+		         (curr_indom->stamp.tv_sec == this->tv_sec &&
+		          curr_indom->stamp.tv_usec <= this->tv_usec))
+		{
+		    /* indom is in list, indom has pdu
+		     * and timestamp in pdu suits us
+		     */
+		    if (othr_indom == NULL) {
+			othr_indom = curr_indom;
+		    }
+		    else if (othr_indom->stamp.tv_sec < curr_indom->stamp.tv_sec ||
+			     (othr_indom->stamp.tv_sec == curr_indom->stamp.tv_sec &&
+			      othr_indom->stamp.tv_usec <= curr_indom->stamp.tv_usec))
 		    {
-			/* indom is in list, indom has pdu
-			 * and timestamp in pdu suits us
+			/* we already have a perfectly good indom,
+			 * but curr_indom has a better timestamp
 			 */
-			if (othr_indom == NULL) {
-			    othr_indom = curr_indom;
-			}
-			else if (ntohl(othr_indom->pdu[2]) < ntohl(curr_indom->pdu[2]) ||
-				 (ntohl(othr_indom->pdu[2]) == ntohl(curr_indom->pdu[2]) &&
-				 ntohl(othr_indom->pdu[3]) <= ntohl(curr_indom->pdu[3])))
-			{
-			    /* we already have a perfectly good indom,
-			     * but curr_indom has a better timestamp
-			     */
-			    othr_indom = curr_indom;
-			}
+			othr_indom = curr_indom;
 		    }
 		}
 		curr_indom = curr_indom->next;
 	    } /*while()*/
 
-	    if (othr_indom != NULL) {
+	    if (othr_indom != NULL && othr_indom->pdu != NULL && othr_indom->written != WRITTEN) {
 		othr_indom->written = MARK_FOR_WRITE;
 		othr_indom->pdu[2] = htonl(this->tv_sec);
 		othr_indom->pdu[3] = htonl(this->tv_usec);
@@ -873,7 +932,7 @@ _createmark(void)
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_APPL0) {
         totalmalloc += sizeof(mark_t);
-        printf ("_createmark : allocated %d\n", (int)sizeof(mark_t));
+        fprintf(stderr, "_createmark : allocated %d\n", (int)sizeof(mark_t));
     }
 #endif
 
@@ -919,7 +978,6 @@ nextmeta(void)
     __pmLogCtl	*lcp;
     __pmContext	*ctxp;
     inarch_t	*iap;			/* pointer to input archive control */
-    __pmHashNode	*hnp;
 
     for (i=0; i<inarchnum; i++) {
 	iap = &inarch[i];
@@ -938,7 +996,10 @@ nextmeta(void)
 	    fprintf(stderr, "    iap->pb[META] is not NULL\n");
 	    abandon();
 	}
-	ctxp = __pmHandleToPtr(iap->ctx);
+	if ((ctxp = __pmHandleToPtr(iap->ctx)) == NULL) {
+	    fprintf(stderr, "%s: botch: __pmHandleToPtr(%d) returns NULL!\n", pmProgname, iap->ctx);
+	    abandon();
+	}
 	lcp = ctxp->c_archctl->ac_log;
 
 againmeta:
@@ -952,6 +1013,7 @@ againmeta:
 			pmProgname, iap->name, pmErrStr(sts));
 		_report(lcp->l_mdfp);
 	    }
+	    PM_UNLOCK(ctxp->c_lock);
 	    continue;
 	}
 
@@ -959,7 +1021,7 @@ againmeta:
 	 *	then append to desc list
 	 */
 	if (ntohl(iap->pb[META][1]) == TYPE_DESC) {
-	    pmid = __ntohpmID(iap->pb[META][2]);
+	    pmid = ntoh_pmID(iap->pb[META][2]);
 
 	    /* if ml is defined, then look for pmid in the list
 	     * if pmid is not in the list then discard it immediately
@@ -975,9 +1037,8 @@ againmeta:
 	    }
 
 	    if (want) {
-		if ((hnp = __pmHashSearch((int)pmid, &mdesc_hash)) == NULL) {
+		if (__pmHashSearch((int)pmid, &mdesc_hash) == NULL)
 		    __pmHashAdd((int)pmid, NULL, &mdesc_hash);
-		}
 		/*
 		 * update the desc list (add first time, check on subsequent
 		 * sightings of desc for this pmid from this source
@@ -996,7 +1057,7 @@ againmeta:
 	    /* if ml is defined, then look for instance domain in the list
 	     * if indom is not in the list then discard it immediately
 	     */
-	    indom = __ntohpmInDom(iap->pb[META][4]);
+	    indom = ntoh_pmInDom(iap->pb[META][4]);
 	    want = 0;
 	    if (ml == NULL)
 	        want = 1;
@@ -1031,6 +1092,7 @@ againmeta:
 	    abandon();
 	}
 
+	PM_UNLOCK(ctxp->c_lock);
     }
 
     if (numeof == inarchnum) return(-1);
@@ -1047,7 +1109,7 @@ nextlog(void)
     int		i;
     int		eoflog = 0;	/* number of log files at eof */
     int		sts;
-    double	curtime;
+    __pmTimeval	curtime;
     __pmLogCtl	*lcp;
     __pmContext	*ctxp;
     inarch_t	*iap;
@@ -1077,11 +1139,14 @@ nextlog(void)
 	    continue;
 	}
 
-againlog:
-	ctxp = __pmHandleToPtr(iap->ctx);
+	if ((ctxp = __pmHandleToPtr(iap->ctx)) == NULL) {
+	    fprintf(stderr, "%s: botch: __pmHandleToPtr(%d) returns NULL!\n", pmProgname, iap->ctx);
+	    abandon();
+	}
 	lcp = ctxp->c_archctl->ac_log;
 
-	if ((sts=__pmLogRead(lcp,PM_MODE_FORW,NULL,&iap->_result)) < 0) {
+againlog:
+	if ((sts=__pmLogRead(lcp, PM_MODE_FORW, NULL, &iap->_result, PMLOGREAD_NEXT)) < 0) {
 	    if (sts != PM_ERR_EOL) {
 		fprintf(stderr, "%s: Error: __pmLogRead[log %s]: %s\n",
 			pmProgname, iap->name, pmErrStr(sts));
@@ -1100,27 +1165,27 @@ againlog:
 		iap->mark = 1;
 		iap->pb[LOG] = _createmark();
 	    }
+	    PM_UNLOCK(ctxp->c_lock);
 	    continue;
 	}
+	assert(iap->_result != NULL);
 
 
 	/* set current log time - this is only done so that we can
 	 * determine whether to keep or discard the log
 	 */
-	curtime = iap->_result->timestamp.tv_sec +
-			(double)iap->_result->timestamp.tv_usec / 1000000.0;
+	curtime.tv_sec = iap->_result->timestamp.tv_sec;
+	curtime.tv_usec = iap->_result->timestamp.tv_usec;
 
 	/* if log time is greater than (or equal to) the current window
 	 * start time, then we may want it
 	 *	(irrespective of the current window end time)
 	 */
-	if (curtime < winstart_time) {
+	if (tvcmp(curtime, winstart) < 0) {
 	    /* log is not in time window - discard result and get next record
 	     */
-	    if (iap->_result != NULL) {
-		pmFreeResult(iap->_result);
-		iap->_result = NULL;
-	    }
+	    pmFreeResult(iap->_result);
+	    iap->_result = NULL;
 	    goto againlog;
 	}
         else {
@@ -1141,13 +1206,13 @@ againlog:
             if (iap->_Nresult == NULL) {
                 /* dont want any of the metrics in _result, try again
                  */
-                if (iap->_result != NULL) {
-                        pmFreeResult(iap->_result);
-                        iap->_result = NULL;
-                }
+		pmFreeResult(iap->_result);
+		iap->_result = NULL;
                 goto againlog;
             }
 	}
+
+	PM_UNLOCK(ctxp->c_lock);
     } /*for(i)*/
 
     /* if we are here, then each archive control struct should either
@@ -1171,7 +1236,7 @@ parseargs(int argc, char *argv[])
     char		*endnum;
     struct stat		sbuf;
 
-    while ((c = getopt(argc, argv, "c:D:dfn:S:s:T:v:wZ:z?")) != EOF) {
+    while ((c = getopt(argc, argv, "c:D:dfS:s:T:v:wZ:z?")) != EOF) {
 	switch (c) {
 
 	case 'c':	/* config file */
@@ -1203,12 +1268,6 @@ parseargs(int argc, char *argv[])
 	    farg = 1;
 	    break;
 
-	case 'n':	/* namespace */
-			/* if namespace is reassigned from config file,
-			 * must reload namespace */
-	    pmnsfile = optarg;
-	    break;
-
 	case 's':	/* number of samples to write out */
 	    sarg = (int)strtol(optarg, &endnum, 10);
 	    if (*endnum != '\0' || sarg < 0) {
@@ -1285,7 +1344,6 @@ int
 parseconfig(void)
 {
     int		errflag = 0;
-    extern FILE * yyin;
 
     if ((yyin = fopen(configfile, "r")) == NULL) {
 	fprintf(stderr, "%s: Cannot open config file \"%s\": %s\n",
@@ -1302,51 +1360,21 @@ parseconfig(void)
     return(-errflag);
 }
 
-void
-adminarch(void)
-{
-    int		i;
-    int		sts;
-
-    if (pmnsfile != PM_NS_DEFAULT || inarchvers == PM_LOG_VERS01) {
-	if ((sts = pmLoadNameSpace (pmnsfile)) < 0) {
-	    fprintf(stderr, "%s: Error: cannot load name space: %s\n",
-		    pmProgname, pmErrStr(sts));
-	    exit(1);
-	}
-
-	for (i=0; i<inarchnum; i++) {
-	    if ((sts = pmUseContext(inarch[i].ctx)) < 0) {
-		fprintf(stderr,
-		    "%s: Error: cannot use context (%d) from archive \"%s\"\n",
-			pmProgname, inarch[i].ctx, inarch[i].name);
-		exit(1);
-	    }
-
-	    if ((sts = pmTrimNameSpace ()) < 0) {
-		fprintf(stderr, "%s: Error: cannot trim name space: %s\n", 
-			pmProgname, pmErrStr(sts));
-		exit(1);
-	    }
-	}
-    }
-}
-
 /*
  *  we are within time window ... return 0
  *  we are outside of time window & mk new window ... return 1
  *  we are outside of time window & exit ... return -1
  */
 static int
-checkwinend(double now)
+checkwinend(__pmTimeval now)
 {
     int		i;
     int		sts;
-    double	tmptime;
+    __pmTimeval	tmptime;
     inarch_t	*iap;
     __pmPDU	*markpdu;	/* mark b/n time windows */
 
-    if (winend_time < 0 || now <= winend_time)
+    if (winend.tv_sec < 0 || tvcmp(now, winend) <= 0)
 	return(0);
 
     /* we have reached the end of a window
@@ -1358,26 +1386,26 @@ checkwinend(double now)
     if (!warg)
 	return(-1);
 
-    winstart_time += NUM_SEC_PER_DAY;
-    winend_time += NUM_SEC_PER_DAY;
+    winstart.tv_sec += NUM_SEC_PER_DAY;
+    winend.tv_sec += NUM_SEC_PER_DAY;
     pre_startwin = 1;
 
     /* if start of next window is later than max termination
      * then bail out here
      */
-    if (winstart_time > logend_time)
+    if (tvcmp(winstart, logend) > 0)
 	    return(-1);
 
     ilog = -1;
     for (i=0; i<inarchnum; i++) {
 	iap = &inarch[i];
 	if (iap->_Nresult != NULL) {
-	    tmptime = iap->_Nresult->timestamp.tv_sec
-			+ (double)iap->_Nresult->timestamp.tv_usec/1000000.0;
-	    if (tmptime < winstart_time) {
+	    tmptime.tv_sec = iap->_Nresult->timestamp.tv_sec;
+	    tmptime.tv_usec = iap->_Nresult->timestamp.tv_usec;
+	    if (tvcmp(tmptime, winstart) < 0) {
 		/* free _result and _Nresult
 		 */
-		if (iap->_result != iap->_Nresult && iap->_Nresult != NULL) {
+		if (iap->_result != iap->_Nresult) {
 		    free(iap->_Nresult);
 		}
 		if (iap->_result != NULL) {
@@ -1389,8 +1417,9 @@ checkwinend(double now)
 	    }
 	}
 	if (iap->pb[LOG] != NULL) {
-	    tmptime = ntohl(iap->pb[LOG][3]) + (double)ntohl(iap->pb[LOG][4])/1000000.0;
-	    if (tmptime < winstart_time) {
+	    tmptime.tv_sec = ntohl(iap->pb[LOG][3]);
+	    tmptime.tv_usec = ntohl(iap->pb[LOG][4]);
+	    if (tvcmp(tmptime, winstart) < 0) {
 		/* free PDU buffer ... it is probably a mark
 		 * and has not been pinned
 		 */
@@ -1418,27 +1447,27 @@ checkwinend(double now)
  *
  */
 void
-writerlist(rlist_t **rlready, double mintime)
+writerlist(rlist_t **rlready, __pmTimeval mintime)
 {
     int		sts;
-    int		needti;		/* need to flush/update */
-    double	titime;		/* time of last temporal index write */
-    double	restime;	/* time of result */
+    int		needti = 0;	/* need to flush/update */
+    __pmTimeval	titime  = {0,0};/* time of last temporal index write */
+    __pmTimeval	restime;	/* time of result */
     rlist_t	*elm;		/* element of rlready to be written out */
     __pmPDU	*pb;		/* pdu buffer */
-    __pmTimeval	this;		/* timeval of this record */
     unsigned long	peek_offset;
 
-    needti = 0;
-    titime = 0.0;
-
     while (*rlready != NULL) {
-	this.tv_sec = (*rlready)->res->timestamp.tv_sec;
-	this.tv_usec = (*rlready)->res->timestamp.tv_usec;
-	restime = this.tv_sec + (double)this.tv_usec / 1000000.0;
+	restime.tv_sec = (*rlready)->res->timestamp.tv_sec;
+	restime.tv_usec = (*rlready)->res->timestamp.tv_usec;
 
-        if (restime > mintime)
+        if (tvcmp(restime, mintime) > 0) {
+#if 0
+fprintf(stderr, "writelist: restime %d.%06d mintime %d.%06d ", restime.tv_sec, restime.tv_usec, mintime.tv_sec, mintime.tv_usec);
+fprintf(stderr, " break!\n");
+#endif
 	    break;
+	}
 
 	/* get the first element from the list
 	 */
@@ -1467,21 +1496,13 @@ writerlist(rlist_t **rlready, double mintime)
 
 	/* convert log record to a pdu
 	 */
-	if (outarchvers == 1)
-	    sts = __pmEncodeResult(PDU_OVERRIDE1, elm->res, &pb);
-	else
-	    sts = __pmEncodeResult(PDU_OVERRIDE2, elm->res, &pb);
-
+	sts = __pmEncodeResult(PDU_OVERRIDE2, elm->res, &pb);
 	if (sts < 0) {
 	    fprintf(stderr, "%s: Error: __pmEncodeResult: %s\n",
 		    pmProgname, pmErrStr(sts));
 	    abandon();
 	}
 
-	/* __pmEncodeResult doesn't pin the PDU buffer, so we have to
-	 */
-	__pmPinPDUBuf(pb);
-
         /* switch volumes if required */
         if (varg > 0) {
             if (written > 0 && (written % varg) == 0) {
@@ -1502,8 +1523,9 @@ writerlist(rlist_t **rlready, double mintime)
 	 */
 	write_metareclist(elm->res, &needti);
 
-
 	/* write out log record */
+	old_log_offset = ftell(logctl.l_mfp);
+	assert(old_log_offset >= 0);
 	if ((sts = __pmLogPutResult(&logctl, pb)) < 0) {
 	    fprintf(stderr, "%s: Error: __pmLogPutResult: log data: %s\n",
 		    pmProgname, pmErrStr(sts));
@@ -1521,7 +1543,7 @@ writerlist(rlist_t **rlready, double mintime)
 	/* make sure that we do not write out the temporal index more
 	 * than once for the same timestamp
 	 */
-	if (needti && titime >= restime)
+	if (needti && tvcmp(titime, restime) >= 0)
 	    needti = 0;
 
 	/* flush/update */
@@ -1535,23 +1557,27 @@ writerlist(rlist_t **rlready, double mintime)
 		old_log_offset = sizeof(__pmLogLabel)+2*sizeof(int);
 
             new_log_offset = ftell(logctl.l_mfp);
+	    assert(new_log_offset >= 0);
             new_meta_offset = ftell(logctl.l_mdfp);
+	    assert(new_meta_offset >= 0);
 
             fseek(logctl.l_mfp, (long)old_log_offset, SEEK_SET);
             fseek(logctl.l_mdfp, (long)old_meta_offset, SEEK_SET);
 
-            __pmLogPutIndex(&logctl, &this);
+            __pmLogPutIndex(&logctl, &restime);
 
             fseek(logctl.l_mfp, (long)new_log_offset, SEEK_SET);
             fseek(logctl.l_mdfp, (long)new_meta_offset, SEEK_SET);
 
             old_log_offset = ftell(logctl.l_mfp);
+	    assert(old_log_offset >= 0);
             old_meta_offset = ftell(logctl.l_mdfp);
+	    assert(old_meta_offset >= 0);
 
             flushsize = ftell(logctl.l_mfp) + 100000;
         }
 
-	/* LOG: free PDU buffer */
+	/* free PDU buffer */
 	__pmUnpinPDUBuf(pb);
 	pb = NULL;
 
@@ -1611,13 +1637,12 @@ main(int argc, char **argv)
     int		sts;
     int		stslog;			/* sts from nextlog() */
     int		stsmeta;		/* sts from nextmeta() */
-    int		breakflag;		/* reason for breaking out of while */
 
     char	*msg;
 
-    double 	now = 0.0;		/* the current time */
-    double 	mintime = 0.0;
-    double 	tmptime = 0.0;
+    __pmTimeval 	now = {0,0};	/* the current time */
+    __pmTimeval 	mintime = {0,0};
+    __pmTimeval 	tmptime = {0,0};
 
     __pmTimeval		tstamp;		/* temporary timestamp */
     inarch_t		*iap;		/* ptr to archive control */
@@ -1657,7 +1682,7 @@ main(int argc, char **argv)
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_APPL0) {
         totalmalloc += (inarchnum * sizeof(inarch_t));
-        printf ("main        : allocated %d\n",
+        fprintf(stderr, "main        : allocated %d\n",
 			(int)(inarchnum * sizeof(inarch_t)));
     }
 #endif
@@ -1728,13 +1753,6 @@ main(int argc, char **argv)
     logctl.l_label.ill_start.tv_sec = logstart_tval.tv_sec;
     logctl.l_label.ill_start.tv_usec = logstart_tval.tv_usec;
 
-
-    /* admin archive loads the namespace if required, and trims it
-     * according to the context of each archive
-     */
-    adminarch();
-
-
     /* process config file
      *	- this includes a list of metrics and their instances
      */
@@ -1804,12 +1822,15 @@ main(int argc, char **argv)
 		pmProgname, msg);
 	abandon();
     }
-    winstart_time = tv2double(&winstart_tval);
-    winend_time = tv2double(&winend_tval);
-    logend_time = tv2double(&logend_tval);
+    winstart.tv_sec = winstart_tval.tv_sec;
+    winstart.tv_usec = winstart_tval.tv_usec;
+    winend.tv_sec = winend_tval.tv_sec;
+    winend.tv_usec = winend_tval.tv_usec;
+    logend.tv_sec = logend_tval.tv_sec;
+    logend.tv_usec = logend_tval.tv_usec;
 
     if (warg) {
-	if (winstart_time + NUM_SEC_PER_DAY < winend_time) {
+	if (winstart.tv_sec + NUM_SEC_PER_DAY < winend.tv_sec) {
 	    fprintf(stderr, "%s: Warning: -S and -T must specify a time window within\nthe same day, for -w to be used.  Ignoring -w flag.\n", pmProgname);
 	    warg = 0;
 	}
@@ -1823,7 +1844,6 @@ main(int argc, char **argv)
     current.tv_usec = 0;
     first_datarec = 1;
     pre_startwin = 1;
-    breakflag = 0;
 
     /* get all meta data first
      * nextmeta() should return 0 (will return -1 when all meta is eof)
@@ -1842,23 +1862,21 @@ main(int argc, char **argv)
 	ilog = -1;
 	curlog.tv_sec = 0;
 	curlog.tv_usec = 0;
-	old_log_offset = ftell(logctl.l_mfp);
 	old_meta_offset = ftell(logctl.l_mdfp);
+	assert(old_meta_offset >= 0);
 
 	/* nextlog() resets ilog, and curlog (to the smallest timestamp)
 	 */
 	stslog = nextlog();
 
-	if (stslog < 0) {
-	    breakflag = BREAK_EOF;
+	if (stslog < 0)
 	    break;
-	}
 
 	/* find the _Nresult (or mark pdu) with the earliest timestamp;
 	 * set ilog
 	 * (this is a bit more complex when tflag is specified)
 	 */
-	mintime = 0.0;
+	mintime.tv_sec = mintime.tv_usec = 0;
 	for (i=0; i<inarchnum; i++) {
 	    if (inarch[i]._Nresult != NULL) {
 		tstamp.tv_sec = inarch[i]._Nresult->timestamp.tv_sec;
@@ -1866,8 +1884,8 @@ main(int argc, char **argv)
 		checklogtime(&tstamp, i);
 
 		if (ilog == i) {
-		    tmptime = curlog.tv_sec + (double)curlog.tv_usec/1000000.0;
-		    if (mintime <= 0 || mintime > tmptime)
+		    tmptime = curlog;
+		    if (mintime.tv_sec <= 0 || tvcmp(mintime, tmptime) > 0)
 		        mintime = tmptime;
 		}
 	    }
@@ -1877,8 +1895,8 @@ main(int argc, char **argv)
 		checklogtime(&tstamp, i);
 
 		if (ilog == i) {
-		    tmptime = curlog.tv_sec + (double)curlog.tv_usec/1000000.0;
-		    if (mintime <= 0 || mintime > tmptime)
+		    tmptime = curlog;
+		    if (mintime.tv_sec <= 0 || tvcmp(mintime, tmptime) > 0)
 		        mintime = tmptime;
 		}
 	    }
@@ -1889,26 +1907,21 @@ main(int argc, char **argv)
 	 * mintime == now or timestamp of the earliest mark
 	 *		(whichever is smaller)
 	 */
-	now = curlog.tv_sec + (double)curlog.tv_usec / 1000000.0;
+	now = curlog;
 
 	/* note - mark (after last archive) will be created, but this
 	 * break, will prevent it from being written out
 	 */
-	if (now > logend_time) {
-	    breakflag = BREAK_LOGEND;
+	if (tvcmp(now, logend) > 0)
 	    break;
-	}
 
 	sts = checkwinend(now);
-	if (sts < 0) {
-	    breakflag = BREAK_WINEND;
+	if (sts < 0)
 	    break;
-	}
 	if (sts > 0)
 	    continue;
 
-	current.tv_sec = (long)now;
-	current.tv_usec = (now - (double)current.tv_sec) * 1000000.0;
+	current = curlog;
 
 	/* prepare to write out log record
 	 */
@@ -1918,6 +1931,7 @@ main(int argc, char **argv)
 	    abandon();
 	}
 
+
 	iap = &inarch[ilog];
 	if (iap->mark)
 	    writemark(iap);
@@ -1930,7 +1944,20 @@ main(int argc, char **argv)
 		abandon();
 	    }
 	    insertresult(&rlready, iap->_Nresult);
-	    writerlist(&rlready, now);
+#if 0
+{
+    rlist_t	*rp;
+    int		i;
+
+    fprintf(stderr, "rlready");
+    for (i = 0, rp = rlready; rp != NULL; i++, rp = rp->next) {
+	fprintf(stderr, " [%d] t=%d.%06d numpmid=%d", i, (int)rp->res->timestamp.tv_sec, (int)rp->res->timestamp.tv_usec, rp->res->numpmid);
+    }
+    fprintf(stderr, " now=%d.%06d\n", now.tv_sec, now.tv_usec);
+}
+#endif
+
+	    writerlist(&rlready, curlog);
 
 	    /* writerlist frees elm (elements of rlready) but does not
 	     * free _result & _Nresult
@@ -1954,9 +1981,10 @@ main(int argc, char **argv)
 		}
 		free(iap->_Nresult);
 	    }
-	    if (iap->_result != NULL)
+	    if (iap->_result != NULL) {
 		pmFreeResult(iap->_result);
-	    iap->_result = NULL;
+		iap->_result = NULL;
+	    }
 	    iap->_Nresult = NULL;
 	}
     } /*while()*/
@@ -1976,13 +2004,16 @@ cleanup:
 	    old_log_offset = sizeof(__pmLogLabel)+2*sizeof(int);
 
 	new_log_offset = ftell(logctl.l_mfp);
+	assert(new_log_offset >= 0);
 	new_meta_offset = ftell(logctl.l_mdfp);
+	assert(new_meta_offset >= 0);
 
 #if 0
-	fprintf(stderr, "*** last tstamp: \n\tmintime=%g \n\ttmptime=%g \n\tlogend_time=%g \n\twinend_time=%g \n\tcurrent=%d.%06d\n",
-	    mintime, tmptime, logend_time, winend_time, current.tv_sec, current.tv_usec);
+	fprintf(stderr, "*** last tstamp: \n\tmintime=%d.%06d \n\ttmptime=%d.%06d \n\tlogend=%d.%06d \n\twinend=%d.%06d \n\tcurrent=%d.%06d\n",
+	    mintime.tv_sec, mintime.tv_usec, tmptime.tv_sec, tmptime.tv_usec, logend.tv_sec, logend.tv_usec, winend.tv_sec, winend.tv_usec, current.tv_sec, current.tv_usec);
 #endif
 
+	fseek(logctl.l_mfp, old_log_offset, SEEK_SET);
 	__pmLogPutIndex(&logctl, &current);
 
 
@@ -1991,10 +2022,9 @@ cleanup:
     }
 #ifdef PCP_DEBUG
     if (pmDebug & DBG_TRACE_APPL0) {
-        printf ("main        : total allocated %ld\n", totalmalloc);
+        fprintf(stderr, "main        : total allocated %ld\n", totalmalloc);
     }
 #endif
 
     exit(exit_status);
-    return(0);
 }
